/***************************************************************************
                          cnf.cpp  -  description
                             -------------------
    begin                : Thu Jul 9 2006
    copyright            : (C) 2006-2009  Marian VooDooMan Meravy
    email                : ghostvoodooman@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************

XTelnet - a simple telnet client
Copyright (C) 2006-2009  Marian VooDooMan Meravy
  (ghostvoodooman@users.sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

#ifndef __CNF_H__
#define __CNF_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>

// for all others, include the necessary headers
#ifndef WX_PRECOMP
#   include <wx/frame.h>
#   include <wx/panel.h>
#   include <wx/layout.h>
#   include <wx/checkbox.h>
#endif

#include <wx/textctrl.h>
#include <wx/event.h>
#include <wx/checkbox.h>
#include <wx/radiobut.h>
#include <wx/stattext.h>
#include <wx/button.h>
#include <wx/filedlg.h>

#ifdef _WIN32
#   include <winsock2.h>
#   include <Ws2tcpip.h>
#   include <windows.h>
#endif

#ifdef __BORLANDC__
#   pragma hdrstop
#endif

#include "ssl.h"
#include "sock.h"

using namespace std;

struct sClient
{
    string host;
    unsigned short port;

    struct
    {
        bool bUTF8;
        bool bLocalEcho;
        bool bUseEscape;
        int iEOL;
        bool bHighlight;
        bool bSSL;
        string ca;
        string client_cert;
        string client_key;
    } Conf;

    bool bHideOutput;

    bool bEnteringPort;

    FILE* log;
    s_socket* sc;
    c_ssl* sc_ssl;

    sClient()
    {
        log=NULL;
        sc=NULL;
        sc_ssl=NULL;

        port=0;

        Conf.bUTF8=false;
        Conf.bLocalEcho=false;
        Conf.bUseEscape=false;
        Conf.iEOL=0;
        Conf.bHighlight=false;
        bHideOutput=false;

        Conf.bSSL=false;
        
        bEnteringPort=false;
    }
};

enum {
    Conf_UTF8 = 1000,
    Conf_Highlight,
    Conf_LocalEcho,
    Conf_EscapeChars,
    Conf_Eol_rn,
    Conf_Eol_n,
    Conf_Eol_r,
    Conf_Eol_nr,
    Conf_ssl,

    Conf_txt1,
    Conf_cas,
    Conf_cas_btn,

    Conf_txt2,
    Conf_crt,
    Conf_crt_btn,

    Conf_txt3,
    Conf_key,
    Conf_key_btn,

    Conf_close,

    Conf_ok_btn,

    Conf_enter_key,
};

class c_cnf_panel : public wxPanel
{
public:
    c_cnf_panel(wxFrame *frame);

    wxCheckBox* m_utf8;
    wxCheckBox* m_highlight;
    wxCheckBox* m_local_echo;
    wxCheckBox* m_escape_chars;

    wxRadioButton* m_eol_rn;
    wxRadioButton* m_eol_n;
    wxRadioButton* m_eol_r;
    wxRadioButton* m_eol_nr;

    wxCheckBox* m_ssl;

    wxStaticText* m_txt1;
    wxTextCtrl* m_cas;
    wxButton* m_cas_btn;

    wxStaticText* m_txt2;
    wxTextCtrl* m_crt;
    wxButton* m_crt_btn;

    wxStaticText* m_txt3;
    wxTextCtrl* m_key;
    wxButton* m_key_btn;

    wxButton* m_ok_btn;

    void OnSslCheckBox(wxCommandEvent& event);
    void OnOkBtn(wxCommandEvent& event);

private:

    DECLARE_EVENT_TABLE()

    wxFileDialog* m_open;
    void OnCAs_btn(wxCommandEvent& event);
    void OnCrt_btn(wxCommandEvent& event);
    void OnKey_btn(wxCommandEvent& event);
};

class c_cnf_frame : public wxFrame
{
public:
    // ctor(s)
    c_cnf_frame(const wxString& title, const wxPoint& pos, const wxSize& size);

public:
    c_cnf_panel *m_panel;

    // any class wishing to process wxWindows events must use this macro
    DECLARE_EVENT_TABLE()

    void OnClose(wxCloseEvent& event);
};

#endif

