/***************************************************************************
                          match.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 1990 by Jarkko Oikarinen
    copyright            : (C) 2004-2009  Marian VooDooMan Meravy
    email                : ghostvoodooman@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************

XTelnet - a simple telnet client
Copyright (C) 1990 by Jarkko Oikarinen
Copyright (C) 2004-2009  Marian VooDooMan Meravy
  (ghostvoodooman@users.sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides wildcard checks
*/

#ifndef _MATCH_H_
#define _MATCH_H_

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "compat.h"

#include "sock_shared.h"

//#include "params.h"

const char *inet_ntop6 (const u_char *src, char *dst, size_t size);
int  inetpton(int af, const char *src, void *dst);
extern int my_inet_pton(int af, const char *src, void *dst);

const char * inet_ntop (int af, const void *src, char *dst, size_t size);

/*!
    \brief Decompresses "::" from IPv6 numeric address; if other address (or possibly hostname given) returns untouched
    \author VooDooMan
    \version 2
    \date 2005, 2005-10-02
    \param src Source address
    \return Returns uncompressed address (must be delete[]'d)
*/
inline char* ipv6_decompress(const char* src)
{
    size_t res_l=1024;
    if(strlen(src)>res_l+1)
        res_l=strlen(src)+1;
    char *res=new char[res_l+1];
    if(strstr(src,":")==NULL) {
        strcpy(res,src);
        return res;
    }
    if(strstr(src,"::")==NULL) {
        strcpy(res,src);
        return res;
    }
    int expected_num_frags=8;
    if(strstr(src,".")!=NULL)
        expected_num_frags=7;
    int num_frags=0;
    int num_size=0;
    bool got=false;
    size_t i1;
    for(i1=0; i1<strlen(src)+1; i1++) { // +1 here is not typo! we need to process zero-terminator as well!
        if(src[i1]==':' || src[i1]==0) {
            if(/*num_size>4 ||*/ num_frags>expected_num_frags) {
                // error
                strcpy(res,src);
                return res;
            }
            num_size=0;
            if(i1>0 && src[i1-1]==':')
                got=true;
            else
                if((i1>0 && src[i1-1]!=':') && i1>0)
                    num_frags++;
        } else {
            num_size++;
        }
    }
    if(!got) {
        strcpy(res,src);
        return res;
    }
    if(num_frags>8) {
        strcpy(res,src);
        return res;
    }
    bool need_colon=false;
    size_t i2=0;
    for(i1=0; i1<strlen(src); i1++) {
        if(i2>res_l-8) {
            strcpy(res,src);
            return res;
        }
        if(src[i1]==':' && i1>0 && src[i1-1]==':') {
            for(int i3=0; i3<expected_num_frags-num_frags; i3++) {
                if(i2>res_l-8) {
                    strcpy(res,src);
                    return res;
                }
                if(i2>0 && res[i2-1]!=':')
                    res[i2++]=':';
                res[i2++]='0';
                need_colon=true;
            }
        } else {
            if(!(i1==0 && src[i1]==':')) {
                if(need_colon)
                    res[i2++]=':';
                if(!need_colon)
                    res[i2++]=src[i1];
                if(need_colon && src[i1]!=':')
                    res[i2++]=src[i1];
            }
            need_colon=false;
        }
    }
    res[i2++]=0;
    return res;
}

//#ifdef INET6

// BEGIN by VooDooMan
#if defined(IN6_IS_ADDR_V4MAPPED)
#   undef IN6_IS_ADDR_V4MAPPED
#endif
// END   by VooDooMan

# define MYDUMMY_SIZE 128
# define bcopy(a,b,c) memmove((b),(a),(c))
#define IN6_IS_ADDR_V4MAPPED(a)         \
    ((((u_int32_t*)(a))[0] == 0) \
    && (((u_int32_t*)(a))[1] == 0) \
    && (((u_int32_t*)(a))[2] == htonl(0xffff)))

#define u_int32_t uint32
#define u_char unsigned char
#define IN6ADDRSZ       16
#define INADDRSZ        4
#define INT16SZ         2

// do NOT touch this define! It doesn't mean if to use IPv6, but rather if to use IPv6 matching!
#define INET6

#define Reg register
#define u_char unsigned char

// BEGIN ADD    - VooDooMan 2004/10/18 - changed
#define u_long unsigned long

inline char* index(char* str, char sub)
{
    char tmp[2];
    tmp[0]=sub;
    tmp[1]=0;
    return strstr(str,tmp);
}
// END   ADD    - VooDooMan 2004/10/18 - changed

#define  MAX_ITERATIONS  512
/*
**  Compare if a given string (name) matches the given
**  mask (which can contain wild cards: '*' - match any
**  number of chars, '?' - match any single character.
**
**  return  0, if match
**    1, if no match
*/

/*

** match()
** Iterative matching function, rather than recursive.
** Written by Douglas A Lewis (dalewis@acsu.buffalo.edu)
*/

inline int  match(char *mask, char *name)
{
  Reg  u_char  *m = (u_char *)mask, *n = (u_char *)name;
  char  *ma = mask, *na = name;
  int  wild = 0, q = 0, calls = 0;

  if (!*mask)
    return 1;

  if (mask[0]=='*' && mask[1]=='\0')
    return 0;

  while (1)
      {
#ifdef  MAX_ITERATIONS
          if (calls++ > MAX_ITERATIONS)
      break;               
#endif

    if (*m == '*')
       {
      while (*m == '*')
        m++;
      wild = 1;
      ma = (char *)m;
      na = (char *)n;
        }

    if (!*m)
        {
        if (!*n)
        return 0;
        for (m--; (m > (u_char *)mask) && (*m == '?'); m--)
        ;
      if ((m > (u_char *)mask) && (*m == '*') &&
          (m[-1] != '\\'))
        return 0;
      if (!wild) 
        return 1;
      m = (u_char *)ma;
      n = (u_char *)++na;
        }
    else if (!*n)
      return 1;
    if ((*m == '\\') &&
      ((m[1] == '*') || (m[1] == '?') || (m[1] == '#')))
        {
      m++;
      q = 1;
        }
    else
      q = 0;

    if ((tolower(*m) == tolower(*n))
      || (*m == '?' && !q)
      || (*m == '#' && !q && isdigit(*n)))
        {
      if (*m)
        m++;
      if (*n)
        n++;
        }
    else
        {
      if (!wild)
        return 1;
      m = (u_char *)ma;
      n = (u_char *)++na;
        }
      }

  return 1;
}

#define  strncpyzt(x, y, N) do{(void)strncpy(x,y,N);x[N-1]='\0';}while(0)

/*
 * Match address by #IP bitmask (10.11.12.128/27)
 * Now should work for IPv6 too.
 * returns -1 on error, 0 on match, 1 when NO match.
 */
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
//int    match_ipmask(char *mask, aClient *cptr, int maskwithusername)
inline int    match_ipmask(char *mask, char* username, in_addr6_ u_addr, int maskwithusername)
// END   CHANGE - VooDooMan 2004/10/18 - changed
{
  int  m;
  char  *p;
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
//  struct  IN_ADDR addr;
  struct  in_addr6_ addr;
// END   CHANGE - VooDooMan 2004/10/18 - changed
  char  dummy[128];
  char  *omask;
  u_long  lmask;
#ifdef  INET6
  int  j;
#endif

  omask = mask;
  strncpyzt(dummy, mask, sizeof(dummy));
  mask = dummy;
  if (maskwithusername && (p = index(mask, '@')))
  {
    *p = '\0';
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
//    if (match(mask, cptr->username))
    if (match(mask, username))
// END   CHANGE - VooDooMan 2004/10/18 - changed
      return 1;
    mask = p + 1;
  }
  if (!(p = index(mask, '/')))
    goto badmask;
  *p = '\0';
  
  if (sscanf(p + 1, "%d", &m) != 1)
  {
    goto badmask;
  }
  if (!m)
    return 0;       /* x.x.x.x/0 always matches */
#ifndef  INET6
  if (m < 0 || m > 32)
    goto badmask;
  lmask = htonl((u_long)0xffffffffL << (32 - m));
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
//  addr.s_addr = inetaddr(mask);
//  return ((addr.s_addr ^ cptr->ip.s_addr) & lmask) ? 1 : 0;
  addr.s_addr=inet_addr(mask);
  return ((addr.s_addr ^ u_addr) & lmask) ? 1 : 0;
// END   CHANGE - VooDooMan 2004/10/18 - changed
#else
  if (m < 0 || m > 128)
    goto badmask;
  if (my_inet_pton(AF_INET6, mask, (void *)addr.bytes) != 1)
  {
    return -1;
  }

  /* Make sure that the ipv4 notation still works. */
  if (IN6_IS_ADDR_V4MAPPED(&addr) && m < 96)
  {
    m += 96;
  }

  j = m & 0x1F;  /* number not mutliple of 32 bits */
  m >>= 5;  /* number of 32 bits */

  if (m && memcmp((void *)(addr.bytes), 
      (void *)(/*cptr->ip.s6_addr*/&(u_addr.bytes)), m << 2))
    return 1;

  if (j)
  {
    lmask = htonl((u_long)0xffffffffL << (32 - j));
    if ((((u_int32_t *)(addr.bytes))[m] ^
        ((u_int32_t *)(/*cptr->ip.s6_addr*/&(u_addr.bytes)))[m]) & lmask)
      return 1;
  }

  return 0;
#endif
badmask:
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
/*  if (maskwithusername)
  sendto_flag(SCH_ERROR, "Ignoring bad mask: %s", omask);*/
// END   CHANGE - VooDooMan 2004/10/18 - changed
  return -1;
}

#endif

