/***************************************************************************
                          sock.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2009  Marian VooDooMan Meravy
    email                : ghostvoodooman@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2009  Marian VooDooMan Meravy
  (ghostvoodooman@users.sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides low-level IP socket communication
*/

//---------------------------------------------------------------------------

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#if defined(_WIN32)
#       define WINVER   0x0500
#       define _WINSOCKAPI_
#   include <windows.h>
#else
#    include <sys/types.h>
#    include <unistd.h>
#endif
#include <string.h>

#ifdef HAVE_SYS_TIME_H
#   include <sys/time.h>
#endif

//#include "utils.h"      // because we want to enable mem leak detection

#define UNUSEDARG(x)    /* */

#include "params.h"

#if defined(_WIN32)
#   include <winsock2.h>
#endif

#if defined(_WIN32)
#   include <Ws2tcpip.h>
#endif

#include <stdlib.h>

#pragma hdrstop

#include "sock.h"
#include "match.h"
#include "log.h"
//#include "md5.h"            // for byteReverse()

#if defined(_WIN32)
#   define ERRNO WSAGetLastError()
#else
#   define ERRNO errno
#endif

//#include "precomp.h"

//---------------------------------------------------------------------------
#ifndef _MSC_VER
#pragma package(smart_init)
#endif

using namespace std;

#if defined(_WIN32)
# ifndef WINSOCK_VERSION
#   define WINSOCK_VERSION 0x0002
# endif
#endif

#if defined(_WIN32)
WSADATA WSAData;                                    //!< Stores "WSAStartup()"'s data
#   define socklen_t int
#endif

short sock_tcp_proto_num;                           //!< Number of TPC protocol

vector<SOCKET> all_sockets;                         //!< List of all sockets

size_t sock_lo_get_input_amount(s_socket& socket, char* buff, size_t len);
size_t sock_lo_recv(s_socket& socket, char* buff, size_t len);
size_t sock_lo_send(s_socket& socket, char* buff, size_t len);

/*!
    \brief Stores data to send/receive later
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_data {
public:
    bool closed;                        //!< Idicates that connection has been closed (if this is true, data && len are undefined)
    size_t len;                         //!< Length of data
    char* data;                         //!< Data

    c_data()
    {
        closed=false;
        len=0;
        data=NULL;
    }
};

/*!
    \brief Stores data to send later
    \author VooDooMan
    \version 2
    \date 2004, 2005
*/
class c_send_later {
public:
    s_socket socket;                        //!< Socket handle
    list<c_data> data;                      //!< Array of data packets

    time_t last_send;                       //!< Timestamp of last send operation

    c_send_later()
    {
        socket.clear();
        data.clear();
        last_send=time(NULL);
    }
};

/*!
    \brief Stores data to receive later
    \author VooDooMan
    \version 2
    \date 2004, 2005
*/
class c_receive_later {
public:
    bool has_error;                         //!< Set to true, if we have data, but there's an error on the socket, or socket has been closed (by remote host)
    int last_error;                         //!< Stores last error code
    s_socket socket;                        //!< Socket handle
    bool using_window;                      //!< Set to true if this socket is using message window callback to receive data
    list<c_data> data;                      //!< Array of data packets

    time_t last_recv;                       //!< Timestamp of last receive operation

    c_receive_later()
    {
        has_error=false;
        socket.clear();
        using_window=false;
        data.clear();
        last_error=0;
        last_recv=time(NULL);
    }
};

vector<c_send_later> send_later;            //!< Packets to send later
vector<c_receive_later> receive_later;      //!< Packets to receive later

#if defined(_WIN32)
#   define WM_TCPASYNCSELECT (WM_USER+123)     //!< Our own window message for receiving data (Windows(TM) only)

/*!
    \brief Callback windows procedure for receiving data (Windows(TM) only)
    \author VooDooMan
    \version 2
    \date 2005
    \param hwnd Handle of window
    \param uMsg Message
    \param wParam Parameter #1
    \param lParam Parameter #2
    \return Should return zero if message has been processed
*/
LRESULT CALLBACK WndProc(/*HWND*/void* UNUSEDARG(hwnd), /*UINT*/int uMsg, /*WPARAM*/int wParam, /*LPARAM*/int lParam)
{
    if(uMsg==WM_NCCREATE)
        return true;

    char* Buf=NULL;

    try {
        if(uMsg!=WM_TCPASYNCSELECT)
            return 0;
        int ErrorCode=WSAGETSELECTERROR(lParam);
        if(ErrorCode>WSABASEERR)
            return 0;

        int SelectEvent=WSAGETSELECTEVENT(lParam);
        switch (SelectEvent) {
            case FD_READ:
                {
                    unsigned int max_len=1024*64;
                    Buf=new char[max_len];
                    for(;;) {
                        s_socket Sock;
                        Sock.clear();
                        Sock.handle=wParam;
                        int Size=recvfrom(Sock.handle,Buf,max_len,0,NULL,NULL);
                        if(Size==SOCKET_ERROR) {
                            delete[] Buf;
                            Buf=NULL;
                            vector<c_receive_later>::iterator i1;
                            for(i1=receive_later.begin(); i1!=receive_later.end(); i1++)
                                if((*i1).socket.handle==Sock.handle) {
                                    int la=WSAGetLastError();
                                    if(la!=WSAEWOULDBLOCK && !(*i1).has_error) {
                                        (*i1).has_error=true;
                                        (*i1).last_error=la;
                                    }
                                }
                            return 0;
                        }
                        if(Size>0)
                            Sock.last_io=time(NULL);
                        vector<c_receive_later>::iterator i1;
                        for(i1=receive_later.begin(); i1!=receive_later.end(); i1++) {
                            if((*i1).socket.handle==Sock.handle) {
                                if((*i1).socket.last_io<Sock.last_io)
                                    (*i1).socket.last_io=Sock.last_io;
                                else
                                    Sock.last_io=(*i1).socket.last_io;

                                if(Size>0)
                                    (*i1).last_recv=time(NULL);

                                c_data d;
                                char* b=new char[Size];
                                memcpy(b,Buf,Size);
                                d.closed=false;
                                d.data=b;
                                d.len=Size;
                                (*i1).data.push_back(d);
                                break;
                            }
                        }
                        if(Size<=0)
                            break;
                    }
                    delete[] Buf;
                    Buf=NULL;
                }
                break;
            case FD_CLOSE:
                {
                    s_socket Sock;
                    Sock.clear();
                    Sock.handle=wParam;
                    vector<c_receive_later>::iterator i1;
                    for(i1=receive_later.begin(); i1!=receive_later.end(); i1++) {
                        if((*i1).socket.handle==Sock.handle) {
                            c_data d;
                            d.closed=true;
                            (*i1).socket.closed=true;
                            d.data=NULL;
                            d.len=0;
                            (*i1).data.push_back(d);
                            break;
                        }
                    }
                }
                break;
        }
    } catch (...) {}
    if(Buf) {
        delete[] Buf;
        Buf=NULL;
    }
    return 0;
}

/*!
    \brief Allocates window handle (Windows(TM) only)
    \author VooDooMan
    \version 1
    \date 2004
    \param m Pointer to window callback message procedure
    \returns Window handle or NULL if something goes wrong
*/
HWND AllocateHWnd(void* m)
{
    WNDCLASSEXA c;
    memset(&c,0,sizeof(c));
    char tmp[1024];
    tmp[0]=0;
    for(int ii=0; ii<5; ii++) {
        int i=rand();
        char tmp2[16];
        sprintf(tmp2,"%x",i);
        strcat(tmp,tmp2);
    }
    c.lpszClassName=tmp;
    c.lpfnWndProc=(WNDPROC)m;
    c.cbSize=sizeof(c);
    RegisterClassExA(&c);
    HWND h=CreateWindowExA(0,tmp,"",0,0,0,0,0,HWND_MESSAGE,NULL,NULL,NULL);
    SetWindowLongA(h, GWL_WNDPROC,(LONG)m);
    return h;
}
#endif

/*!
    \brief Writes data to connected socket
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle of connected socket
    \param data Buffer to write
    \param size Size of buffer
    \param error_code Returns OS's socket error
    \return Returns amount of sent data in bytes
*/
size_t sock_send(s_socket& socket, const char* data, size_t size, int& error_code)
{
    {
        {
            {
            again:
                bool got=false;
                vector<c_send_later>::iterator i1;
                for(i1=send_later.begin(); i1!=send_later.end(); i1++) {
                    if((*i1).socket.handle==socket.handle) {
                        if((*i1).socket.last_io<socket.last_io)
                            (*i1).socket.last_io=socket.last_io;
                        else
                            socket.last_io=(*i1).socket.last_io;
                        (*i1).last_send=time(NULL);
                        got=true;
                        break;
                    }
                }
                if(!got) {
                    c_send_later s;
                    s.socket=socket;
                    s.last_send=time(NULL);
                    send_later.push_back(s);
                    goto again;
                }
                c_data d;
                d.closed=false;
                d.len=size;
                d.data=new char[size];
                memcpy(d.data,data,size);
                (*i1).data.push_back(d);

                error_code=0;
                return size;
            }
        }
    }
}

/*!
    \brief Reads data from connected socket
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle of connected socket
    \param data Buffer to be filled up to "size" bytes
    \param size Size of buffer (maximum bytes to be read)
    \param error_code Returns OS's socket error
    \param connection_closed Returns true if data was successfuly read, but remote host closed the connection (this is not error and read of data is working correctly)
    \return Returns amount of read data in bytes
*/
size_t sock_read(s_socket& socket, char* data, size_t size, int& error_code, bool& connection_closed)
{
    error_code=0;
    connection_closed=false;
    {
        vector<c_receive_later>::iterator i1;
        for(i1=receive_later.begin(); i1!=receive_later.end(); i1++) {
            if((*i1).socket.handle==socket.handle) {
                if((*i1).socket.last_io<socket.last_io)
                    (*i1).socket.last_io=socket.last_io;
                else
                    socket.last_io=(*i1).socket.last_io;

                if((*i1).data.begin()==(*i1).data.end() && (*i1).last_error) {
                    error_code=(*i1).last_error;
                    return 0;
                }

                list<c_data>::iterator i2;
                connection_closed=false;
again:
                for(i2=(*i1).data.begin(); i2!=(*i1).data.end(); i2++) {
                    (*i1).last_recv=time(NULL);

                    error_code=0;
                    if((*i2).closed) {
                        connection_closed=true;
                        (*i1).socket.closed=true;
                        (*i1).data.pop_front();
                        goto again;
                    }
                    connection_closed=false;

                    size_t l=size;
                    if((*i2).len<=l) {
                        l=(*i2).len;
                        if((*i2).len)
                            memcpy(data,(*i2).data,l);
                        delete[] (*i2).data;
                        (*i1).data.pop_front();
                        error_code=0;
                        return l;
                    }
                    if((*i2).len>l) {
                        memcpy(data,(*i2).data,l);
                        char* buf=new char[(*i2).len-l];
                        memcpy(buf,&((*i2).data[l]),(*i2).len-l);
                        delete[] (*i2).data;
                        (*i2).len-=l;
                        (*i2).data=buf;
                        error_code=0;
                        return l;
                    }
                }
                error_code=0;
                return 0;
            }
        }
    }

    return 0;
}

/*!
    \brief Puts buffer back to queue as if it weren't read
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle of connected socket
    \param data Buffer to be filled up to "size" bytes
    \param size Size of buffer (maximum bytes to be read)
*/
void sock_unread(s_socket& socket, char* data, size_t size)
{
    {
        vector<c_receive_later>::iterator i1;
        for(i1=receive_later.begin(); i1!=receive_later.end(); i1++) {
            if((*i1).socket.handle==socket.handle) {
                if((*i1).socket.last_io<socket.last_io)
                    (*i1).socket.last_io=socket.last_io;
                else
                    socket.last_io=(*i1).socket.last_io;
                c_data d;
                d.closed=false;
                d.data=data;
                d.len=size;
                (*i1).data.push_front(d);
                return;
            }
        }
    }
}

/*!
    \brief Get size of input buffer on socket
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle
    \param closed Returns true if connection was closed
    \return Number of bytes waiting to receive
*/
size_t sock_get_receive_size(s_socket& socket, bool& closed)
{
    size_t result=0;
    closed=false;

    vector<c_receive_later>::iterator i1;
    for(i1=receive_later.begin(); i1!=receive_later.end(); i1++) {
        if((*i1).socket.handle==socket.handle) {
            if((*i1).socket.last_io<socket.last_io)
                (*i1).socket.last_io=socket.last_io;
            else
                socket.last_io=(*i1).socket.last_io;

            list<c_data>::iterator i2;
            for(i2=(*i1).data.begin(); i2!=(*i1).data.end(); i2++) {
                if(!(*i2).closed)
                    result+=(*i2).len;
                else
                    closed=true;
            }
            break;
        }
    }
    return result;
}

/*!
    \brief Sends and receive all pending data (in the cache), should be called in loop
    \author VooDooMan
    \version 2
    \date 2005
*/
void sock_send_cache()
{
    unsigned int _tmp_size_=1024*64;
    char* _tmp_=NULL;

    vector<c_receive_later>::iterator i2;
    for(i2=receive_later.begin(); i2!=receive_later.end(); i2++) {
        if(!(*i2).socket.cmp())
            continue;

        {
            {
                if(!_tmp_)
                    _tmp_=new char[_tmp_size_];
                // do a dummy operation
                size_t res=sock_lo_get_input_amount((*i2).socket,_tmp_,_tmp_size_);

                int error_code=WSAGetLastError();
                if(res!=SOCKET_ERROR)
                    error_code=0;
                if(error_code && error_code!=WSAEWOULDBLOCK) {
                    if(!(*i2).has_error) {
                        (*i2).has_error=true;
                        (*i2).last_error=error_code;
                    }
                }
            }
        }

        if((*i2).using_window)
            continue;

        char tmp[1024];
        socklen_t tmp2=1024;
        getsockopt((*i2).socket.handle,SOL_SOCKET,SO_ERROR,tmp,&tmp2);

        for(;;) {
            {
                if(!_tmp_)
                    _tmp_=new char[_tmp_size_];
                // do a dummy operation
                size_t res=sock_lo_get_input_amount((*i2).socket,_tmp_,_tmp_size_);

                int error_code=WSAGetLastError();
                if(res!=SOCKET_ERROR)
                    error_code=0;
                if(error_code && error_code!=WSAEWOULDBLOCK) {
                    if(!(*i2).has_error) {
                        (*i2).has_error=true;
                        (*i2).last_error=error_code;
                    }
                }
            }

            fd_set a;
            timeval t;

            t.tv_sec=0;
            t.tv_usec=1;

            FD_ZERO(&a);
            FD_SET((*i2).socket.handle,&a);

            int i=select((int)(*i2).socket.handle+1,&a,NULL,NULL,&t);
            if(i==0)
                break;

            unsigned int max_size=1024*64;
            char* buf=new char[max_size];
            size_t size_=sock_lo_recv((*i2).socket,buf,max_size);
            if(size_==SOCKET_ERROR) {
                delete[] buf;
                int error_code=WSAGetLastError();
                if(error_code==WSAEWOULDBLOCK)
                    break;
                if(!(*i2).has_error) {
                    (*i2).has_error=true;
                    (*i2).last_error=error_code;
                    break;
                }
                break;
            }
            if(size_>0) {
                (*i2).socket.last_io=time(NULL);
                (*i2).last_recv=time(NULL);
            }
            if(size_==0) {
                delete[] buf;
                if(!(*i2).has_error) {
                    c_data d;
                    d.closed=true;
                    //d.error=0;
                    (*i2).data.push_back(d);
                    (*i2).has_error=true;
                }
                break;
            }
            if(size_) {
                c_data d;
                d.closed=false;
                d.len=size_;
                d.data=buf;
                (*i2).data.push_back(d);
            }
        }
    }

    vector<c_send_later>::iterator i1;
    for(i1=send_later.begin(); i1!=send_later.end(); i1++) {
    again:
        {
            if(!(*i1).socket.cmp())
                continue;

            list<c_data>::iterator i2;
            try {
                i2=(*i1).data.begin();
            } catch(...) {
                continue;
            }
            for(i2=(*i1).data.begin(); i2!=(*i1).data.end(); ) {
                size_t l=1024;
                if((*i2).len<l)
                    l=(*i2).len;
                size_t Sent=sock_lo_send((*i1).socket,(*i2).data,l);
                if(Sent==SOCKET_ERROR) {
                    int la=WSAGetLastError();
                    if(la==WSAEWOULDBLOCK)
                        break;
                    {
                        vector<c_receive_later>::iterator i2;
                        for(i2=receive_later.begin(); i2!=receive_later.end(); i2++)
                            if((*i2).socket.handle==(*i1).socket.handle) {
                                (*i2).has_error=true;
                                if((*i2).last_error==0)
                                    (*i2).last_error=la;
                                break;
                            }
                    }
                    break;
                }
                if(Sent>0) {
                    (*i1).socket.last_io=time(NULL);
                    (*i1).last_send=time(NULL);
                }
                if(Sent<(*i2).len) {
                    char* buf=new char[(*i2).len-Sent];
                    memcpy(buf,&(*i2).data[Sent],(*i2).len-Sent);
                    delete[] (*i2).data;
                    (*i2).len-=Sent;
                    (*i2).data=buf;
                    if(Sent==0)
                        goto next;
                } else {
                    delete[] (*i2).data;
                    (*i1).data.pop_front();
                }
                goto again;
            }
        }
    next:
        ;
    }
    if(_tmp_)
        delete[] _tmp_;
}

/*!
    \brief Creates cache entries
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle to create cache for
    \param using_window Are we using window message callback to receive? (Windows(TM) only)
*/
void sock_create_later(s_socket& socket, bool using_window)
{
#if !defined(_WIN32)
    using_window=false;
#endif
    c_receive_later r;
    r.has_error=false;
    r.socket=socket;
    r.using_window=using_window;
    r.last_error=0;
    r.last_recv=time(NULL);
    receive_later.push_back(r);

    c_send_later s;
    s.socket=socket;
    s.last_send=time(NULL);
    send_later.push_back(s);
}

/*!
    \brief Sets asynchronous mode for the socket
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle
*/
void sock_async(s_socket& socket)
{
#if defined(_WIN32)
#   ifndef __GNUC__
    HWND WindowHandle=AllocateHWnd(WndProc);
#   else
    HWND WindowHandle=AllocateHWnd((void*)WndProc);
#   endif
    socket.wnd=WindowHandle;
    WSAAsyncSelect(socket.handle,WindowHandle,WM_TCPASYNCSELECT,FD_READ | FD_CLOSE);
#endif
    sock_set_nonblocking(socket,true);
}

/*!
    \brief Flushes cache data (receives and sends)
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle to flush cache for
*/
void sock_flush_later(s_socket& socket)
{
    vector<c_send_later>::iterator i1;
    for(i1=send_later.begin(); i1!=send_later.end(); i1++) {
        if((*i1).socket.handle==socket.handle) {
            if((*i1).socket.last_io<socket.last_io)
                (*i1).socket.last_io=socket.last_io;
            else
                socket.last_io=(*i1).socket.last_io;

            (*i1).last_send=time(NULL);

            list<c_data>::iterator i2;
        again:
            for(i2=(*i1).data.begin(); i2!=(*i1).data.end(); ) {
                if(!(*i2).closed) {
                    size_t l=1024;
                    if((*i2).len<l)
                        l=(*i2).len;
                    size_t Sent=sock_lo_send((*i1).socket,(*i2).data,l);
                    if(Sent==SOCKET_ERROR) {
                        int la=WSAGetLastError();
                        if(la!=WSAEWOULDBLOCK) {
                            vector<c_receive_later>::iterator i3;
                            for(i3=receive_later.begin(); i3!=receive_later.end(); i3++)
                                if((*i3).socket.handle==(*i1).socket.handle) {
                                    (*i3).has_error=true;
                                    if((*i3).last_error==0)
                                        (*i3).last_error=la;
                                    break;
                                }
                            (*i1).data.pop_front();
                            goto again;
                        } else {
                            break;
                        }
                    }
                    if(Sent>0)
                        (*i1).socket.last_io=time(NULL);
                    if(Sent<(*i2).len) {
                        char* buf=new char[(*i2).len-Sent];
                        memcpy(buf,&(*i2).data[Sent],(*i2).len-Sent);
                        delete[] (*i2).data;
                        (*i2).len-=Sent;
                        (*i2).data=buf;
                    } else {
                        delete[] (*i2).data;
                        (*i1).data.pop_front();
                    }
                    goto again;
                }

                (*i1).data.pop_front();
                goto again;
            }
            break;
        }
    }
}

/*!
    \brief Frees cache data (receives and sends), without send / receive
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle to free cache for
*/
void sock_erase_later(s_socket& socket)
{
    vector<c_send_later>::iterator i1;
again_a:
    for(i1=send_later.begin(); i1!=send_later.end(); i1++) {
        if((*i1).socket.handle==socket.handle) {
            list<c_data>::iterator ii;
again_1:
            for(ii=(*i1).data.begin(); ii!=(*i1).data.end(); ) {
                if(!(*ii).closed)
                    delete[] (*ii).data;
                (*i1).data.erase(ii);
                goto again_1;
            }
            send_later.erase(i1);
            goto again_a;
        }
    }
    vector<c_receive_later>::iterator i2;
again_b:
    for(i2=receive_later.begin(); i2!=receive_later.end(); i2++) {
        if((*i2).socket.handle==socket.handle) {
            if((*i2).socket.last_io<socket.last_io)
                (*i2).socket.last_io=socket.last_io;
            else
                socket.last_io=(*i2).socket.last_io;

            list<c_data>::iterator ii;
again_2:
            for(ii=(*i2).data.begin(); ii!=(*i2).data.end(); ) {
                if(!(*ii).closed)
                    delete[] (*ii).data;
                (*i2).data.erase(ii);
                goto again_2;
            }

            receive_later.erase(i2);
            goto again_b;
        }
    }
}

/*!
    \brief Cancels caching
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle
*/
void sock_cancel_caching(s_socket& socket)
{
    sock_flush_later(socket);
    sock_erase_later(socket);
#if defined(_WIN32)
    WSAAsyncSelect(socket.handle,socket.wnd,0,0);
#endif
}

/*!
    \brief Returns last read/receive or send timestamp on socket
    \author VooDooMan
    \version 1
    \date 2005
    \param socket Socket handle
    \param mode Set to 'r' to get timestamp of last receive; set to 's' to get time of last send; set to 'a' (default) (stands for "all") to get most recent timestamp on both send and receive
    \return Timestamp of last send or receive on socket
    \warning If argument "mode" contain neither 'r', nor 's', nor 'a', the returned value is zero
*/
time_t sock_get_last_io(s_socket& socket, char mode)
{
    time_t last_io=0;

    if(mode=='r') {
        vector<c_receive_later>::iterator i2;
        for(i2=receive_later.begin(); i2!=receive_later.end(); i2++)
            if((*i2).socket.handle==socket.handle) {
                last_io=(*i2).last_recv;
                break;
            }
    }

    if(mode=='s') {
        vector<c_send_later>::iterator i2;
        for(i2=send_later.begin(); i2!=send_later.end(); i2++)
            if((*i2).socket.handle==socket.handle) {
                last_io=(*i2).last_send;
                break;
            }
    }

    if(mode=='a') {
        last_io=socket.last_io;

        vector<c_receive_later>::iterator i2;
        for(i2=receive_later.begin(); i2!=receive_later.end(); i2++)
            if((*i2).socket.handle==socket.handle) {
                if(last_io<(*i2).socket.last_io)
                    last_io=(*i2).socket.last_io;
                if(last_io<(*i2).last_recv)
                    last_io=(*i2).last_recv;
                break;
            }

        vector<c_send_later>::iterator i1;
        for(i1=send_later.begin(); i1!=send_later.end(); i1++)
            if((*i1).socket.handle==socket.handle) {
                if(last_io<(*i1).socket.last_io)
                    last_io=(*i1).socket.last_io;
                if(last_io<(*i1).last_send)
                    last_io=(*i1).last_send;
                break;
            }
    }

    return last_io;
}

/*!
    \brief Sets last read/receive or send timestamp on socket
    \author VooDooMan
    \version 1
    \date 2006-07-16
    \param socket Socket handle
    \param last_io Set timestamp of last I/O
    \param mode Set to 'r' to set timestamp of last receive; set to 's' to set time of last send
    \return Timestamp of last send or receive on socket
*/
void sock_set_last_io(s_socket& socket, time_t last_io, char mode)
{
    socket.last_io=last_io;
    if(mode=='r') {
        vector<c_receive_later>::iterator i2;
        for(i2=receive_later.begin(); i2!=receive_later.end(); i2++)
            if((*i2).socket.handle==socket.handle) {
                (*i2).last_recv=last_io;
                break;
            }
    }

    if(mode=='s') {
        vector<c_send_later>::iterator i2;
        for(i2=send_later.begin(); i2!=send_later.end(); i2++)
            if((*i2).socket.handle==socket.handle) {
                (*i2).last_send=last_io;
                break;
            }
    }
}

/*!
    \brief Discovers type of address
    \author VooDooMan
    \version 1
    \date 2005
    \param host IP of host name
    \retval 0 Given host is DNS name of host
    \retval 4 Given host is numeric IPv4 address
    \retval 6 Given host is numeric IPv6 address
*/
int sock_get_address_type(string host)
{
    if(!host.compare("0.0.0.0")) // special case
        return 4;
    uint32 x=sock_ipv4_from_string(host.c_str());
    if(x)
        return 4;
    if(host.find(":",0)!=string::npos) {
        in_addr6_ addr6;
        if(my_inet_pton(AF_INET6,host.c_str(),&addr6)==1)
            return 6;
    }
    return 0;
}

uint32 sock_ipv4_from_string(const char* str)
{
    uint32 ip=0;
    if(my_inet_pton(AF_INET,str,&ip)!=1)
        ip=0;
    return ip;
}

/*!
    \brief Retrieves statistics about sock I/O cache
    \author VooDooMan
    \version 1
    \date 2006-01-21
    \param socket Socket
    \param to_receive Returns number of bytes waiting in cache to receive them
    \param to_send Returns number of bytes waiting in cache to send them
*/
void sock_cache_stats(s_socket& socket, size_t& to_receive, size_t& to_send)
{
    to_receive=to_send=0;

    {
        vector<c_receive_later>::iterator i2;
        for(i2=receive_later.begin(); i2!=receive_later.end(); i2++) {
            if((*i2).socket.handle==socket.handle) {
                list<c_data>::iterator i1;
                for(i1=(*i2).data.begin(); i1!=(*i2).data.end(); i1++)
                    if(!(*i1).closed)
                        to_receive+=(*i1).len;
                break;
            }
        }
    }

    {
        vector<c_send_later>::iterator i2;
        for(i2=send_later.begin(); i2!=send_later.end(); i2++) {
            if((*i2).socket.handle==socket.handle) {
                list<c_data>::iterator i1;
                for(i1=(*i2).data.begin(); i1!=(*i2).data.end(); i1++)
                    if(!(*i1).closed)
                        to_send+=(*i1).len;
                break;
            }
        }
    }
}

/*!
    \brief Returns OS's error string from socket I/O error
    \author VooDooMan
    \version 2
    \date 2004, 2005-05-19
    \param code Socket I/O error
    \param type Type of error code
    \return OS's error string (must be delete[]'d)
    \warning Result must be delete[]'d
*/
char* sock_error(int code, c_sock_log::e_type
#if !defined(_WIN32)
type
#endif
)
{
    char* result=new char[1024+1];
    result[0]='\0';
    switch((unsigned long)code) {
        case MY_SELECT_ERROR:
            strcpy(result,"(Unknown error, based on select() fuction)");
            return result;
        case MY_NO_IPV6:
            strcpy(result,"(Not compiled with IPv6 support)");
            return result;
        case MY_SSL_ERROR:
            strcpy(result,"(SSL error)");
            return result;
        case MY_UNKNOWN_ERROR:
            strcpy(result,"(Unknown error MY_UNKNOWN_ERROR)");
            return result;
        case MY_DNS_ERROR:
            strcpy(result,"(DNS lookup failed)");
            return result;
    }
#if defined(_WIN32)
    switch(code) {
        case WSAEACCES:
            strcpy(result,"Permission denied"); break;
        case WSAEADDRINUSE:
            strcpy(result,"Address already in use"); break;
        case WSAEADDRNOTAVAIL:
            strcpy(result,"Cannot assign requested address"); break;
        case WSAEAFNOSUPPORT:
            strcpy(result,"Address family not supported by protocol family"); break;
        case WSAEALREADY:
            strcpy(result,"Operation already in progress"); break;
        case WSAECONNABORTED:
            strcpy(result,"Software caused connection abort"); break;
        case WSAECONNREFUSED:
            strcpy(result,"Connection refused"); break;
        case WSAECONNRESET:
            strcpy(result,"Connection reset by peer"); break;
        case WSAEDESTADDRREQ:
            strcpy(result,"Destination address required"); break;
        case WSAEFAULT:
            strcpy(result,"Bad address"); break;
        case WSAEHOSTDOWN:
            strcpy(result,"Host is down"); break;
        case WSAEHOSTUNREACH:
            strcpy(result,"No route to host"); break;
        case WSAEINPROGRESS:
            strcpy(result,"Operation now in progress"); break;
        case WSAEINTR:
            strcpy(result,"Interrupted function call"); break;
        case WSAEINVAL:
            strcpy(result,"Invalid argument"); break;
        case WSAEISCONN:
            strcpy(result,"Socket is already connected"); break;
        case WSAEMFILE:
            strcpy(result,"Too many open files"); break;
        case WSAEMSGSIZE:
            strcpy(result,"Message too long"); break;
        case WSAENETDOWN:
            strcpy(result,"Network is down"); break;
        case WSAENETRESET:
            strcpy(result,"Network dropped connection on reset"); break;
        case WSAENETUNREACH:
            strcpy(result,"Network is unreachable"); break;
        case WSAENOBUFS:
            strcpy(result,"No buffer space available"); break;
        case WSAENOPROTOOPT:
            strcpy(result,"Bad protocol option"); break;
        case WSAENOTCONN:
            strcpy(result,"Socket is not connected"); break;
        case WSAENOTSOCK:
            strcpy(result,"Socket operation on non-socket"); break;
        case WSAEOPNOTSUPP:
            strcpy(result,"Operation not supported"); break;
        case WSAEPFNOSUPPORT:
            strcpy(result,"Protocol family not supported"); break;
        case WSAEPROCLIM:
            strcpy(result,"Too many processes"); break;
        case WSAEPROTONOSUPPORT:
            strcpy(result,"Protocol not supported"); break;
        case WSAEPROTOTYPE:
            strcpy(result,"Protocol wrong type for socket"); break;
        case WSAESHUTDOWN:
            strcpy(result,"Cannot send after socket shutdown"); break;
        case WSAESOCKTNOSUPPORT:
            strcpy(result,"Socket type not supported"); break;
        case WSAETIMEDOUT:
            strcpy(result,"Connection timed out"); break;
        case WSATYPE_NOT_FOUND:
            strcpy(result,"Class type not found"); break;
        case WSAEWOULDBLOCK:
            strcpy(result,"Resource temporarily unavailable"); break;
        case WSAHOST_NOT_FOUND:
            strcpy(result,"Host not found"); break;
        case WSANOTINITIALISED:
            strcpy(result,"Successful WSAStartup not yet performed"); break;
        case WSANO_DATA:
            strcpy(result,"Valid name, no data record of requested type"); break;
        case WSANO_RECOVERY:
            strcpy(result,"This is a non-recoverable error"); break;
        case WSASYSCALLFAILURE:
            strcpy(result,"System call failure"); break;
        case WSASYSNOTREADY:
            strcpy(result,"Network subsystem is unavailable"); break;
        case WSATRY_AGAIN:
            strcpy(result,"Non-authoritative host not found"); break;
        case WSAVERNOTSUPPORTED:
            strcpy(result,"WINSOCKDLL version out of range"); break;
        case WSAEDISCON:
            strcpy(result,"Graceful shutdown in progress"); break;

        default:
            strcpy(result,"(Unknown error)"); break;
    }
#else
    switch(type) {
        case c_sock_log::TYPE_ERRNO:
            {
#ifndef HAVE_STRERROR_R
                strcpy(result,strerror(code));
#else
#   ifdef STRERROR_R_CHAR_P
                char buff[1024];
                strcpy(result,strerror_r(code,buff,1024));
#   else
                strerror_r(code,result,1024);
#   endif
                errno=0;
#endif
            }
            break;
        case c_sock_log::TYPE_GAI:
            strcpy(result,gai_strerror(code));
            break;
        case c_sock_log::TYPE_GETHOSTBY:
            strcpy(result,hstrerror(code));
            break;
        default:
            strcpy(result,"(Unknown error)");
            break;
    }

#endif
    return(result);
}

/*!
    \brief Finds out if it is a temporary error
    \author VooDooMan
    \version 1
    \date 2006-05-19
    \param code Socket I/O error
    \param type Type of error code
    \retval true Temporary error
    \retval false Not temporary error
*/
bool sock_temporary_error(int code, c_sock_log::e_type
#if !defined(_WIN32)
type
#endif
)
{
#if defined(_WIN32)
    switch(code) {
        case WSAEACCES:
        case WSAEADDRINUSE:
        case WSAEADDRNOTAVAIL:
        case WSAEAFNOSUPPORT:
        case WSAEALREADY:
        case WSAECONNABORTED:
        case WSAECONNREFUSED:
        case WSAECONNRESET:
        case WSAEDESTADDRREQ:
        case WSAEFAULT:
        case WSAEHOSTDOWN:
        case WSAEHOSTUNREACH:
        case WSAEINPROGRESS:
        case WSAEINTR:
        case WSAEINVAL:
        case WSAEISCONN:
            return false;

        case WSAEMFILE:
            return true;

        case WSAEMSGSIZE:
            return false;

        case WSAENETDOWN:
            return true;

        case WSAENETRESET:
        case WSAENETUNREACH:
            return false;

        case WSAENOBUFS:
            return true;

        case WSAENOPROTOOPT:
        case WSAENOTCONN:
        case WSAENOTSOCK:
        case WSAEOPNOTSUPP:
        case WSAEPFNOSUPPORT:
            return false;

        case WSAEPROCLIM:
            return true;

        case WSAEPROTONOSUPPORT:
        case WSAEPROTOTYPE:
        case WSAESHUTDOWN:
        case WSAESOCKTNOSUPPORT:
        case WSAETIMEDOUT:
        case WSATYPE_NOT_FOUND:
            return false;

        case WSAEWOULDBLOCK:
            return true;

        case WSAHOST_NOT_FOUND:
            return false;

        case WSANOTINITIALISED:
            return true;

        case WSANO_DATA:
        case WSANO_RECOVERY:
        case WSASYSCALLFAILURE:
        case WSASYSNOTREADY:
            return false;

        case WSATRY_AGAIN:
            return true;

        case WSAVERNOTSUPPORTED:
        case WSAEDISCON:
            return false;

        case MY_SELECT_ERROR:
            return true;
        case MY_NO_IPV6:
            return false;
        case MY_SSL_ERROR:
            return true;
        case MY_UNKNOWN_ERROR:
            return false;
        default:
            return true; // hm?
    }
#else
    if(MY_NO_IPV6==code)
        return false;
    else if(MY_UNKNOWN_ERROR==code)
        return false;

    switch(type) {
        case c_sock_log::TYPE_ERRNO:
            {
                if(EWOULDBLOCK==code) // to avoid confilict of multiple case's in case EWOULDBLOCK==EAGAIN==TRY_AGAIN
                    return true;

                if(TRY_AGAIN==code) // to avoid confilict of multiple case's in case EWOULDBLOCK==EAGAIN==TRY_AGAIN
                    return true;

                switch(code) {
                    case EAGAIN:
                    case EBUSY:
                    case EMFILE:
                    case EMLINK:
                    case ENETDOWN:
                    case ENFILE:
#if defined(ENODATA)
                    case ENODATA:
#endif
                    case ENOMEM:
                    case ENOMSG:
                    case ENOSPC:
                        return true;
                }
            }
            break;
        case c_sock_log::TYPE_GETHOSTBY:
            {
                if(NO_DATA==code) // to avoid confilict of multiple case's in case NO_ADDRESS==NO_DATA
                    return false;

                switch(code) {
                    case HOST_NOT_FOUND:
                    case NO_ADDRESS:
                    case NO_RECOVERY:
                        return false;
                    case TRY_AGAIN:
                        return true;
                    default:
                        return true;
                }
            }
            break;
        case c_sock_log::TYPE_GAI:
            return true;
        case c_sock_log::TYPE_MY:
            {
                switch(code) {
                    case MY_SELECT_ERROR:
                        return true;
                    case MY_NO_IPV6:
                        return false;
                    case MY_SSL_ERROR:
                        return true;
                    case MY_UNKNOWN_ERROR:
                        return false;
                    default:
                        return true;
                }
            }
            break;
        default:
            return true;
    }
    return false;
#endif
}

/*!
    \brief Initializes WinSock
    \author VooDooMan
    \version 1
    \date 2004
    \return OS's socket error
    \warning Only for Windows(TM)
*/
int sock_startup()
{
    errno=0;
    int la=0;
#if defined(_WIN32)
    int ec=WSAStartup(WINSOCK_VERSION,&WSAData);
    la=WSAGetLastError();
    if(ec==0)
        la=0;
    errno=0;
#endif

    protoent* pe=getprotobyname("tcp");
    if(pe)
        sock_tcp_proto_num=pe->p_proto;
    else {
#if defined(_WIN32)
        la=WSAGetLastError();
#else
        la=errno;
#endif
    }
    errno=0;

    return la;
}

/*!
    \brief Shuts down WinSock
    \author VooDooMan
    \version 1
    \date 2004
    \return OS's socket error
    \warning Only for Windows(TM)
*/
int sock_shutdown()
{
    errno=0;
#if defined(_WIN32)
    int ec=WSACleanup();
    int la=WSAGetLastError();
    if(ec==0)
        la=0;
    errno=0;
    return la;
#else
    return 0;
#endif
}

/*!
    \brief Sets TCP keep-alive messages to be sent on socket
    \author VooDooMan
    \version 1
    \date 2006-05-19
    \param socket Socket
*/
void sock_set_tcp_keepalive(s_socket& socket)
{
    int set=1;
    socklen_t size=sizeof(set);
    setsockopt(socket.handle,sock_tcp_proto_num,SO_KEEPALIVE,(char*)&set,size);
}

/*!
    \brief Connects to remote host
    \author VooDooMan
    \version 1
    \date 2004
    \param localIP Local IPv4 address as text to bind, if we are using IPv4 as destination, otherwise ignored
    \param remote Remote IPv4/IPv6 address as text, or IPv4/IPv6 host name
    \param remote_port Remote port
    \param error_code Returns OS's socket error
    \param async If true, it calls WSAAsyncSelect and creates cache buffers
    \param comment Comment for socket
    \return If error_code == zero, returns socket handle of connected socket, else undefined
*/
s_socket sock_connect(const char* localIP, const char* remote, unsigned short remote_port, int& error_code, bool async, const char* comment)
{
    errno=0;
    s_socket Socket;

#ifdef IPv6

    s_socket nul;
    nul.clear();
    nul.comment=comment;

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    char port_name[64];
    sprintf(port_name,"%u",(unsigned int)remote_port);

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=PF_UNSPEC;
    Hints.ai_socktype=SOCK_STREAM;
    errno=0;
    if(getaddrinfo(remote,port_name,&Hints,&AddrInfo)) {
        int ec=WSAGetLastError();
        string log="Cannot resolve address and port, while calling getaddrinfo() in file " __FILE__ " in function ";
        log+=__func__;
        char* errstr=NULL;
        if(ec)
            log_socket(ec,errstr=sock_error(ec),log.c_str(),&nul);
        delete[] errstr;
        error_code=ec;
        s_socket nul;
        errno=0;
        return nul;
    }
    errno=0;

    //
    // Try each address getaddrinfo returned, until we find one to which
    // we can sucessfully connect.
    //
    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {

        // Open a socket with the correct address family for this address.
        errno=0;
        Socket.handle=socket(AI->ai_family,AI->ai_socktype,AI->ai_protocol);
        if(Socket.handle==INVALID_SOCKET) {
            string log="Error opening socket, while calling socket() in file " __FILE__ " in function ";
            log+=__func__;
            int ec=WSAGetLastError();
            char* errstr=NULL;
            if(ec)
                log_socket(ec,errstr=sock_error(ec),log.c_str(),&nul);
            delete[] errstr;
            errno=0;
            continue;
        }
        errno=0;

        if(strstr(localIP,":")!=NULL) {
            sockaddr_in6 LocalName;
            memset(&LocalName,0,sizeof(LocalName));
            LocalName.sin6_family=AF_INET6;
            LocalName.sin6_port=0;
            errno=0;
            my_inet_pton(AF_INET6,localIP,&LocalName.sin6_addr);
            errno=0;
#if SOCKET_BIND_WILDCARD==1
            memset(&LocalName.sin6_addr,0,sizeof(LocalName.sin6_addr));
            memset(&LocalName.sin6_port,0,sizeof(LocalName.sin6_port));
#endif
            int ErrorCode=bind(Socket.handle,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
            if(ErrorCode!=0) {
                int la=error_code=WSAGetLastError();
                if(la!=0) {
                    s_socket nul;
                    nul.clear();
                    nul.comment=comment;
                    errno=0;
                    return nul;
                }
                errno=0;
            }
            errno=0;
        } else {
            sockaddr_in LocalName;
            memset(&LocalName,0,sizeof(LocalName));
            LocalName.sin_family=AF_INET;
            LocalName.sin_port=0;
            errno=0;
#if defined(_WIN32)
            my_inet_pton(AF_INET,localIP,&LocalName.sin_addr.S_un.S_addr);
#else
            my_inet_pton(AF_INET,localIP,&LocalName.sin_addr.s_addr);
#endif
            errno=0;

#if SOCKET_BIND_WILDCARD==1
            memset(&LocalName.sin_addr,0,sizeof(LocalName.sin_addr));
            memset(&LocalName.sin_port,0,sizeof(LocalName.sin_port));
#endif

            int ErrorCode=bind(Socket.handle,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
            if(ErrorCode!=0) {
                int la=error_code=WSAGetLastError();
                if(la!=0) {
                    s_socket nul;
                    nul.clear();
                    nul.comment=comment;
                    errno=0;
                    return nul;
                }
                errno=0;
            }
            errno=0;
        }

        errno=0;
        if(async)
            sock_create_later(Socket,true);
        errno=0;

        errno=0;
        if(connect(Socket.handle,AI->ai_addr,AI->ai_addrlen)==SOCKET_ERROR) {
            sock_erase_later(Socket);
            closesocket(Socket.handle);
            string log="Error connecting socket, while calling connect() in file " __FILE__ " in function ";
            log+=__func__;
            int ec=WSAGetLastError();
            char* errstr=NULL;
            if(ec)
                log_socket(ec,errstr=sock_error(ec),log.c_str(),&nul);
            delete[] errstr;
            errno=0;
            continue;
        }

        errno=0;
        break;
    }

    errno=0;
    freeaddrinfo(AddrInfo);
    errno=0;

    if(!AI) {
        errno=0;
        sock_erase_later(Socket);
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        errno=0;
        return nul;
    }

#else

    sockaddr_in SockAddrIn;
    SockAddrIn.sin_family=AF_INET;
    SockAddrIn.sin_addr=sock_resolve(remote);
    SockAddrIn.sin_port=htons(remote_port);

    uint32 a=0;
    memcpy(&a,&(SockAddrIn.sin_addr),sizeof(a));
    if(!a) {
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        errno=0;
        error_code=MY_DNS_ERROR;
        return nul;
    }

    errno=0;

    Socket.handle=socket(PF_INET,SOCK_STREAM,sock_tcp_proto_num);
    if(Socket.handle==SOCKET_ERROR) {
        int la=error_code=WSAGetLastError();
        if(la!=0) {
            s_socket nul;
            nul.clear();
            nul.comment=comment;
            errno=0;
            return nul;
        }
        errno=0;
    }
    errno=0;

    sockaddr_in LocalName;
    memset(&LocalName,0,sizeof(LocalName));
    LocalName.sin_family=AF_INET;
    LocalName.sin_port=0;
    errno=0;
#if defined(_WIN32)
    my_inet_pton(AF_INET,localIP,&LocalName.sin_addr.S_un.S_addr);
#else
    my_inet_pton(AF_INET,localIP,&LocalName.sin_addr.s_addr);
#endif
    errno=0;

#if SOCKET_BIND_WILDCARD==1
    memset(&LocalName.sin_addr,0,sizeof(LocalName.sin_addr));
    memset(&LocalName.sin_port,0,sizeof(LocalName.sin_port));
#endif

    int ErrorCode=bind(Socket.handle,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
    if(ErrorCode!=0) {
        int la=error_code=WSAGetLastError();
        if(la!=0) {
            s_socket nul;
            nul.clear();
            nul.comment=comment;
            errno=0;
            return nul;
        }
        errno=0;
    }
    errno=0;

    if(async)
        sock_create_later(Socket,true);
    errno=0;

    ErrorCode=connect(Socket.handle,reinterpret_cast<sockaddr *>(&SockAddrIn),sizeof(SockAddrIn));
    if(ErrorCode!=0) {
        int la=error_code=WSAGetLastError();
        if(la!=0) {
            sock_erase_later(Socket);
            closesocket(Socket.handle);

            s_socket nul;
            nul.clear();
            nul.comment=comment;
            errno=0;
            return nul;
        }
        errno=0;
    }
    errno=0;

#endif

    errno=0;
    if(async)
        sock_async(Socket);
    errno=0;

    Socket.comment=comment;

    if(Socket.cmp()) {
        sock_set_tcp_keepalive(Socket);
        sock_register_socket(Socket.handle);
    }

    Socket.closed=false;

    error_code=0;
    return Socket;
}

/*!
    \brief Connects to remote host (for use in thread)
    \author VooDooMan
    \version 2
    \date 2004, 2005-12-07
    \param localIP Local IPv4 address as text to bind, if we are using IPv4 as destination, otherwise ignored
    \param remote Remote IPv4/IPv6 address as text, or IPv4/IPv6 host name
    \param remote_port Remote port
    \param error_code Returns OS's socket error
    \param comment Comment for socket
    \param log Returns data that should be logged
    \param errnum Returns error number (can be NULL)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \return If error_code == zero, returns socket handle of connected socket, else undefined
*/
s_socket sock_thread_connect(const string& localIP, const string& remote, unsigned short remote_port, int& error_code, const string& comment, vector<c_sock_log>& log, int* errnum, c_sock_log::e_type* errtype)
{
    error_code=0;

    s_socket Socket;

#ifdef IPv6

    s_socket nul;
    nul.clear();
    nul.comment=comment;

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    char port_name[64];
    sprintf(port_name,"%u",(unsigned int)remote_port);

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=PF_UNSPEC;
    Hints.ai_socktype=SOCK_STREAM;
    int err=0;
    if((err=getaddrinfo(remote.c_str(),port_name,&Hints,&AddrInfo))) {
#if defined(_WIN32)
        int ec=WSAGetLastError();
#else
        int ec=err;
#endif
        string log_="Cannot resolve address and port, while calling getaddrinfo() in file " __FILE__ " in function ";
        log_+=__func__;
        error_code=ec;
        if(ec)
            log.push_back(c_sock_log(ec,c_sock_log::TYPE_GAI,log_));
        s_socket nul;
        if(errnum) {
            *errnum=ec;
            *errtype=c_sock_log::TYPE_GAI;
        }
        return nul;
    }

    //
    // Try each address getaddrinfo returned, until we find one to which
    // we can sucessfully connect.
    //
    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {
        // Open a socket with the correct address family for this address.
        Socket.handle=socket(AI->ai_family,AI->ai_socktype,AI->ai_protocol);
        if(Socket.handle==INVALID_SOCKET) {
            string log_="Error opening socket, while calling socket() in file " __FILE__ " in function ";
            log_+=__func__;
#if defined(_WIN32)
            int ec=WSAGetLastError();
            if(ec)
                log.push_back(c_sock_log(ec,c_sock_log::TYPE_ERRNO,log_));
#else
            int ec=MY_UNKNOWN_ERROR;
            log.push_back(c_sock_log(ec,c_sock_log::TYPE_MY,log_));
#endif
            if(errnum) {
                *errnum=ec;
                *errtype=c_sock_log::TYPE_MY;
            }
            continue;
        }

        if(localIP.find(':')!=string::npos) {
            sockaddr_in6 LocalName;
            memset(&LocalName,0,sizeof(LocalName));
            LocalName.sin6_family=AF_INET6;
            LocalName.sin6_port=0;
            my_inet_pton(AF_INET6,localIP.c_str(),&LocalName.sin6_addr);

#if SOCKET_BIND_WILDCARD==1
            memset(&LocalName.sin6_addr,0,sizeof(LocalName.sin6_addr));
            memset(&LocalName.sin6_port,0,sizeof(LocalName.sin6_port));
#endif

            int ErrorCode=bind(Socket.handle,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
            if(ErrorCode!=0) {
#if defined(_WIN32)
                int la=error_code=WSAGetLastError();
#else
                int la=error_code=ErrorCode;
#endif
                if(la!=0) {
                    s_socket nul;
                    nul.clear();
                    nul.comment=comment;
                    if(errtype)
                        *errtype=c_sock_log::TYPE_ERRNO;
                    return nul;
                }
            }
        } else {
            sockaddr_in LocalName;
            memset(&LocalName,0,sizeof(LocalName));
            LocalName.sin_family=AF_INET;
            LocalName.sin_port=0;
#if defined(_WIN32)
            my_inet_pton(AF_INET,localIP.c_str(),&LocalName.sin_addr.S_un.S_addr);
#else
            my_inet_pton(AF_INET,localIP.c_str(),&LocalName.sin_addr.s_addr);
#endif

#if SOCKET_BIND_WILDCARD==1
            memset(&LocalName.sin_addr,0,sizeof(LocalName.sin_addr));
            memset(&LocalName.sin_port,0,sizeof(LocalName.sin_port));
#endif

            int ErrorCode=bind(Socket.handle,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
            if(ErrorCode!=0) {
#if defined(_WIN32)
                int la=error_code=WSAGetLastError();
#else
                int la=error_code=ErrorCode;
#endif
                if(la!=0) {
                    s_socket nul;
                    nul.clear();
                    nul.comment=comment;
                    if(errtype)
                        *errtype=c_sock_log::TYPE_ERRNO;
                    return nul;
                }
            }
        }

        int err=0;
        if((err=connect(Socket.handle,AI->ai_addr,AI->ai_addrlen))!=0) {
#if defined(_WIN32)
            int la=error_code=WSAGetLastError();
#else
            int la=error_code=errno;
#endif
            closesocket(Socket.handle);
            string log_="Error connecting socket, while calling connect() in file " __FILE__ " in function ";
            log_+=__func__;
            if(la) {
                log_+=", ";
                char* e=NULL;
                log_+=e=sock_error(la,c_sock_log::TYPE_ERRNO);
                delete[] e;
                log.push_back(c_sock_log(la,c_sock_log::TYPE_ERRNO,log_));
            }
            continue;
        }
        break;
    }

    freeaddrinfo(AddrInfo);

    if(!AI) {
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        return nul;
    }

#else

    vector<in_addr> a=sock_resolve_multi(remote.c_str(),log,errnum,errtype);
    if(a.size()==0) {
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        return nul;
    }
    int r=rand()%a.size();

    sockaddr_in SockAddrIn;
    SockAddrIn.sin_family=AF_INET;

    memcpy(&(SockAddrIn.sin_addr),&(a.at(r)),sizeof(SockAddrIn.sin_addr));

    SockAddrIn.sin_port=htons(remote_port);

    Socket.handle=socket(PF_INET,SOCK_STREAM,sock_tcp_proto_num);
    if(Socket.handle==SOCKET_ERROR) {
#if defined(_WIN32)
        int la=error_code=WSAGetLastError();
#else
        int la=error_code=Socket.handle;
#endif
        if(la!=0) {
            s_socket nul;
            nul.clear();
            nul.comment=comment;
            if(errtype)
                *errtype=c_sock_log::TYPE_ERRNO;
            return nul;
        }
    }

    sockaddr_in LocalName;
    memset(&LocalName,0,sizeof(LocalName));
    LocalName.sin_family=AF_INET;
    LocalName.sin_port=0;
#if defined(_WIN32)
    my_inet_pton(AF_INET,localIP.c_str(),&LocalName.sin_addr.S_un.S_addr);
#else
    my_inet_pton(AF_INET,localIP.c_str(),&LocalName.sin_addr.s_addr);
#endif

#if SOCKET_BIND_WILDCARD==1
    memset(&LocalName.sin_addr,0,sizeof(LocalName.sin_addr));
    memset(&LocalName.sin_port,0,sizeof(LocalName.sin_port));
#endif

    int ErrorCode=bind(Socket.handle,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
    if(ErrorCode!=0) {
#if defined(_WIN32)
        int la=error_code=WSAGetLastError();
#else
        int la=error_code=ErrorCode;
#endif
        if(la!=0) {
            s_socket nul;
            nul.clear();
            nul.comment=comment;
            if(errtype)
                *errtype=c_sock_log::TYPE_ERRNO;
            return nul;
        }
    }

    ErrorCode=connect(Socket.handle,reinterpret_cast<sockaddr *>(&SockAddrIn),sizeof(SockAddrIn));
    if(ErrorCode!=0) {
#if defined(_WIN32)
        int la=error_code=WSAGetLastError();
#else
        int la=error_code=errno;
#endif
        if(la!=0) {
            closesocket(Socket.handle);

            s_socket nul;
            nul.clear();
            nul.comment=comment;
            string log_="Error connecting socket, while calling connect() in file " __FILE__ " in function ";
            log_+=__func__;
            if(la) {
                log_+=", ";
                char* e=NULL;
                log_+=e=sock_error(la,c_sock_log::TYPE_ERRNO);
                delete[] e;
                log.push_back(c_sock_log(la,c_sock_log::TYPE_ERRNO,log_));
            }
            return nul;
        }
    }

#endif

    Socket.comment=comment;

    error_code=0;
    return Socket;
}

/*!
    \brief Closes connected socket
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle of connected socket
*/
void sock_close(s_socket& socket)
{
    sock_unregister_socket(socket.handle);

    errno=0;
    sock_flush_later(socket);
    errno=0;
    sock_erase_later(socket);
    errno=0;
    if(socket.handle)
        closesocket(socket.handle);
    errno=0;
    socket.handle=0;
}

/*!
    \brief Resolves DNS name or text IPv4 address to IPv4 address via DNS
    \author VooDooMan
    \version 1
    \date 2006-07-03
    \param domain DNS name or IPv4 address as text to resolve
    \param errnum Error number (could be NULL if don't care)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \param log Returns data that should be logged
    \return Returns IPv4 address(es)
*/
vector<in_addr> sock_resolve_multi(const char* domain, vector<c_sock_log>& log, int* errnum, c_sock_log::e_type* errtype)
{
    vector<in_addr> res;

    if(errnum) {
        *errnum=0;
        *errtype=c_sock_log::TYPE_ERRNO;
    }
    char servInfo[256];

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    char port_name[64];
    sprintf(port_name,"%u",(unsigned int)80); // something must be set as port...

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=PF_INET;
    Hints.ai_socktype=SOCK_STREAM;
    Hints.ai_flags=AI_PASSIVE;

    int err=0;

    if((err=getaddrinfo(domain,port_name,&Hints,&AddrInfo))) {
        string log_="Cannot resolve address and port, while calling getaddrinfo() in file " __FILE__ " in function ";
        log_+=__func__;
#if defined(_WIN32)
        int ec=WSAGetLastError();
#else
        int ec=err;
#endif
        if(ec)
            log.push_back(c_sock_log(ec,c_sock_log::TYPE_GAI,log_));
        if(errnum) {
            *errnum=ec;
            *errtype=c_sock_log::TYPE_GAI;
        }
        return res;
    }

    char host[2048];

    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {
        if(AI->ai_family!=AF_INET)
            continue;

        host[0]='\0';

        if((err=getnameinfo(AI->ai_addr,AI->ai_addrlen, host, sizeof(host)-1, servInfo, sizeof(servInfo), NI_NUMERICHOST))!=0) {
            string log_="Cannot resolve address and port, while calling getnameinfo() in file " __FILE__ " in function ";
            log_+=__func__;
#if defined(_WIN32)
            int ec=WSAGetLastError();
#else
            int ec=err;
#endif
            if(ec)
                log.push_back(c_sock_log(ec,c_sock_log::TYPE_GAI,log_));
            if(errnum) {
                *errnum=ec;
                *errtype=c_sock_log::TYPE_GAI;
            }
            continue;
        }
        in_addr a4;
        if(host[0] && my_inet_pton(AF_INET,(const char*)host,&a4)==1) {
            res.push_back(a4);
        }
    }
    freeaddrinfo(AddrInfo);

    if(res.empty() && errnum) {
#if defined(_WIN32)
        *errnum=WSANO_DATA;
        *errtype=c_sock_log::TYPE_ERRNO;
#else
#   if defined(ENODATA)
        *errnum=ENODATA;
#   else
        *errnum=ENOENT;
#   endif
        *errtype=c_sock_log::TYPE_ERRNO;
#endif
    }

    return res;
}

/*!
    \brief Resolves IPv4 address to DNS name via DNS (reverse lookup)
    \author VooDooMan
    \version 1
    \date 2004
    \param address IPv4 address as text (dotted format)
    \param dns_name Buffer that receives DNS name (should be long enough)
    \param dns_max_size Length of dns_name buffer
    \param errnum Error number (could be NULL if don't care)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
*/
void sock_reverse(const char* address, char* dns_name, int dns_max_size, int* errnum, c_sock_log::e_type* errtype)
{
    unsigned int addr=0;
    if(errnum) {
        *errnum=0;
        *errtype=c_sock_log::TYPE_ERRNO;
    }
    my_inet_pton(AF_INET,address,&addr);
#ifndef HAVE_GETHOSTBYADDR_R
    errno=0;
    hostent* remoteHost=gethostbyaddr((char *)&addr, 4, AF_INET);
    if(errnum) {
        *errnum=ERRNO;
        *errtype=c_sock_log::TYPE_GETHOSTBY;
    }

    if(WSAGetLastError()==0 && remoteHost) {
        errno=0;
        strncpy(dns_name,remoteHost->h_name,dns_max_size);
        dns_name[dns_max_size-1]=0;
    } else
        dns_name[0]=0;
    errno=0;
#else
    hostent remoteHost;
    size_t tmpl=1024*32;
    char* tmp=new char[tmpl];
    hostent* tmp2=NULL;
    int errnop=0;
    if(gethostbyaddr_r((char*)&addr,4,AF_INET,&remoteHost,tmp,tmpl,&tmp2,&errnop)==0 && tmp2) {
        strncpy(dns_name,tmp2->h_name,dns_max_size);
        dns_name[dns_max_size-1]=0;
    } else
        dns_name[0]=0;

    if(errnum) {
        *errnum=errnop;
        *errtype=c_sock_log::TYPE_GETHOSTBY;
    }

    delete[] tmp;
#endif
}

/*!
    \brief Binds server socket and start listening
    \author VooDooMan
    \version 2
    \date 2004, 2005-08-01
    \param local Local IPv4 address to bind to
    \param local_port Local port to bind
    \param error_code Returns OS's socket error
    \param comment Comment for socket
    \return If error_code == zero, returns server socket handle, else undefined
*/
s_socket sock_bind(in_addr local, unsigned short local_port, int& error_code, const char* comment)
{
    errno=0;
    sockaddr_in SockAddrIn;

    SockAddrIn.sin_family=AF_INET;
    SockAddrIn.sin_addr=local;
    SockAddrIn.sin_port=htons(local_port);
    errno=0;

    s_socket Socket;
    Socket.handle=socket(PF_INET,SOCK_STREAM,sock_tcp_proto_num);
    if(Socket.handle==SOCKET_ERROR) {
        error_code=WSAGetLastError();
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
    int ErrorCode=0;
#if !defined(_WIN32)
    // on Win32 this would be harmfull, only on POSIX it makes sense,
    // if socket is in TIME_WAIT state; on Win32 we would never get
    // "address already in use" error while socket is in TIME_WAIT,
    // but more dangerous on Win32 is that we will never get that error even
    // if address/port is really busy listening after setting socket option SO_REUSEADDR.
    int tmp=1;
    errno=0;
    ErrorCode=setsockopt(Socket.handle,SOL_SOCKET,SO_REUSEADDR,(const char*)&tmp,sizeof(tmp));
    if(ErrorCode!=0) {
        error_code=WSAGetLastError();
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
#endif
    errno=0;

#if SOCKET_BIND_WILDCARD==1
    memset(&SockAddrIn.sin_addr,0,sizeof(SockAddrIn.sin_addr));
    // do NOT reset port to 0! tihs is server socket!
#endif

    ErrorCode=bind(Socket.handle,reinterpret_cast<sockaddr*>(&SockAddrIn),sizeof(SockAddrIn));
    if(ErrorCode!=0) {
        error_code=WSAGetLastError();
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
    socklen_t i1=sizeof(SockAddrIn);
    ErrorCode=getsockname(Socket.handle,reinterpret_cast<sockaddr*>(&SockAddrIn),&i1);
    if(ErrorCode!=0) {
        error_code=WSAGetLastError();
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
    ErrorCode=listen(Socket.handle,5);
    if(ErrorCode!=0) {
        error_code=WSAGetLastError();
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
#if defined(_WIN32)
    errno=0;
    unsigned long non_blocking=1;
    ErrorCode=ioctlsocket(Socket.handle,FIONBIO,&non_blocking);
#else
    errno=0;
    int flags=fcntl(Socket.handle,F_GETFL,0);
    flags|=O_NONBLOCK;
    fcntl(Socket.handle,F_SETFL,flags);
    ErrorCode=0;
#endif
    if(ErrorCode!=0) {
        error_code=WSAGetLastError();
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;

    Socket.comment=comment;
    return Socket;
}

/*!
    \brief Binds server socket and start listening
    \author VooDooMan
    \version 1
    \date 2004
    \param local Local IPv6 address to bind to
    \param port Local port to bind
    \param error_code Returns OS's socket error
    \param comment Comment for socket
    \return If error_code == zero, returns server socket handle, else undefined
*/
s_socket sock_bind6(char* /*local*/, unsigned short /*port*/, int& error_code, const char* comment)
{
    errno=0;
#ifdef IPv6
    s_socket nul;
    nul.clear();
    nul.comment=comment;

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    char port_name[64];
    sprintf(port_name,"%u",(unsigned int)port);

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=PF_INET6;
    Hints.ai_socktype=SOCK_STREAM;
    Hints.ai_flags=AI_NUMERICHOST /*AI_PASSIVE*/;

    if(getaddrinfo(local,port_name,&Hints,&AddrInfo)) {
        string log="Cannot resolve address and port, while calling getaddrinfo() in file " __FILE__ " in function ";
        log+=__func__;
        int ec=WSAGetLastError();
        char* errstr=NULL;
        if(ec)
            log_socket(ec,errstr=sock_error(ec),log.c_str(),&nul);
        delete[] errstr;
        s_socket nul;
        nul.clear();
        nul.comment=comment;
        errno=0;
        return nul;
    }
    errno=0;

    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {
        errno=0;
        if(AI->ai_family!=AF_INET6)
            continue;

        s_socket Socket;

        Socket.handle=socket(AI->ai_family,AI->ai_socktype,AI->ai_protocol);
        if(Socket.handle==SOCKET_ERROR)
            continue;

        errno=0;

        addrinfo AI2;
        memcpy(&AI2,AI,sizeof(AI2));

#if SOCKET_BIND_WILDCARD==1
        memset(&(AI2.ai_addr),0,sizeof(AI2.ai_addr));
        // do NOT reset port to 0! tihs is server socket!
#endif

        int ErrorCode=bind(Socket.handle,AI2.ai_addr,AI2.ai_addrlen);
        if(ErrorCode!=0) {
            errno=0;
            closesocket(Socket.handle);
            errno=0;
            continue;
        }
        errno=0;

        ErrorCode=listen(Socket.handle,5);
        if(ErrorCode!=0) {
            errno=0;
            closesocket(Socket.handle);
            errno=0;
            continue;
        }
        errno=0;
    
#if defined(_WIN32)
        unsigned long non_blocking=1;
        errno=0;
        ErrorCode=ioctlsocket(Socket.handle,FIONBIO,&non_blocking);
        errno=0;
#else
        errno=0;
        int flags=fcntl(Socket.handle,F_GETFL,0);
        errno=0;
        flags|=O_NONBLOCK;
        fcntl(Socket.handle,F_SETFL,flags);
        ErrorCode=0;
        errno=0;
#endif
        if(ErrorCode!=0) {
            errno=0;
            closesocket(Socket.handle);
            errno=0;
            continue;
        }
        errno=0;

        freeaddrinfo(AddrInfo);
        errno=0;

        Socket.comment=comment;
        error_code=0;
        return Socket;
    }

    errno=0;
    freeaddrinfo(AddrInfo);
    errno=0;

    error_code=WSAGetLastError();
    nul.clear();
    nul.comment=comment;
    errno=0;
    return nul;

#else
    error_code=MY_NO_IPV6;
    s_socket nul;
    nul.clear();
    nul.comment=comment;
    errno=0;
    return nul;
#endif
}

/*!
    \brief Accepts client connection on listening server socket
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle of listening server socket
    \param comment Comment for socket
    \param async true for asynchronous mode
    \return Returns socket handle of client connection
*/
s_socket sock_accept(s_socket& socket, const char* comment, bool async)
{
    errno=0;
    fd_set s_;
    FD_ZERO(&s_);
    FD_SET(socket.handle,&s_);
    timeval timeout;
    timeout.tv_sec=0;
    timeout.tv_usec=100;
    int num=select((int)socket.handle+1,&s_,NULL,NULL,&timeout);
    errno=0;

    s_socket nul;
    nul.clear();
    nul.comment=comment;

    if(num==0)
        return nul;
    sockaddr_in SockAddrIn;
    socklen_t i1=sizeof(SockAddrIn);
    errno=0;
    SOCKET ec=accept(socket.handle,reinterpret_cast<sockaddr *>(&SockAddrIn),&i1);

    if(ec!=INVALID_SOCKET) {
        s_socket s;
        s.clear();
        s.handle=ec;

        if(async) {
            sock_create_later(s,true);
            sock_async(s);
        }

        sock_set_tcp_keepalive(s);

        sock_register_socket(s.handle);

        s.comment=comment;
        errno=0;
        return s;
    }
    errno=0;

    return nul;
}

/*!
    \brief Accepts client connection on listening server socket
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle of listening server socket
    \param comment Comment for socket
    \param async true for asynchronous mode
    \return Returns socket handle of client connection
*/
s_socket sock_accept6(s_socket& /*socket*/, const char* comment, bool /*async*/)
{
    errno=0;
    s_socket nul;
    nul.clear();
    nul.comment=comment;
#ifdef IPv6
    fd_set s_;
    FD_ZERO(&s_);
    FD_SET(socket.handle,&s_);
    timeval timeout;
    timeout.tv_sec=0;
    timeout.tv_usec=100;
    int num=select(socket.handle+1,&s_,NULL,NULL,&timeout);
    errno=0;

    if(num==0)
        return nul;
    sockaddr_in6 SockAddrIn;
    socklen_t i1=sizeof(SockAddrIn);
    SOCKET ec=accept(socket.handle,reinterpret_cast<sockaddr *>(&SockAddrIn),&i1);
    errno=0;

    if(ec!=INVALID_SOCKET) {
        s_socket s;
        s.clear();
        s.handle=ec;

        errno=0;
        if(getnameinfo(reinterpret_cast<sockaddr *>(&SockAddrIn),sizeof(SockAddrIn),s.host,sizeof(s.host),NULL,0,NI_NUMERICHOST)!=0)
            s.host[0]=0;
        errno=0;

        if(async) {
            sock_create_later(s,true);
            sock_async(s);
        }

        sock_set_tcp_keepalive(s);

        sock_register_socket(s.handle);

        s.comment=comment;
        errno=0;
        return s;
    }

#endif
    errno=0;
    return nul;
}

/*!
    \brief Gets remote IPv4 address of remote host
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle of connected socket
    \param error_code Returns OS's socket error
    \return If error_code == zero, returns IPv4 address of remote host, else undefined
*/
in_addr sock_get_remote_addr(s_socket& socket, int& error_code)
{
    errno=0;
    sockaddr_in addr;
    socklen_t i2=sizeof(addr);
    error_code=getpeername(socket.handle,reinterpret_cast<sockaddr*>(&addr),&i2);
    errno=0;
    return addr.sin_addr;
}

/*!
    \brief Gets remote IPv6 address of remote host
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle of connected socket
    \return Returns IPv6 address of remote host, or zero
*/
in_addr6_ sock_get_remote_addr6(s_socket& /*socket*/)
{
    errno=0;
#ifdef IPv6
    char tmp[2048];
    strcpy(tmp,socket.host);
    if(strstr(tmp,"%"))
        *strstr(tmp,"%")=0;
    in_addr6_ addr;
    errno=0;
    inetpton(AF_INET6,(const char*)tmp,&addr);
    errno=0;
    return addr;
#else
    in_addr6_ nul;
    memset(&nul,0,sizeof(nul));
    errno=0;
    return nul;
#endif
}

/*!
    \brief Low-level send to socket (Windows(TM) only)
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle
    \param buff Buffer to send
    \param len Length of the buffer
    \return Returns number of bytes actually sent
*/
size_t sock_lo_send(s_socket& socket, char* buff, size_t len)
{
    errno=0;
    size_t sent=send(socket.handle,buff,(int)len,0);
    errno=0;
    if((signed)sent>0) {
        socket.last_io=time(NULL);
        socket.sent+=sent;
    }
    return sent;
}

/*!
    \brief Low-level: receives data from the socket input queue, but leaves it in it (can be used for check how mush data is waiting)
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle
    \param buff Buffer to receive
    \param len Length of the buffer
    \return Returns number of bytes waiting in queue
*/
size_t sock_lo_get_input_amount(s_socket& socket, char* buff, size_t len)
{
    errno=0;
    fd_set s_;
    FD_ZERO(&s_);
    FD_SET(socket.handle,&s_);
    timeval timeout;
    timeout.tv_sec=0;
    timeout.tv_usec=100;
    int num=select((int)socket.handle+1,&s_,NULL,NULL,&timeout);
    if(ERRNO)
        num=0;
    errno=0;
    size_t r=0;
    if(num)
        r=recvfrom(socket.handle,buff,(int)len,MSG_PEEK,NULL,NULL);
    if(r<0)
        r=0;
    errno=0;
    return r;
}

/*!
    \brief Low-level read from socket (Windows(TM) only)
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle
    \param buff Buffer to store received data
    \param len Maximum length of the buffer
    \return Returns number of bytes actually received
*/
size_t sock_lo_recv(s_socket& socket, char* buff, size_t len)
{
    errno=0;
    fd_set s_;
    FD_ZERO(&s_);
    FD_SET(socket.handle,&s_);
    timeval timeout;
    timeout.tv_sec=0;
    timeout.tv_usec=100;
    int num=select((int)socket.handle+1,&s_,NULL,NULL,&timeout);
    if(ERRNO)
        num=0;
    errno=0;
    size_t r=0;
    if(num)
        r=recvfrom(socket.handle,buff,(int)len,0,NULL,NULL);
    if(r<0)
        r=0;
    errno=0;
    if((signed)r>0) {
        socket.last_io=time(NULL);
        socket.received+=r;
    }
    return r;
}

/*!
    \brief Resolves DNS name or text IPv6 address to IPv6 address via DNS
    \author VooDooMan
    \version 1
    \date 2006-07-03
    \param domain DNS name or IPv6 address as text to resolve
    \param log Returns messages that should be logged
    \param errnum Returns error number (can be NULL)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \return Returns IPv6 address(es)
*/
vector<in_addr6_> sock_resolve6_multi(const char* /*domain*/, vector<c_sock_log>& /*log*/, int* errnum, c_sock_log::e_type* errtype)
{
    vector<in_addr6_> res;

    if(errnum) {
        *errnum=0;
        *errtype=c_sock_log::TYPE_ERRNO;
    }
#ifdef IPv6
    char servInfo[256];

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    char port_name[64];
    sprintf(port_name,"%u",(unsigned int)80); // something must be set as port...

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=PF_INET;
    Hints.ai_socktype=SOCK_STREAM;
    Hints.ai_flags=AI_PASSIVE;

    int err=0;

    if((err=getaddrinfo(domain,port_name,&Hints,&AddrInfo))) {
        string log_="Cannot resolve address and port, while calling getaddrinfo() in file " __FILE__ " in function ";
        log_+=__func__;
#if defined(_WIN32)
        int ec=WSAGetLastError();
#else
        int ec=err;
#endif
        if(ec)
            log.push_back(c_sock_log(ec,c_sock_log::TYPE_GAI,log_));
        if(errnum) {
            *errnum=ec;
            *errtype=c_sock_log::TYPE_GAI;
        }
        return res;
    }

    char host[2048];

    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {
        if(AI->ai_family!=AF_INET6)
            continue;

        host[0]='\0';

        if((err=getnameinfo(AI->ai_addr,AI->ai_addrlen, host, sizeof(host)-1, servInfo, sizeof(servInfo), NI_NUMERICHOST))!=0) {
            string log_="Cannot resolve address and port, while calling getnameinfo() in file " __FILE__ " in function ";
            log_+=__func__;
#if defined(_WIN32)
            int ec=WSAGetLastError();
#else
            int ec=err;
#endif
            if(ec)
                log.push_back(c_sock_log(ec,c_sock_log::TYPE_GAI,log_));
            if(errnum) {
                *errnum=ec;
                *errtype=c_sock_log::TYPE_GAI;
            }
            continue;
        }
        in_addr6_ a6;
        if(host[0] && my_inet_pton(AF_INET6,(const char*)host,&a6)==1) {
            res.push_back(a6);
        }
    }
    freeaddrinfo(AddrInfo);
#endif

    if(res.empty() && errnum) {
#if defined(_WIN32)
        *errnum=WSANO_DATA;
        *errtype=c_sock_log::TYPE_ERRNO;
#else
#   if defined(ENODATA)
        *errnum=ENODATA;
#   else
        *errnum=ENOENT;
#   endif
        *errtype=c_sock_log::TYPE_ERRNO;
#endif
    }

    return res;
}

/*!
    \brief Resolves IPv6 address to DNS name via DNS (reverse lookup)
    \author VooDooMan
    \version 1
    \date 2004
    \param address IPv6 address
    \param dns_name Buffer that receives DNS name (should be long enough)
    \param dns_max_size Length of dns_name buffer
    \param log Returns messages that should be logged
    \param errnum Error number (could be NULL if don't care)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \bug Doesn't work :-(
*/
void sock_reverse(in_addr6_ /*address*/, char* dns_name, int /*dns_max_size*/, vector<c_sock_log>& /*log*/, int* errnum, c_sock_log::e_type* errtype)
{
    if(errnum) {
        *errnum=0;
        *errtype=c_sock_log::TYPE_ERRNO;
    }
#ifdef IPv6
    char host[2048];
    char servInfo[256];

    sockaddr_in6 a;
    memset(&a,0,sizeof(a));
    memcpy(&a.sin6_addr,&address,sizeof(a.sin6_addr));
    a.sin6_family=AF_INET6;
    a.sin6_port=80; // something for port...
    int err=0;
    if((err=getnameinfo((sockaddr*)&a, sizeof(a), host, sizeof(host), servInfo, sizeof(servInfo), NI_NAMEREQD))!=0) {
        string log_="Error while calling getnameinfo() in file " __FILE__ " in function ";
        log_+=__func__;
#if defined(_WIN32)
        int ec=WSAGetLastError();
#else
        int ec=err;
#endif
        if(ec)
            log.push_back(c_sock_log(ec,c_sock_log::TYPE_GAI,log_));
        if(errnum) {
            *errnum=ec;
            *errtype=c_sock_log::TYPE_GAI;
        }
        host[0]=0;
    }
    memset(dns_name,0,dns_max_size);
    strncpy(dns_name,host,(dns_max_size<strlen(host))?dns_max_size:strlen(host));
#else
    dns_name[0]=0;
#endif
}

/*!
    \brief Sets or unsets the non-blocking mode for the socket
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle
    \param non_blocking true if to set non-blocking mode
*/
void sock_set_nonblocking(s_socket& socket, bool non_blocking)
{
#if defined(_WIN32)
    unsigned long non_blocking_=non_blocking?1:0;
    errno=0;
    ioctlsocket(socket.handle,FIONBIO,&non_blocking_);
    errno=0;
#else
    errno=0;
    int flags=fcntl(socket.handle,F_GETFL,0);
    errno=0;
    flags&=~O_NONBLOCK;
    flags|=non_blocking?O_NONBLOCK:0;
    fcntl(socket.handle,F_SETFL,flags);
    errno=0;
#endif
}

/*!
    \brief Sets comment for the socket
    \author VooDooMan
    \version 1
    \date 2005-08-01
    \param socket Socket handle
    \param comment Comment for socket
*/
void sock_comment(s_socket& socket, const string& comment)
{
    socket.comment=comment;
}

/*!
    \brief Gets remote port number
    \author VooDooMan
    \version 1
    \date 2006-01-11
    \param socket Socket handle of connected socket
*/
unsigned short sock_get_remote_port(s_socket& socket)
{
    sockaddr_in addr;
    socklen_t i2=sizeof(addr);
    errno=0;
    int error_code=getpeername(socket.handle,reinterpret_cast<sockaddr*>(&addr),&i2);
    errno=0;
    if(error_code)
        return 0;
    unsigned short r=ntohs(addr.sin_port);
    errno=0;
    return r;
}

/*!
    \brief Gets local port number
    \author VooDooMan
    \version 1
    \date 2006-01-11
    \param socket Socket handle of connected socket
*/
unsigned short sock_get_local_port(s_socket& socket)
{
    sockaddr_in addr;
    socklen_t i2=sizeof(addr);
    errno=0;
    int error_code=getsockname(socket.handle,reinterpret_cast<sockaddr*>(&addr),&i2);
    errno=0;
    if(error_code)
        return 0;
    unsigned short r=ntohs(addr.sin_port);
    errno=0;
    return r;
}

void sock_register_socket(SOCKET s)
{
    all_sockets.push_back(s);
}

void sock_unregister_socket(SOCKET s)
{
    for(vector<SOCKET>::iterator i1=all_sockets.begin(); i1!=all_sockets.end(); i1++) {
        if((*i1)==s) {
            all_sockets.erase(i1);
            break;
        }
    }
}

/*!
    \brief Checks if socket I/O queue is empty or not
    \author VooDooMan
    \version 1
    \date 2006-05-08
    \return Returns 'true' is it is empty, otherwise 'false'
*/
static bool sock_is_queue_empty()
{
    {
        for(vector<c_receive_later>::iterator i1=receive_later.begin(); i1!=receive_later.end(); i1++)
            if(!(*i1).data.empty() && !(*i1).socket.closed && !(*i1).has_error && !(*i1).using_window)
                return false;
    }
    {
        for(vector<c_send_later>::iterator i1=send_later.begin(); i1!=send_later.end(); i1++)
            if(!(*i1).data.empty()
#if defined(_WIN32)
                && (*i1).socket.wnd
#endif
                && !(*i1).socket.closed
                )
                return false;
    }

    return true;
}

#ifndef XTELNET

void sock_interruptible_sleep(unsigned int ms)
{
    if(ms<50) {
    common_sleep:
#if !defined(_WIN32)
        timeval t;
        t.tv_sec=0;
        t.tv_usec=ms*1000;
        select(1,NULL,NULL,NULL,&t);
#else
        SleepEx((DWORD)ms,true);
#endif
        return;
    }

    if(sock_is_queue_empty()) {
        goto common_sleep;
    }

    int one_more=1;
    fd_set s;
    FD_ZERO(&s);

    /*
    static unsigned last=0;
    if(last>all_sockets.size())
        last=0;
    */
    unsigned last=0;

    unsigned i2=0, i3=0, cnt=0;
    unsigned max=FD_SETSIZE;
    if(max>FD_SETSIZE_MAX)
        max=FD_SETSIZE_MAX;
    vector<SOCKET>::iterator i1=all_sockets.begin();

    while(i3<all_sockets.size()) {
        if(all_sockets.end()==i1) {
            i1=all_sockets.begin();
            i2=0;
            continue;
        }

        if(i3>=last) {
            last=i3;
            if(!FD_ISSET(*i1,&s)) {
                FD_SET(*i1,&s);
                cnt++;
                if(one_more<=(int)*i1)
                    one_more=(int)*i1+1;
            }
            if(cnt==max)
                break;
        }

        i1++;
        i2++;
        i3++;
    }

    timeval t;
    t.tv_sec=ms/1000;
    t.tv_usec=(ms%1000)*1000;

    select(one_more,&s,NULL,NULL,&t);
    // <- ...
    // 1. we want NULL for write request because we want to sleep!
    // 2. we want NULL for exceptions, to avoid bug in case socket is already invalid
    //    and forgotten to be removed from array
}

/*!
    \brief Report FD_SETSIZE size to logs
    \version 1
    \author VooDooMan
    \date 2006-05-31
    \param anyway Argument 'anyway' to pass to logging function
*/
void sock_log_fd_array_size(bool anyway)
{
    unsigned max=FD_SETSIZE;
    if(max>FD_SETSIZE_MAX)
        max=FD_SETSIZE_MAX;

    s_socket nul;
    nul.clear();
    string log="Using ";
    char tmp[64];
    log+=ltoa((long)max,tmp,10);
    log+=" rotating sockets as maximum for idle sleeper (FD_SETSIZE==";
    log+=ltoa((long)FD_SETSIZE,tmp,10);
    log+=")";
    log_socket(0,"",log.c_str(),&nul,anyway);
}

#endif

/*!
    \brief Resolves DNS name or text IPv4 address to IPv4 address via DNS
    \author VooDooMan
    \version 2
    \date 2004, 2006-07-03
    \param domain DNS name or IPv4 address as text to resolve
    \param errnum Error number (could be NULL if don't care)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \return Returns IPv4 address (if there are more than one, returns randomly selected one)
*/
in_addr sock_resolve(const char* domain, int* errnum, c_sock_log::e_type* errtype)
{
    in_addr addr;

    vector<c_sock_log> log;
    vector<in_addr> a=sock_resolve_multi(domain,log,errnum,errtype);

    vector<c_sock_log>::iterator i1;
    s_socket nul;
    nul.clear();
    for(i1=log.begin(); i1!=log.end(); i1++) {
#ifndef XTELNET
        char* errstr=NULL;
        log_socket((*i1).ec,errstr=sock_error((*i1).ec,(*i1).type),(*i1).log.c_str(),&nul);
        delete[] errstr;
#endif
    }

    if(a.size()==0) {
        memset(&addr,0,sizeof(addr));
        return addr;
    }
    int r=rand()%a.size();
    memcpy(&addr,&(a.at(r)),sizeof(addr));
    return addr;
}

/*!
    \brief Resolves DNS name or text IPv6 address to IPv6 address via DNS
    \author VooDooMan
    \version 2
    \date 2004, 2006-07-03
    \param domain DNS name or IPv6 address as text to resolve
    \param log Returns messages that should be logged
    \param errnum Returns error number (can be NULL)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \return Returns IPv6 address (if there are more than one, returns randomly selected one)
*/
in_addr6_ sock_resolve6(const char* domain, vector<c_sock_log>& log, int* errnum, c_sock_log::e_type* errtype)
{
    in_addr6_ addr;

    vector<in_addr6_> a=sock_resolve6_multi(domain,log,errnum,errtype);
    if(a.size()==0) {
        memset(&addr,0,sizeof(addr));
        return addr;
    }
    int r=rand()%a.size();
    memcpy(&addr,&(a.at(r)),sizeof(addr));
    return addr;
}

/*!
    \brief Check wether connection was closed
    \author VooDooMan
    \version 1
    \date 2006-07-10
    \param socket Socket handle of connected socket
    \return Return true for error, or connection was closed
*/
bool sock_closed(s_socket& socket)
{
    return socket.closed;
}

