/***************************************************************************
                          sock_shared.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2009  Marian VooDooMan Meravy
    email                : ghostvoodooman@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2009  Marian VooDooMan Meravy
  (ghostvoodooman@users.sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides low-level implementation of TCP IP sockets
*/

#ifndef SOCK_SHARED_H__
#define SOCK_SHARED_H__

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <time.h>
#include <string>

using namespace std;

#if !defined(_WIN32)
#   include <sys/types.h>
#   include <sys/socket.h>
#   include <netinet/in.h>
#   include <arpa/inet.h>
#   include <netdb.h>

#   include <errno.h>
#   include <fcntl.h>

#   define WSAGetLastError() errno
#   define WSAEWOULDBLOCK EAGAIN
#   define closesocket close
#   define ioctlsocket ioctl
#   define SOCKET int
#   define PROTOENT protoent
#   define SOCKET_ERROR -1
#   define INVALID_SOCKET -1
#else
#   include <winsock2.h>
#endif

//#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"
#include "compat.h"

#include "assert.h"

#ifndef USE_OPENSSL
#   include <gnutls/gnutls.h>
#endif

/*!
    \brief Stores IPv6 address
    \author VooDooMan
    \version 1
    \date 2004
*/
struct in_addr6_ {
    unsigned char bytes[16];            //!< 16 bytes of IPv6 address
    inline bool operator< (const in_addr6_& than) const
    {
        for(size_t i1=0; i1<16; i1++) {
            if(bytes[i1]<than.bytes[i1])
                return true;
        }
        return false;
    }
    inline bool operator== (const in_addr6_& as) const
    {
        for(size_t i1=0; i1<16; i1++) {
            if(bytes[i1]!=as.bytes[i1])
                return false;
        }
        return true;
    }
};

/*!
    \brief Stores informnations about socket handle
    \author VooDooMan
    \version 1
    \date 2004
*/
struct s_socket {
    SOCKET handle;                          //!< OS Handle to socket
#if defined(_WIN32)
    HWND wnd;                               //!< Window handle for socket event messages (Windows(TM) only)
#endif

    string comment;                         //!< Comment for socket

    size_t received;                        //!< Number of bytes received
    size_t sent;                            //!< Number of bytes sent

#ifndef USE_OPENSSL
    bool got_xcred;                         //!< If to free "xcred"
    gnutls_certificate_credentials xcred;   //!< GNUTLS credentials
#endif

    time_t last_io;                         //!< Timestamp of last receive, or send data (only for client sockets)

    bool closed;                            //!< true if it is closed

    char host[2048];                        //!< If it is IPv6 server, there is numeric host of remote peer

    /*!
        \brief Clears the structure
        \author VooDooMan
        \version 1
        \date 2004
    */
    void clear() {
        handle=(unsigned)-1;
#if defined(_WIN32)
        wnd=NULL;
#endif
        comment="";
        received=sent=0;
#ifndef USE_OPENSSL
        got_xcred=false;
#endif
        last_io=time(NULL);
        closed=true;
        host[0]='\0';
    }

    /*!
        \brief Returns true if handle to socket is not -1
        \author VooDooMan
        \version 1
        \date 2004
    */
    bool cmp()
    {
        return handle!=-1;
    }

    /*!
        \brief Compares two handles for identicality
        \author VooDooMan
        \version 1
        \date 2004
    */
    inline bool operator== (s_socket& x) {
        return this->handle==x.handle;
    }

    s_socket() {
        clear();
    }

    s_socket(const s_socket& orig)
    {
        handle=orig.handle;
#if defined(_WIN32)
        wnd=orig.wnd;
#endif

        comment=orig.comment;

        received=orig.received;
        sent=orig.sent;

#ifndef USE_OPENSSL
        got_xcred=orig.got_xcred;
        xcred=orig.xcred;
#endif

        last_io=orig.last_io;

        closed=orig.closed;

        memcpy(host,orig.host,sizeof(host));
    }
};

#endif
