/***************************************************************************
                          ssl.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2009  Marian VooDooMan Meravy
    email                : ghostvoodooman@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2009  Marian VooDooMan Meravy
  (ghostvoodooman@users.sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides interface to SSL library
*/

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <deque>

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <errno.h>

#if defined(_WIN32)
#   include <winsock2.h>
#   include <process.h>
#else
#    include <sys/types.h>
#    include <unistd.h>
#endif

#ifdef HAVE_SYS_TIME_H
#   include <sys/time.h>
#endif

#ifdef USE_OPENSSL
#   include <openssl/rsa.h>       /* SSLeay stuff */
#   include <openssl/crypto.h>
#   include <openssl/x509.h>
#   include <openssl/pem.h>
#   include <openssl/ssl.h>
#   include <openssl/err.h>
#   include <openssl/x509v3.h>
#   include <openssl/rand.h>
#endif

//#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"

#if !defined(_WIN32)
#   ifdef USE_PTHREAD
#       ifdef HAVE_LIBPTHREAD
#           include <pthread.h>
#       endif
#   else
#       ifdef HAVE_LIBPTHREAD
#           undef HAVE_LIBPTHREAD
#       endif
#   endif
#endif

#include "compat.h"

#ifndef USE_OPENSSL
#   include <gnutls/gnutls.h>
#   include <gnutls/x509.h>
#   include <gcrypt.h>
#endif

#if !defined(_MSC_VER) && !defined(__BORLANDC__)
#   ifndef USE_OPENSSL
#       include <gcrypt.h>
#       if defined(_WIN32)
#           define gnutls_init_msvc_thread gnutls_init_threads
#       endif
#   endif
#endif

#ifndef __BORLANDC__
#   define MINGW_EAGAIN EAGAIN
#else
// BCB has non-standard EAGAIN, so set correct value
// as it complians with MinGW
#   define MINGW_EAGAIN 11L
#endif

#pragma hdrstop

using namespace std;

#include "ssl.h"
#include "sock.h"
#include "log.h"
#include "compat.h"

//#include "precomp.h"

int num_transactions=0;

#ifndef USE_OPENSSL

#define ERR_error_string XXXXXXXXXXX /* to prohibit using this function */

ssize_t pull_func(gnutls_transport_ptr handle, void* data, size_t size)
{
    s_socket* s=(s_socket*)(((c_ssl*)handle)->s);
    bool closed=false;
    size_t got=sock_get_receive_size(*s,closed);
    if(got<size) {
        if(closed) {
#if defined(_MSC_VER) || defined(__BORLANDC__)
            gnutls_transport_set_errno(((c_ssl*)handle)->handle,-1);
#else
            errno=-1;
#endif
            return -1;
        }
#if defined(_MSC_VER) || defined(__BORLANDC__)
        gnutls_transport_set_errno(((c_ssl*)handle)->handle,MINGW_EAGAIN);
#else
        errno=EAGAIN;
#endif
        return -1;
    }
    int ec=0;
    ssize_t res=0;
    char* d=(char*)data;
    while(res<size) {
        res+=sock_read(*s,&((char*)data)[res],size-res,ec,closed);
        if(ec || closed) {
#if defined(_MSC_VER) || defined(__BORLANDC__)
            gnutls_transport_set_errno(((c_ssl*)handle)->handle,-1);
#else
            errno=-1;
#endif
            return -1;
        }
    }
    return res;
}

ssize_t push_func(gnutls_transport_ptr handle, const void* data, size_t size)
{
    s_socket* s=(s_socket*)(((c_ssl*)handle)->s);
    int ec=0;
    ssize_t res=sock_send(*s,(const char*)data,size,ec);
    if(ec)
        res=-1;
    return res;
}

/*!
    \brief Provides better error string
    \param rc Error code
    \return return error string
*/
static string my_ERR_error_string(int rc)
{
    char tmp[64];
    string err=ltoa((long)rc,tmp,10);
    err+=", 0x";
    ltoa((long)rc,tmp,16);
    err+=tmp;
    err+=": ";
    switch(rc) {
        case GNUTLS_E_SUCCESS:
            err+="GNUTLS_E_SUCCESS";
            break;
        case GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM:
            err+="GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM";
            break;
        case GNUTLS_E_UNKNOWN_CIPHER_TYPE:
            err+="GNUTLS_E_UNKNOWN_CIPHER_TYPE";
            break;
        case GNUTLS_E_LARGE_PACKET:
            err+="GNUTLS_E_LARGE_PACKET";
            break;
        case GNUTLS_E_UNSUPPORTED_VERSION_PACKET:
            err+="GNUTLS_E_UNSUPPORTED_VERSION_PACKET";
            break;
        case GNUTLS_E_UNEXPECTED_PACKET_LENGTH:
            err+="GNUTLS_E_UNEXPECTED_PACKET_LENGTH";
            break;
        case GNUTLS_E_INVALID_SESSION:
            err+="GNUTLS_E_INVALID_SESSION";
            break;
        case GNUTLS_E_FATAL_ALERT_RECEIVED:
            err+="GNUTLS_E_FATAL_ALERT_RECEIVED";
            break;
        case GNUTLS_E_UNEXPECTED_PACKET:
            err+="GNUTLS_E_UNEXPECTED_PACKET";
            break;
        case GNUTLS_E_WARNING_ALERT_RECEIVED:
            err+="GNUTLS_E_WARNING_ALERT_RECEIVED";
            break;
        case GNUTLS_E_ERROR_IN_FINISHED_PACKET:
            err+="GNUTLS_E_ERROR_IN_FINISHED_PACKET";
            break;
        case GNUTLS_E_UNEXPECTED_HANDSHAKE_PACKET:
            err+="GNUTLS_E_UNEXPECTED_HANDSHAKE_PACKET";
            break;
        case GNUTLS_E_UNKNOWN_CIPHER_SUITE:
            err+="GNUTLS_E_UNKNOWN_CIPHER_SUITE";
            break;
        case GNUTLS_E_UNWANTED_ALGORITHM:
            err+="GNUTLS_E_UNWANTED_ALGORITHM";
            break;
        case GNUTLS_E_MPI_SCAN_FAILED:
            err+="GNUTLS_E_MPI_SCAN_FAILED";
            break;
        case GNUTLS_E_DECRYPTION_FAILED:
            err+="GNUTLS_E_DECRYPTION_FAILED";
            break;
        case GNUTLS_E_MEMORY_ERROR:
            err+="GNUTLS_E_MEMORY_ERROR";
            break;
        case GNUTLS_E_DECOMPRESSION_FAILED:
            err+="GNUTLS_E_DECOMPRESSION_FAILED";
            break;
        case GNUTLS_E_COMPRESSION_FAILED:
            err+="GNUTLS_E_COMPRESSION_FAILED";
            break;
        case GNUTLS_E_AGAIN:
            err+="GNUTLS_E_AGAIN";
            break;
        case GNUTLS_E_EXPIRED:
            err+="GNUTLS_E_EXPIRED";
            break;
        case GNUTLS_E_DB_ERROR:
            err+="GNUTLS_E_DB_ERROR";
            break;
        case GNUTLS_E_SRP_PWD_ERROR:
            err+="GNUTLS_E_SRP_PWD_ERROR";
            break;
        case GNUTLS_E_INSUFFICIENT_CREDENTIALS:
        //case GNUTLS_E_INSUFICIENT_CREDENTIALS:
        //case GNUTLS_E_INSUFFICIENT_CRED:
        //case GNUTLS_E_INSUFICIENT_CRED:
            err+="GNUTLS_E_INSUFFICIENT_CREDENTIALS";
            break;
        case GNUTLS_E_HASH_FAILED:
            err+="GNUTLS_E_HASH_FAILED";
            break;
        case GNUTLS_E_BASE64_DECODING_ERROR:
            err+="GNUTLS_E_BASE64_DECODING_ERROR";
            break;
        case GNUTLS_E_MPI_PRINT_FAILED:
            err+="GNUTLS_E_MPI_PRINT_FAILED";
            break;
        case GNUTLS_E_REHANDSHAKE:
            err+="GNUTLS_E_REHANDSHAKE";
            break;
        case GNUTLS_E_GOT_APPLICATION_DATA:
            err+="GNUTLS_E_GOT_APPLICATION_DATA";
            break;
        case GNUTLS_E_RECORD_LIMIT_REACHED:
            err+="GNUTLS_E_RECORD_LIMIT_REACHED";
            break;
        case GNUTLS_E_ENCRYPTION_FAILED:
            err+="GNUTLS_E_ENCRYPTION_FAILED";
            break;
        case GNUTLS_E_PK_ENCRYPTION_FAILED:
            err+="GNUTLS_E_PK_ENCRYPTION_FAILED";
            break;
        case GNUTLS_E_PK_DECRYPTION_FAILED:
            err+="GNUTLS_E_PK_DECRYPTION_FAILED";
            break;
        case GNUTLS_E_PK_SIGN_FAILED:
            err+="GNUTLS_E_PK_SIGN_FAILED";
            break;
        case GNUTLS_E_X509_UNSUPPORTED_CRITICAL_EXTENSION:
            err+="GNUTLS_E_X509_UNSUPPORTED_CRITICAL_EXTENSION";
            break;
        case GNUTLS_E_KEY_USAGE_VIOLATION:
            err+="GNUTLS_E_KEY_USAGE_VIOLATION";
            break;
        case GNUTLS_E_NO_CERTIFICATE_FOUND:
            err+="GNUTLS_E_NO_CERTIFICATE_FOUND";
            break;
        case GNUTLS_E_INVALID_REQUEST:
            err+="GNUTLS_E_INVALID_REQUEST";
            break;
        case GNUTLS_E_SHORT_MEMORY_BUFFER:
            err+="GNUTLS_E_SHORT_MEMORY_BUFFER";
            break;
        case GNUTLS_E_INTERRUPTED:
            err+="GNUTLS_E_INTERRUPTED";
            break;
        case GNUTLS_E_PUSH_ERROR:
            err+="GNUTLS_E_PUSH_ERROR";
            break;
        case GNUTLS_E_PULL_ERROR:
            err+="GNUTLS_E_PULL_ERROR";
            break;
        case GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER:
            err+="GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER";
            break;
        case GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE:
            err+="GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE";
            break;
        case GNUTLS_E_PKCS1_WRONG_PAD:
            err+="GNUTLS_E_PKCS1_WRONG_PAD";
            break;
        case GNUTLS_E_RECEIVED_ILLEGAL_EXTENSION:
            err+="GNUTLS_E_RECEIVED_ILLEGAL_EXTENSION";
            break;
        case GNUTLS_E_INTERNAL_ERROR:
            err+="GNUTLS_E_INTERNAL_ERROR";
            break;
        case GNUTLS_E_DH_PRIME_UNACCEPTABLE:
            err+="GNUTLS_E_DH_PRIME_UNACCEPTABLE";
            break;
        case GNUTLS_E_FILE_ERROR:
            err+="GNUTLS_E_FILE_ERROR";
            break;
        case GNUTLS_E_TOO_MANY_EMPTY_PACKETS:
            err+="GNUTLS_E_TOO_MANY_EMPTY_PACKETS";
            break;
        case GNUTLS_E_UNKNOWN_PK_ALGORITHM:
            err+="GNUTLS_E_UNKNOWN_PK_ALGORITHM";
            break;
        case GNUTLS_E_INIT_LIBEXTRA:
            err+="GNUTLS_E_INIT_LIBEXTRA";
            break;
        case GNUTLS_E_LIBRARY_VERSION_MISMATCH:
            err+="GNUTLS_E_LIBRARY_VERSION_MISMATCH";
            break;
        case GNUTLS_E_NO_TEMPORARY_RSA_PARAMS:
            err+="GNUTLS_E_NO_TEMPORARY_RSA_PARAMS";
            break;
        case GNUTLS_E_LZO_INIT_FAILED:
            err+="GNUTLS_E_LZO_INIT_FAILED";
            break;
        case GNUTLS_E_NO_COMPRESSION_ALGORITHMS:
            err+="GNUTLS_E_NO_COMPRESSION_ALGORITHMS";
            break;
        case GNUTLS_E_NO_CIPHER_SUITES:
            err+="GNUTLS_E_NO_CIPHER_SUITES";
            break;
        case GNUTLS_E_OPENPGP_GETKEY_FAILED:
            err+="GNUTLS_E_OPENPGP_GETKEY_FAILED";
            break;
        case GNUTLS_E_PK_SIG_VERIFY_FAILED:
            err+="GNUTLS_E_PK_SIG_VERIFY_FAILED";
            break;
        case GNUTLS_E_ILLEGAL_SRP_USERNAME:
            err+="GNUTLS_E_ILLEGAL_SRP_USERNAME";
            break;
        case GNUTLS_E_SRP_PWD_PARSING_ERROR:
            err+="GNUTLS_E_SRP_PWD_PARSING_ERROR";
            break;
        case GNUTLS_E_NO_TEMPORARY_DH_PARAMS:
            err+="GNUTLS_E_NO_TEMPORARY_DH_PARAMS";
            break;
        case GNUTLS_E_ASN1_ELEMENT_NOT_FOUND:
            err+="GNUTLS_E_ASN1_ELEMENT_NOT_FOUND";
            break;
        case GNUTLS_E_ASN1_IDENTIFIER_NOT_FOUND:
            err+="GNUTLS_E_ASN1_IDENTIFIER_NOT_FOUND";
            break;
        case GNUTLS_E_ASN1_DER_ERROR:
            err+="GNUTLS_E_ASN1_DER_ERROR";
            break;
        case GNUTLS_E_ASN1_VALUE_NOT_FOUND:
            err+="GNUTLS_E_ASN1_VALUE_NOT_FOUND";
            break;
        case GNUTLS_E_ASN1_GENERIC_ERROR:
            err+="GNUTLS_E_ASN1_GENERIC_ERROR";
            break;
        case GNUTLS_E_ASN1_VALUE_NOT_VALID:
            err+="GNUTLS_E_ASN1_VALUE_NOT_VALID";
            break;
        case GNUTLS_E_ASN1_TAG_ERROR:
            err+="GNUTLS_E_ASN1_TAG_ERROR";
            break;
        case GNUTLS_E_ASN1_TAG_IMPLICIT:
            err+="GNUTLS_E_ASN1_TAG_IMPLICIT";
            break;
        case GNUTLS_E_ASN1_TYPE_ANY_ERROR:
            err+="GNUTLS_E_ASN1_TYPE_ANY_ERROR";
            break;
        case GNUTLS_E_ASN1_SYNTAX_ERROR:
            err+="GNUTLS_E_ASN1_SYNTAX_ERROR";
            break;
        case GNUTLS_E_ASN1_DER_OVERFLOW:
            err+="GNUTLS_E_ASN1_DER_OVERFLOW";
            break;
#ifdef GNUTLS_E_OPENPGP_TRUSTDB_VERSION_UNSUPPORTED
        case GNUTLS_E_OPENPGP_TRUSTDB_VERSION_UNSUPPORTED:
            err+="GNUTLS_E_OPENPGP_TRUSTDB_VERSION_UNSUPPORTED";
            break;
#endif
        case GNUTLS_E_OPENPGP_UID_REVOKED:
            err+="GNUTLS_E_OPENPGP_UID_REVOKED";
            break;
        //case GNUTLS_E_X509_CERTIFICATE_ERROR:
        case GNUTLS_E_CERTIFICATE_ERROR:
            err+="GNUTLS_E_CERTIFICATE_ERROR";
            break;
        case GNUTLS_E_CERTIFICATE_KEY_MISMATCH:
            err+="GNUTLS_E_CERTIFICATE_KEY_MISMATCH";
            break;
        case GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE:
            err+="GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE";
            break;
        case GNUTLS_E_X509_UNKNOWN_SAN:
            err+="GNUTLS_E_X509_UNKNOWN_SAN";
            break;
        case GNUTLS_E_OPENPGP_FINGERPRINT_UNSUPPORTED:
            err+="GNUTLS_E_OPENPGP_FINGERPRINT_UNSUPPORTED";
            break;
        case GNUTLS_E_X509_UNSUPPORTED_ATTRIBUTE:
            err+="GNUTLS_E_X509_UNSUPPORTED_ATTRIBUTE";
            break;
        case GNUTLS_E_UNKNOWN_HASH_ALGORITHM:
            err+="GNUTLS_E_UNKNOWN_HASH_ALGORITHM";
            break;
        case GNUTLS_E_UNKNOWN_PKCS_CONTENT_TYPE:
            err+="GNUTLS_E_UNKNOWN_PKCS_CONTENT_TYPE";
            break;
        case GNUTLS_E_UNKNOWN_PKCS_BAG_TYPE:
            err+="GNUTLS_E_UNKNOWN_PKCS_BAG_TYPE";
            break;
        case GNUTLS_E_INVALID_PASSWORD:
            err+="GNUTLS_E_INVALID_PASSWORD";
            break;
        case GNUTLS_E_MAC_VERIFY_FAILED:
            err+="GNUTLS_E_MAC_VERIFY_FAILED";
            break;
        /*
        case GNUTLS_E_CONSTRAINT_ERROR:
            err+="GNUTLS_E_CONSTRAINT_ERROR";
            break;
        */
        case GNUTLS_E_BASE64_ENCODING_ERROR:
            err+="GNUTLS_E_BASE64_ENCODING_ERROR";
            break;
        case GNUTLS_E_INCOMPATIBLE_GCRYPT_LIBRARY:
        //case GNUTLS_E_INCOMPATIBLE_CRYPTO_LIBRARY:
            err+="GNUTLS_E_INCOMPATIBLE_CRYPTO_LIBRARY";
            break;
        case GNUTLS_E_INCOMPATIBLE_LIBTASN1_LIBRARY:
            err+="GNUTLS_E_INCOMPATIBLE_LIBTASN1_LIBRARY";
            break;
        case GNUTLS_E_OPENPGP_KEYRING_ERROR:
            err+="GNUTLS_E_OPENPGP_KEYRING_ERROR";
            break;
        case GNUTLS_E_X509_UNSUPPORTED_OID:
            err+="GNUTLS_E_X509_UNSUPPORTED_OID";
            break;
        /*case GNUTLS_E_RANDOM_FAILED:
            err+="GNUTLS_E_RANDOM_FAILED";
            break;*/
        case GNUTLS_E_UNIMPLEMENTED_FEATURE:
            err+="GNUTLS_E_UNIMPLEMENTED_FEATURE";
            break;
        default:
            err+="(other error)";
            break;
    }
    err+=": ";
    err+=gnutls_strerror(rc);
    return err;
}
#endif

/*!
    \def verify_depth_
    \brief Depth of verifying chained certificates
*/
#define verify_depth_    128

/*!
    \brief Stores data for verification purposes
    \author VooDooMan
    \version 2
    \date 2004, 2005-11-07
*/
class ssl_meta_data {
public:
    bool ok;                                        //!< If true, there's no verification error yet
    string error;                                   //!< Verification error message

    ssl_meta_data()
    {
        ok=true;
        error="";
    }
};

#ifdef USE_OPENSSL
map<SSL*,pair<int,ssl_meta_data*> > extra_data;      //!< SSL descriptor to extra data index
#endif

#ifndef USE_OPENSSL
/* Helper functions to load a certificate and key
 * files into memory.
 */
static gnutls_datum load_file(const char *file)
{
  FILE *f;
  gnutls_datum loaded_file = { NULL, 0 };
  long filelen;
  void *ptr;

  if (!(f = fopen(file, "r"))
      || fseek(f, 0, SEEK_END) != 0
      || (filelen = ftell(f)) < 0
      || fseek(f, 0, SEEK_SET) != 0
      || !(ptr = malloc((size_t)filelen))
      || fread(ptr, 1, (size_t)filelen, f) < (size_t)filelen)
    {
      return loaded_file;
    }

  loaded_file.data = (unsigned char*)ptr;
  loaded_file.size = (unsigned int)filelen;
  return loaded_file;
}

static void unload_file(gnutls_datum data)
{
  free(data.data);
}
#endif

/*!
    \brief Loads a certificate from file
    \author VooDooMan
    \version 1
    \date 2005
    \param filename Name of file
*/
#ifdef USE_OPENSSL
X509 *ssl_read_cert(char* filename)
#else
gnutls_x509_crt *ssl_read_cert(char* filename)
#endif
{
#ifdef USE_OPENSSL
    X509* cert;
    BIO* s=BIO_new_file(filename,"r");

    if(!s)
        return NULL;
          
    BIO* f=BIO_new(BIO_f_base64());
    if(!f) {
        BIO_free(s);
        return NULL;
    }
    s=BIO_push(f,s);
    cert=d2i_X509_bio(s,NULL);
    BIO_free_all(s);

    return cert;
#else
    gnutls_x509_crt crt;

    gnutls_datum data=load_file(filename);
    gnutls_x509_crt_init(&crt);

    int ret=gnutls_x509_crt_import(crt,&data,GNUTLS_X509_FMT_PEM);
    if(ret<0)
    {
      return NULL;
    }

    unload_file(data);

    gnutls_x509_crt* c=new gnutls_x509_crt(crt);

    return c;
#endif
}

#ifndef USE_OPENSSL
/* This function will try to verify the peer's certificate, and
 * also check if the hostname matches, and the activation, expiration dates.
 */
string verify_certificate(gnutls_session session, const char *hostname, const char* expected_cert)
{
    string res;

    unsigned int status;
    const gnutls_datum *cert_list;
    int ret;
    unsigned cert_list_size;
    gnutls_x509_crt cert;


    /* This verification function uses the trusted CAs in the credentials
    * structure. So you must have installed one or more CA certificates.
    */
    ret = gnutls_certificate_verify_peers2(session, &status);

    if (ret < 0)
        {
        //printf ("Error\n");
            if(!res.empty())
                res+=", ";
        res+="Certificate verifycation error";
        }

        if (status & GNUTLS_CERT_INVALID) {
            //printf ("The certificate is not trusted.\n");
            if(!res.empty())
                res+=", ";
            res+="The certificate is not trusted";
        }

        if (status & GNUTLS_CERT_SIGNER_NOT_FOUND) {
            //printf ("The certificate hasn't got a known issuer.\n");
            if(!res.empty())
                res+=", ";
            res+="The certificate hasn't got a known issuer";
        }

        if (status & GNUTLS_CERT_REVOKED) {
            //printf ("The certificate has been revoked.\n");
            if(!res.empty())
                res+=", ";
            res+="The certificate has been revoked";
        }


    /* Up to here the process is the same for X.509 certificates and
    * OpenPGP keys. From now on X.509 certificates are assumed. This can
    * be easily extended to work with openpgp keys as well.
    */
        if (gnutls_certificate_type_get (session) != GNUTLS_CRT_X509) {
            if(!res.empty())
                res+=", ";
            res+="Not an X.509 certificate";
            //gnutls_x509_crt_deinit (cert);
            return res;
        }

    if (gnutls_x509_crt_init (&cert) < 0)
        {
        //printf ("error in initialization\n");
            if(!res.empty())
                res+=", ";
            res+="Error in initialization";
            gnutls_x509_crt_deinit (cert);
        return res;
        }

    cert_list = gnutls_certificate_get_peers (session, &cert_list_size);
    if (cert_list == NULL)
        {
        //printf ("No certificate was found!\n");
            if(!res.empty())
                res+=", ";
            res+="No certificate found";
            gnutls_x509_crt_deinit (cert);
        return res;
        }

    /* This is not a real world example, since we only check the first 
    * certificate in the given chain.
    */
    if (gnutls_x509_crt_import (cert, &cert_list[0], GNUTLS_X509_FMT_DER) < 0)
        {
        //printf ("error parsing certificate\n");
            if(!res.empty())
                res+=", ";
            res+="Error parsing certificate";
            gnutls_x509_crt_deinit (cert);
        return res;
        }

    time_t t;
    /* Beware here we do not check for errors.
    */
    t=gnutls_x509_crt_get_expiration_time (cert);
    t&=0xFFFFFFFFUL; // MSVC has 64 bit my_time_t
    if (t < time (0))
        {
        //printf ("The certificate has expired\n");
            if(!res.empty())
                res+=", ";
            res+="The certificate has expired";
            gnutls_x509_crt_deinit (cert);
        return res;
        }

    t=gnutls_x509_crt_get_activation_time (cert);
    t&=0xFFFFFFFFUL; // MSVC has 64 bit my_time_t
    if (t > time (0))
        {
        //printf ("The certificate is not yet activated\n");
            if(!res.empty())
                res+=", ";
            res+="The certificate is not yet activated";
            gnutls_x509_crt_deinit (cert);
        return res;
        }

        /*
    if (!gnutls_x509_crt_check_hostname (cert, hostname))
        {
        printf ("The certificate's owner does not match hostname '%s'\n",
            hostname);
        return;
        }
        */

    if(expected_cert && expected_cert[0]!=0) {
        FILE* f=fopen(expected_cert,"r");
        if(!f) {
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="Could not open file \"";
            res+=expected_cert;
            res+="\"";
            return res;
        }
        string read_cert;
        while(!feof(f)) {
            char tmp[1024+1];
            tmp[1024]=0;
            fgets(tmp,1024,f);
            read_cert+=tmp;
        }
        fclose(f);
        f=NULL;
        char* tmp=new char[read_cert.length()+1];
        strcpy(tmp,read_cert.c_str());
        gnutls_datum read_cert2;
        read_cert2.data=(unsigned char*)tmp;
        read_cert2.size=read_cert.length();

        gnutls_x509_crt r_cert;
        int ret=gnutls_x509_crt_init (&r_cert);
        if(ret!=0) {
            delete[] tmp;
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="gnutls_x509_crt_init() failed: ";
            res+=my_ERR_error_string(ret);
            return res;
        }

        ret=gnutls_x509_crt_import(r_cert,&read_cert2,GNUTLS_X509_FMT_PEM);
        if(ret!=0) {
            gnutls_x509_crt_deinit (r_cert);
            delete[] tmp;
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="gnutls_x509_crt_import() failed: ";
            res+=my_ERR_error_string(ret);
            return res;
        }
        char buff_1[1024];
        char buff_2[1024];
        size_t buff_1_size=1024-1;
        size_t buff_2_size=1024-1;
        
        ret=gnutls_x509_crt_get_fingerprint (cert,GNUTLS_DIG_SHA,buff_1,&buff_1_size);
        if(ret!=0) {
            gnutls_x509_crt_deinit (r_cert);
            delete[] tmp;
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="gnutls_x509_crt_get_fingerprint() failed: ";
            res+=my_ERR_error_string(ret);
            return res;
        }

        ret=gnutls_x509_crt_get_fingerprint (r_cert,GNUTLS_DIG_SHA,buff_2,&buff_2_size);
        if(ret!=0) {
            gnutls_x509_crt_deinit (r_cert);
            delete[] tmp;
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="gnutls_x509_crt_get_fingerprint() failed: ";
            res+=my_ERR_error_string(ret);
            return res;
        }

        bool ok=true;
        if(buff_1_size!=buff_2_size)
            ok=false;
        if(ok)
            for(int ii=0; ii<buff_1_size; ii++)
                if(buff_1[ii]!=buff_2[ii]) {
                    ok=false;
                    break;
                }

        {
            gnutls_x509_crt_deinit (r_cert);
            delete[] tmp;
            gnutls_x509_crt_deinit (cert);

            if(!ok) {
                if(!res.empty())
                    res+=", ";
                res+="certificate verification failed: SHA1 hash doesn't match";
            }
            return res;
        }
    }

    gnutls_x509_crt_deinit (cert);

    return res;
}
#endif

#ifdef USE_OPENSSL
/*!
    \brief SSL verification callback
    \author Sampo Kellomaki <sampo@iki.fi>; Wade Scholine <wades@mail.cybg.com>; VooDooMan
    \version 1
    \date 1996-1998, 2004
    \param preverify_ok I don't actually know what the SSL wants to say by this parameter ;-)
    \param ctx Context of stored certificates
    \return Returns 1 (one) to continue the verification process, 0 (zero) to terminate
*/
int c_ssl::my_verify_callback(int preverify_ok, X509_STORE_CTX *ctx)
{
#   define buf__len    (2048L*2L)
    char    buf__[buf__len+16];
    X509   *err_cert;
    int     err, depth;
    SSL    *ssl;
    c_ssl *mydata;

    clear_error_queue();

    err_cert = X509_STORE_CTX_get_current_cert(ctx);
    err = X509_STORE_CTX_get_error(ctx);
    depth = X509_STORE_CTX_get_error_depth(ctx);

    /*
     * Retrieve the pointer to the SSL of the connection currently treated
     * and the application specific data stored into the SSL object.
     */
    ssl = (SSL*)X509_STORE_CTX_get_ex_data(ctx, SSL_get_ex_data_X509_STORE_CTX_idx());
    map<SSL*,pair<int,ssl_meta_data*> >::iterator i1=extra_data.find(ssl);
    if(extra_data.end()==i1)
        return 0;
    int index=(*i1).second.first;
    mydata = (c_ssl*) SSL_get_ex_data(ssl, index);
    if(!mydata)
        return 0;

    X509_NAME_oneline(X509_get_subject_name(err_cert), buf__, buf__len);

    /*
     * Catch a too long certificate chain. The depth limit set using
     * SSL_CTX_set_verify_depth() is by purpose set to "limit+1" so
     * that whenever the "depth>verify_depth" condition is met, we
     * have violated the limit and want to log this error condition.
     * We must do it here, because the CHAIN_TOO_LONG error would not
     * be found explicitly; only errors introduced by cutting off the
     * additional certificates would be logged.
     */
    if (depth > verify_depth_) {
        preverify_ok = 0;
        err = X509_V_ERR_CERT_CHAIN_TOO_LONG;
        X509_STORE_CTX_set_error(ctx, err);
    }
    if (!preverify_ok) {
        mydata->cert_ok_so_far=false;
        string error="verify error:num=";
        char tmp[64];
        error+=ltoa((long)err,tmp,10);
        error+=(string)":"+X509_verify_cert_error_string(err);
        error+=":depth=";
        error+=ltoa((long)depth,tmp,10);
        error+=(string)":"+buf__;

        X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert), buf__, buf__len);
        error+=" - remote certificate: ";
        error+=buf__;
        mydata->set_error(error);
    }

    return preverify_ok;
}
#endif

#ifdef USE_OPENSSL
SSL_CTX* ctx;                               //!< SSL context, unique for program's lifetime
SSL_METHOD *meth;                           //!< SSL method for context
#else
#endif

string CA_file;                             //!< File name with CA list

vector<c_ssl*> ssl_descriptors;             //!< All SSL descriptors
vector<pair<c_ssl*,time_t> > ssl_garbage;   //!< All SSL descriptors taht should be closed after some time

#ifndef USE_OPENSSL

gnutls_dh_params dh_params;
gnutls_rsa_params rsa_params;

gnutls_dh_params thread_dh_params;              //!< DH params for thread
gnutls_rsa_params thread_rsa_params;            //!< RSA params for thread
static volatile bool cs_init;                   //! Has been "cs"/"mutex" variable initialized?
#if defined(_WIN32)
static CRITICAL_SECTION cs;                     //!< Critical section (thread locking)
#else
#   ifdef HAVE_LIBPTHREAD
static pthread_mutex_t mutex;                   //!< Mutex for phtread (thread locking)
#   endif
#endif

volatile bool ssl_dh_thread_running=false;

/*!
    \brief Thread function for DH and RSA params regeneration
    \author VooDooMan
    \version 1
    \date 2005-12-16
    \param d Data for thread
*/
#if defined(_WIN32)
void __cdecl
#else
void*
#endif
ssl_dh_thread(void* d)
{
    gnutls_dh_params_init(&thread_dh_params);
    gnutls_rsa_params_init(&thread_rsa_params);
    gnutls_dh_params_generate2(thread_dh_params, DH_BITS);
    gnutls_rsa_params_generate2(thread_rsa_params, RSA_BITS);

#if defined(_WIN32)
again:
    try {
        EnterCriticalSection(&cs);
    } catch(...) {
        SleepEx(10,true);
        goto again;
    }
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_lock(&mutex);
#   endif
#endif

    ssl_dh_thread_running=false;

#if defined(_WIN32)
    LeaveCriticalSection(&cs);
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_unlock(&mutex);
#   endif
#endif

#if !defined(_WIN32)
    return NULL;
#endif
}

#endif

#if !defined(_MSC_VER) && !defined(__BORLANDC__)
#ifndef USE_OPENSSL
#ifdef USE_PTHREAD

static int gcry_pthread_mutex_init(void **priv)
{
    int err = 0;
    pthread_mutex_t *lock = (pthread_mutex_t *)malloc (sizeof (pthread_mutex_t));
    if (!lock)
        err = 12;
    if (!err) {
        err = pthread_mutex_init (lock, __null);
        if (err)
            free (lock);
        else
            *priv = lock;
    } return err;
}

static int gcry_pthread_mutex_destroy (void **lock)
{
    int err = pthread_mutex_destroy ((pthread_mutex_t *)*lock);
    free (*lock);
    return err;
}

static int gcry_pthread_mutex_lock (void **lock)
{
    return pthread_mutex_lock ((pthread_mutex_t *)*lock);
}

static int gcry_pthread_mutex_unlock (void **lock)
{
    return pthread_mutex_unlock ((pthread_mutex_t *)*lock);
}

static struct gcry_thread_cbs gcry_threads_pthread = { GCRY_THREAD_OPTION_PTHREAD, __null, gcry_pthread_mutex_init, gcry_pthread_mutex_destroy, gcry_pthread_mutex_lock, gcry_pthread_mutex_unlock };

#endif
#endif
#endif

void gnutls_init_msvc_thread();

/*!
    \brief Initializes SSL
    \author VooDooMan
    \version 1
    \date 2004
    \param err_str If return value is true, there's error string
    \param trustedCAs Path to file containing certificates of CA's tu trust
    \return true for error
*/
bool ssl_init(string& err_str, char* trustedCAs)
{
    CA_file=trustedCAs;
#ifdef USE_OPENSSL
    /* SSL preliminaries. We keep the certificate and key with the context. */

    SSL_load_error_strings();
    ERR_load_crypto_strings();

    SSL_library_init();                      /* initialize library */
    do {
        bool file_ok=false;
        FILE* f=fopen("openssl.rnd","rb");
        if(f) {
            file_ok=true;
            fseek(f,0,SEEK_END);
            long size=ftell(f);
            fclose(f);
            RAND_load_file("openssl.rnd",size);
        }
        if(!file_ok || !RAND_status()) {
            do {
                char tmp[1024*2+sizeof(int)+128];
                // not too much random... uninitialized stack, current time and PID
                for(size_t i1=0; i1<1024; i1++) {
                    int* p=(int*)(&(tmp[i1*2+0]));
                    *p+=rand()^((int)time(NULL));
                    p=(int*)(&(tmp[i1*2+1]));
                    *p+=rand()^((int)getpid());
                }
                RAND_seed(tmp,1024*2);
            } while(!RAND_status());
            RAND_write_file("openssl.rnd");
        }
    } while(!RAND_status());

    SSLeay_add_ssl_algorithms();
    meth = SSLv3_method();
    ctx = SSL_CTX_new (meth);

    char es[1024];

    if (!ctx) {
        es[0]=0;
        ERR_error_string(ERR_get_error(),es);
        err_str="SSL_CTX_new(): ";
        err_str+=es;
        log_ssl(err_str);
        return true;
    }

    STACK_OF(X509_NAME) *cert_names=SSL_load_client_CA_file(trustedCAs);
    if(!cert_names) {
        es[0]=0;
        ERR_error_string(ERR_get_error(),es);
        err_str="SSL_load_client_CA_file(): ";
        err_str+=es;
        err_str+=" (using file \"";
        err_str+=trustedCAs;
        err_str+="\")";
        log_ssl(err_str);
        return true;
    }
    SSL_CTX_set_client_CA_list(ctx,cert_names);

    if(!SSL_CTX_load_verify_locations(ctx,trustedCAs,NULL)) {
        es[0]=0;
        ERR_error_string(ERR_get_error(),es);
        err_str="SSL_CTX_load_verify_locations(): ";
        err_str+=es;
        log_ssl(err_str);
        return true;
    }

    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE|SSL_VERIFY_FAIL_IF_NO_PEER_CERT,c_ssl::my_verify_callback);
    SSL_CTX_set_verify_depth(ctx,verify_depth_ + 1);

    return false;
#else

#if !defined(_WIN32)
#   ifdef USE_PTHREAD
    gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
#   endif
#else
#   if defined(_MSC_VER) || defined(__BORLANDC__)
    gnutls_init_msvc_thread();
#   else
    gnutls_init_threads();
#   endif
#endif
    gnutls_global_init();

    /*
    gnutls_certificate_allocate_credentials(&x509_cred);
    gnutls_certificate_set_x509_trust_file(x509_cred,trustedCAs,GNUTLS_X509_FMT_PEM);
    */

    /*
    gnutls_certificate_set_x509_crl_file (x509_cred, CRLFILE,
                    GNUTLS_X509_FMT_PEM);
    */

    /*
    gnutls_certificate_set_x509_key_file (x509_cred, CERTFILE, KEYFILE,
                    GNUTLS_X509_FMT_PEM);
    */

    ssl_loop();

    /*
    gnutls_certificate_set_dh_params (x509_cred, dh_params);
    gnutls_certificate_set_rsa_export_params (x509_cred, rsa_params);
    */

    return false;
#endif
}

string ssl_get_ssl_version()
{
#ifdef USE_OPENSSL
    return OPENSSL_VERSION_TEXT;
#else
    string res="GNUTLS ";
    res+=LIBGNUTLS_VERSION;
    return res;
#endif
}

#if !defined(_MSC_VER) && !defined(__BORLANDC__)
#if defined(_WIN32)
#ifndef USE_OPENSSL
static struct gcry_thread_cbs gcry_threads_other;   //!< Descriptor for thread locking

/*!
    \brief Initialization of thread library
    \author VooDooMan
    \version 1
*/
static int thread_init()
{
    return 0;
}

/*!
    \brief Initialization of mutex
    \author VooDooMan
    \version 1
*/
static int thread_mutex_init(void **priv)
{
    CRITICAL_SECTION* cs=(CRITICAL_SECTION*)malloc(sizeof(CRITICAL_SECTION));
    InitializeCriticalSection(cs);
    *priv=(void*)cs;
    return 0;
}

/*!
    \brief Closing of mutex
    \author VooDooMan
    \version 1
*/
static int thread_mutex_destroy(void **priv)
{
    CRITICAL_SECTION* cs=(CRITICAL_SECTION*)*priv;
    DeleteCriticalSection(cs);
    free(cs);
    cs=NULL;
    *priv=(void*)cs;
    return 0;
}

/*!
    \brief Obtaining a mutex exclusion
    \author VooDooMan
    \version 1
*/
static int thread_mutex_lock(void **priv)
{
    CRITICAL_SECTION* cs=(CRITICAL_SECTION*)*priv;
    EnterCriticalSection(cs);
    *priv=(void*)cs;
    return 0;
}

/*!
    \brief Releasing of mutex exclusion
    \author VooDooMan
    \version 1
*/
static int thread_mutex_unlock(void **priv)
{
    CRITICAL_SECTION* cs=(CRITICAL_SECTION*)*priv;
    LeaveCriticalSection(cs);
    *priv=(void*)cs;
    return 0;
}

/*!
    \brief Introducing threads into gcrypto library
    \author VooDooMan
    \version 1
*/
void gnutls_init_msvc_thread()
{
    memset(&gcry_threads_other,0,sizeof(gcry_threads_other));
    gcry_threads_other.option=GCRY_THREAD_OPTION_USER;
    gcry_threads_other.init=thread_init;
    gcry_threads_other.mutex_init=thread_mutex_init;
    gcry_threads_other.mutex_destroy=thread_mutex_destroy;
    gcry_threads_other.mutex_lock=thread_mutex_lock;
    gcry_threads_other.mutex_unlock=thread_mutex_unlock;
    gcry_control (GCRYCTL_SET_THREAD_CBS, &gcry_threads_other);
}
#endif
#endif
#endif

/*!
    \brief Retrieves statistics about sock I/O cache
    \author VooDooMan
    \version 2
    \date 2006-01-22, 2006-07-11
    \param ssl SSL handle
    \param to_receive Returns number of bytes waiting in cache to receive them
    \param to_send Returns number of bytes waiting in cache to send them
*/
void c_ssl::get_cache_stats(size_t& to_receive, size_t& to_send)
{
    to_receive=to_send=0;

    deque<s_ssl_data>::iterator i1;

    for(i1=to_read.begin(); i1!=to_read.end(); i1++)
        to_receive+=(*i1).len;

    for(i1=to_write.begin(); i1!=to_write.end(); i1++)
        to_send+=(*i1).len;
}

void c_ssl::set_error(const string& err_string)
{
    if(error)
        return;
    error=true;
    err_str=err_string;
}

c_ssl::c_ssl(s_socket& socket, s_socket* server, const string& cert, const string& key)
{
#ifdef USE_OPENSSL
    rbio=wbio=NULL;
#endif

    clear_error_queue();

    s=new s_socket(socket);
    cert_ok_so_far=true;
    error=false;
    state=disconnected;
    force_write=false;

    num_transactions++;
    ssl_descriptors.push_back(this);

    if(server) {
#ifdef USE_OPENSSL
        handle = SSL_new (ctx);
        if(!handle) {
            set_error("SSL_new() failed");
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            handle=NULL;
            return;
        }

        //SSL_set_mode(handle,SSL_MODE_AUTO_RETRY);

        rbio=BIO_new(BIO_s_mem());
        wbio=BIO_new(BIO_s_mem());
        if(!rbio || !wbio) {
            set_error("BIO_new(BIO_s_mem()) failed");
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }

        int err;

        if ((err=SSL_use_certificate_file(handle, cert.c_str(), SSL_FILETYPE_PEM)) <= 0) {
            string err_str="SSL_use_certificate_file(): ";
            err_str+=get_openssl_error(err);
            set_error(err_str);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }
        if ((err=SSL_use_RSAPrivateKey_file(handle, key.c_str(), SSL_FILETYPE_PEM)) <= 0) {
            string err_str="SSL_use_RSAPrivateKey_file(): ";
            err_str+=get_openssl_error(err);
            set_error(err_str);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }

        if (!(err=SSL_check_private_key(handle))) {
            string err_str="SSL_check_private_key(): Private key does not match the certificate public key; ";
            err_str+=get_openssl_error(err);
            set_error(err_str);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }

        int mydata_index = SSL_get_ex_new_index(0,NULL , NULL, NULL, NULL);
        ssl_meta_data* mydata=new ssl_meta_data;
        extra_data.insert(make_pair(handle,make_pair(mydata_index,mydata)));

        SSL_set_ex_data(handle, mydata_index, mydata);

        /*
        SSL_set_fd (handle, (int)s->handle);
        */
        SSL_set_bio(handle,rbio,wbio);

        err=SSL_accept(handle);

        if(err<0)
            state=want_accept;
        else if(1==err) {
            state=verify_cert;
        } else {
            // err == 0 || err > 1
            set_error(get_openssl_error(SSL_get_error(handle,err)));
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }

        return;
#else
        /* X509 stuff */
        int err=gnutls_certificate_allocate_credentials (&(s->xcred));
        if(err<0) {
            set_error("gnutls_certificate_allocate_credentials() failed");
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            return;
        }

        /* sets the trusted cas file
        */
        err=gnutls_certificate_set_x509_trust_file (s->xcred,CA_file.c_str(), GNUTLS_X509_FMT_PEM);
        if(err<0) {
            set_error("gnutls_certificate_set_x509_trust_file() failed");
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            gnutls_certificate_free_credentials (s->xcred);
            return;
        }

        /*
        gnutls_certificate_set_x509_crl_file (x509_cred, CRLFILE,
                        GNUTLS_X509_FMT_PEM);
        */

        err=gnutls_certificate_set_x509_key_file (s->xcred, cert.c_str(), key.c_str(),
                        GNUTLS_X509_FMT_PEM);
        if(err<0) {
            set_error("gnutls_certificate_set_x509_key_file() failed");
            gnutls_deinit(handle);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            gnutls_certificate_free_credentials (s->xcred);
            destroy();
            handle=NULL;
            return;
        }

        gnutls_certificate_set_dh_params (s->xcred, dh_params);
        gnutls_certificate_set_rsa_export_params (s->xcred, rsa_params);

        gnutls_session session;

        gnutls_init(&session, GNUTLS_SERVER);

        const int cert_type_priority[10] = { GNUTLS_CRT_X509, 0 };
        const int cipher_priority[10] = {GNUTLS_CIPHER_AES_256_CBC, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_CIPHER_3DES_CBC, 0};
        const int protocol_priority[10] = {GNUTLS_SSL3, GNUTLS_TLS1_1, GNUTLS_TLS1_0, 0};
        const int kx_priority[10] = {GNUTLS_KX_RSA, 0};
        const int mac_priority[10] = {GNUTLS_MAC_SHA, GNUTLS_MAC_RMD160, GNUTLS_MAC_MD5, 0};
        const int compression_priority[10] = {GNUTLS_COMP_NULL, 0};

        /* Use the default priorities, plus, export cipher suites.
        */
        //gnutls_set_default_export_priority (session);
        gnutls_set_default_priority (session);

        /*
        gnutls_protocol_set_priority(ssl,protocol_priority);
        gnutls_kx_set_priority(ssl,kx_priority);
        gnutls_mac_set_priority(ssl,mac_priority);
        gnutls_cipher_set_priority(ssl,cipher_priority);
        gnutls_certificate_type_set_priority (ssl, cert_type_priority);
        gnutls_compression_set_priority(ssl,compression_priority);
        */

        gnutls_credentials_set (session, GNUTLS_CRD_CERTIFICATE, s->xcred);

        /* request client certificate if any.
        */
        gnutls_certificate_server_set_request (session, GNUTLS_CERT_REQUEST);

        gnutls_dh_set_prime_bits (session, DH_BITS);

        //gnutls_transport_set_ptr(session,(gnutls_transport_ptr)client->handle);
        gnutls_transport_set_pull_function(session,pull_func);
        gnutls_transport_set_push_function(session,push_func);
        gnutls_transport_set_ptr(session,(gnutls_transport_ptr)this);

        s->got_xcred=true;

        /*
        int ret=gnutls_handshake(session);

        gnutls_certificate_free_credentials(x509_cred);

        if(ret<0) {
            gnutls_deinit(session);
            session=NULL;
            string log="Handshake failed: ";
            log+=gnutls_strerror(ret);
            log_ssl(log);
            return true;
        }
        */

        state=want_handshake;

        handle=session;
        
        return;
#endif
    } else {
#ifdef USE_OPENSSL
        handle = SSL_new (ctx);
        if(!handle) {
            set_error("SSL_new() failed");
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            handle=NULL;
            return;
        }

        //SSL_set_mode(handle,SSL_MODE_AUTO_RETRY);

        rbio=BIO_new(BIO_s_mem());
        wbio=BIO_new(BIO_s_mem());
        if(!rbio || !wbio) {
            set_error("BIO_new(BIO_s_mem()) failed");
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }

        SSL_set_verify(handle, SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE|SSL_VERIFY_FAIL_IF_NO_PEER_CERT,my_verify_callback);
        SSL_set_verify_depth(handle,verify_depth_ + 1);

        int err;

        if ((err=SSL_use_certificate_file(handle, cert.c_str(), SSL_FILETYPE_PEM)) <= 0) {
            string err_str="SSL_use_certificate_file(): ";
            err_str+=get_openssl_error(err);
            set_error(err_str);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }
        if ((err=SSL_use_RSAPrivateKey_file(handle, key.c_str(), SSL_FILETYPE_PEM)) <= 0) {
            string err_str="SSL_use_RSAPrivateKey_file(): ";
            err_str+=get_openssl_error(err);
            set_error(err_str);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }

        if (!(err=SSL_check_private_key(handle))) {
            string err_str="SSL_check_private_key(): Private key does not match the certificate public key; ";
            err_str+=get_openssl_error(err);
            set_error(err_str);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }

        int mydata_index = SSL_get_ex_new_index(0,NULL , NULL, NULL, NULL);
        ssl_meta_data* mydata=new ssl_meta_data;
        extra_data.insert(make_pair(handle,make_pair(mydata_index,mydata)));

        SSL_set_ex_data(handle, mydata_index, mydata);

        /*
        SSL_set_fd (handle, (int)s->handle);
        */
        SSL_set_bio(handle,rbio,wbio);
        
        err=SSL_connect(handle);

        if(err<0)
            state=want_handshake;
        else if(1==err) {
            state=verify_cert;
        } else {
            // err == 0 || err > 1
            set_error(get_openssl_error(SSL_get_error(handle,err)));
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }

        return;
#else
        gnutls_certificate_credentials xcred;

        /* X509 stuff */
        int err=gnutls_certificate_allocate_credentials (&xcred);
        if(err<0) {
            set_error("gnutls_certificate_allocate_credentials() failed");
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            return;
        }

        /* sets the trusted cas file
        */
        err=gnutls_certificate_set_x509_trust_file (xcred,CA_file.c_str(), GNUTLS_X509_FMT_PEM);
        if(err<0) {
            set_error("gnutls_certificate_set_x509_trust_file() failed");
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            gnutls_certificate_free_credentials (xcred);
            return;
        }

        destroy();
        handle=NULL;

        /* Initialize TLS session 
        */
        gnutls_init (&handle, GNUTLS_CLIENT);

        err=gnutls_certificate_set_x509_key_file(xcred,cert.c_str(), key.c_str(), GNUTLS_X509_FMT_PEM);
        if(err<0) {
            set_error("gnutls_certificate_set_x509_key_file() failed");
            gnutls_deinit(handle);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            gnutls_certificate_free_credentials (xcred);
            destroy();
            handle=NULL;
            return;
        }

        gnutls_certificate_set_dh_params (xcred, dh_params);
        gnutls_certificate_set_rsa_export_params (xcred, rsa_params);

        /* put the x509 credentials to the current session
        */
        gnutls_credentials_set (handle, GNUTLS_CRD_CERTIFICATE, xcred);

        const int cert_type_priority[10] = { GNUTLS_CRT_X509, 0 };
        const int cipher_priority[10] = {GNUTLS_CIPHER_AES_256_CBC, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_CIPHER_3DES_CBC, 0};
        const int protocol_priority[10] = {GNUTLS_SSL3, GNUTLS_TLS1_1, GNUTLS_TLS1_0, 0};
        const int kx_priority[10] = {GNUTLS_KX_RSA, 0};
        const int mac_priority[10] = {GNUTLS_MAC_SHA, GNUTLS_MAC_RMD160, GNUTLS_MAC_MD5, 0};
        const int compression_priority[10] = {GNUTLS_COMP_NULL, 0};

        /* Use default priorities */
        //gnutls_set_default_priority (ssl);

        gnutls_protocol_set_priority(handle,protocol_priority);
        gnutls_kx_set_priority(handle,kx_priority);
        gnutls_mac_set_priority(handle,mac_priority);
        gnutls_cipher_set_priority(handle,cipher_priority);
        gnutls_certificate_type_set_priority (handle, cert_type_priority);
        gnutls_compression_set_priority(handle,compression_priority);

        /* connect to the peer
        */
        //gnutls_transport_set_ptr(ssl,(gnutls_transport_ptr)server->handle);

        gnutls_transport_set_pull_function(handle,pull_func);
        gnutls_transport_set_push_function(handle,push_func);
        gnutls_transport_set_ptr(handle,(gnutls_transport_ptr)this);

        socket.got_xcred=true;
        socket.xcred=xcred;

        state=want_handshake;

        return;
#endif
    }

    // NOT REACH
}

size_t c_ssl::read(void* buff, size_t len)
{
    if(state!=disconnected) { // to allow read something from the read queue, even if disconnected
        if(state!=session_ok)
            return 0;
    }
    if(to_read.empty())
        return 0;

    deque<s_ssl_data>::iterator i1=to_read.begin();
    if((*i1).len<=len) {
        memcpy(buff,(*i1).data,len=(*i1).len);
        free((*i1).data);
        to_read.pop_front();
        return len;
    }
    // resize
    s_ssl_data n;
    n.len=(*i1).len-len;
    n.data=malloc(n.len);
    if(!n.data)
        return 0;
    memcpy(n.data,&(((char*)(*i1).data)[len]),n.len);
    memcpy(buff,(*i1).data,len);
    free((*i1).data);
    to_read.pop_front();
    to_read.push_front(n);
    return len;
}

void c_ssl::write(const void* buff, size_t len)
{
    if(!len)
        return;
    s_ssl_data n;
    n.len=len;

    char buf[256];

    if(to_write.size()>=2) {
        deque<s_ssl_data>::reverse_iterator i1=to_write.rbegin();
        if((*i1).len+len<1024) {
            void* b=realloc((*i1).data,(*i1).len+len);
            if(b) {
                (*i1).data=b;
                memcpy(&((char*)b)[(*i1).len],buff,len);
                (*i1).len+=len;
                return;
            }
        }
    }
    
    n.data=malloc(n.len);
    if(!n.data) {
        sprintf(buf,"malloc(%u) failed, out of memory",n.len);
        set_error(buf);
#ifdef USE_OPENSSL
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }

        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
        handle=NULL;
        state=disconnected;
        return;
#else
        //gnutls_deinit(handle);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        state=disconnected;
        return;
#endif
        // NOT REACH
    }

    memcpy(n.data,buff,n.len);

    to_write.push_back(n);
}

bool c_ssl::get_error(string& str)
{
    if(!to_read.empty()) {
        // to allow read something from the read queue, even if disconnected
        str="";
        return false;
    }
    if(error)
        str=err_str;
    else
        str="";
    return error;
}

void c_ssl::set_expected_cert_file(const string& cert_file)
{
    expected_cert_file=cert_file;
    if(session_ok==state)
        state=verify_cert;
}

c_ssl::~c_ssl()
{
    clear_error_queue();

#ifdef USE_OPENSSL
    if(handle) {
        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
    }
    handle=NULL;
#else
    /*if(handle)
        gnutls_deinit(handle);*/
    destroy();
    handle=NULL;
#endif

    if(s) {
        sock_close(*s);
        delete s;
        s=NULL;
    }

    while(!to_read.empty()) {
        free((*(to_read.begin())).data);
        to_read.pop_front();
    }
    while(!to_write.empty()) {
        free((*(to_write.begin())).data);
        to_write.pop_front();
    }

    vector<c_ssl*>::iterator i1;
    for(i1=ssl_descriptors.begin(); i1!=ssl_descriptors.end(); i1++) {
        if(this==*i1) {
            ssl_descriptors.erase(i1);
            break;
        }
    }
}

void c_ssl::process()
{
    clear_error_queue();

    if(!handle) {
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        state=disconnected;
        return;
    }

    if(!s) {
#ifdef USE_OPENSSL
        rbio=NULL;
        wbio=NULL;

        destroy();

        if(handle)
            SSL_free(handle);
#else
        /*if(handle)
            gnutls_deinit(handle);*/
        destroy();
#endif
        handle=NULL;
        state=disconnected;
        return;
    }

#ifdef USE_OPENSSL
    if(rbio && wbio && s) {
        char buff[1024*16];

        while(!BIO_eof(wbio)) {
            int ret=BIO_read(wbio,buff,sizeof(buff));
            if(ret>0) {
                int ec=0;
                size_t sent=sock_send(*s,buff,ret,ec);
                if(ec || sent!=ret) {
                    rbio=NULL;
                    wbio=NULL;

                    destroy();

                    if(handle)
                        SSL_free(handle);
                    handle=NULL;
                    state=disconnected;
                    return;
                }
            } else {
                bool retry=BIO_should_retry(wbio);
                if(ret<0 && !retry) {
                    rbio=NULL;
                    wbio=NULL;

                    destroy();

                    if(handle)
                        SSL_free(handle);
                    handle=NULL;
                    state=disconnected;
                    return;
                }
                if(ret<0 && retry) {
                    // okay
                }
                break;
            }
        }

        for(;;) {
            int ec=0;
            bool closed=false;
            size_t len=sock_read(*s,buff,sizeof(buff),ec,closed);
            if(len && !ec && !closed) {
                int ret=BIO_write(rbio,buff,len);
                bool retry=BIO_should_retry(rbio);
                if(ret<0 && !retry) {
                    rbio=NULL;
                    wbio=NULL;

                    destroy();

                    if(handle)
                        SSL_free(handle);
                    handle=NULL;
                    state=disconnected;
                    return;
                }
                if(ret<0 && retry) {
                    sock_unread(*s,buff,len);
                    break;
                }
            } else {
                if(ec || closed) {
                    rbio=NULL;
                    wbio=NULL;

                    destroy();

                    if(handle)
                        SSL_free(handle);
                    handle=NULL;
                    state=disconnected;

                    string log="in file " __FILE__ " in function ";
                    log+=__func__;
                    if(ec) {
                        char* errstr=NULL;
                        //log_socket(ec,errstr=sock_error(ec),log,s);
                        set_error(errstr);
                        delete[] errstr;
                    } else {
                        // closed
                        //log_socket(0,"SSL socket EOF",log,s);
                        set_error("SSL socket EOF");
                    }
                }
                break;
            }
        }
    }
#endif

    switch(state) {
        case disconnected:
            return;
        case want_handshake:
            do_handshake();
            return;
        case want_accept:
            do_accept();
            return;
        case verify_cert:
            do_verify();
            return;
        case session_ok:
            do_transfer();
            return;
        case in_shutdown:
            {
                size_t in, out;
                get_cache_stats(in,out);
                if(in || out) {
                    do_transfer();
                    return;
                }
                do_shutdown();
                return;
            }
    }
    // huh?!
    if(s) {
        sock_close(*s);
        delete s;
        s=NULL;
    }
    return;
}

void c_ssl::do_handshake()
{
    clear_error_queue();

#ifdef USE_OPENSSL
    int err=SSL_connect(handle);

    unsigned long err2=SSL_get_error(handle,err);

    if(err<0 && err2!=1)
        state=want_handshake;
    else if(1==err) {
        state=verify_cert;
    } else {
        // err == 0 || err > 1 || fatal || (err==-1 && err2==1)
        set_error(get_openssl_error(err2));
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }

        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
        handle=NULL;
        state=disconnected;
        return;
    }
    return;
#else
    int err=gnutls_handshake(handle);
    if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
        return;
    }
    if(err!=0 && gnutls_error_is_fatal(err)) {
        set_error(my_ERR_error_string(err));
        //gnutls_deinit(handle);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        state=disconnected;
        return;
    }
    if(0==err)
        state=verify_cert;
    return;
#endif
    // NOT REACH
}

void c_ssl::do_accept()
{
    clear_error_queue();

#ifdef USE_OPENSSL
    int err=SSL_accept(handle);

    unsigned long err2=SSL_get_error(handle,err);

    if(err<0)
        state=want_accept;
    else if(1==err) {
        state=verify_cert;
    } else {
        // err == 0 || err > 1
        set_error(get_openssl_error(err2));
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }

        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
        handle=NULL;
        state=disconnected;
        return;
    }
    return;
#else
    int err=gnutls_handshake(handle);
    if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
        return;
    }
    if(err!=0 && gnutls_error_is_fatal(err)) {
        set_error(my_ERR_error_string(err));
        //gnutls_deinit(handle);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        state=disconnected;
        return;
    }
    if(0==err)
        state=verify_cert;
    return;
#endif
    // NOT REACH
}

void c_ssl::do_transfer()
{
    clear_error_queue();

    char buf[1024*10];
    size_t max_size=1024*10;
    int err;

    if(force_write)
        goto do_write;

    // read

#ifdef USE_OPENSSL
    err=(int)SSL_read(handle,buf,max_size);
    if(err>0) {
        s_ssl_data n;
        n.len=(size_t)err;
        n.data=malloc(n.len);
        if(!n.data) {
            sprintf(buf,"malloc(%u) failed, out of memory",n.len);
            set_error(buf);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            state=disconnected;
            return;
        }
        memcpy(n.data,buf,n.len);
        to_read.push_back(n);
    }
    if(err<0) {
        // ok, no problem
    }
    if(0==err) {
        set_error(get_openssl_error(SSL_get_error(handle,err)));
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }

        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
        handle=NULL;
        state=disconnected;
        return;
    }
#else
    err=gnutls_record_recv(handle,buf,max_size);
    if(err<0 && gnutls_error_is_fatal(err)) {
        set_error(my_ERR_error_string(err));
        //gnutls_deinit(handle);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        state=disconnected;
        return;
    } else if(err<0) {
        if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
            // force_read
        }
        goto do_write;
    } else if(0==err) {
        set_error("EOF");
        //gnutls_deinit(handle);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        state=disconnected;
        return;
    }
    // err > 0
    {
        s_ssl_data n;
        n.len=(size_t)err;
        n.data=malloc(n.len);
        if(!n.data) {
            sprintf(buf,"malloc(%u) failed, out of memory",n.len);
            set_error(buf);
            //gnutls_deinit(handle);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            destroy();
            handle=NULL;
            state=disconnected;
            return;
        }
        memcpy(n.data,buf,n.len);
        to_read.push_back(n);
    }
#endif

do_write:
    // write
    clear_error_queue();
    force_write=false;
    if(to_write.empty())
        return;
    void* w=(*(to_write.begin())).data;
    size_t len=(*(to_write.begin())).len;
    if(!len) {
        to_write.pop_front();
        goto do_write;
    }

#ifdef USE_OPENSSL
    err=(int)SSL_write(handle,w,len);
    if(err>0) {
        if((signed)len==err) {
            free(w);
            to_write.pop_front();
            return;
        }
        s_ssl_data n;
        n.len=((signed)len)-err;
        n.data=malloc(n.len);
        if(!n.data) {
            sprintf(buf,"malloc(%u) failed, out of memory",n.len);
            set_error(buf);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            state=disconnected;
            return;
        }
        memcpy(n.data,&(((char*)w)[err]),n.len);
        free(w);
        to_write.pop_front();
        to_write.push_front(n);
        return;
    }
    if(err<0) {
        // okay, no problem
        force_write=true;
        return;
    }
    if(0==err) {
        set_error(get_openssl_error(SSL_get_error(handle,err)));
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }

        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
        handle=NULL;
        state=disconnected;
        return;
    }
    return;
#else
    err=gnutls_record_send(handle,w,len);
    if(err<0 && gnutls_error_is_fatal(err)) {
        set_error(my_ERR_error_string(err));
        //gnutls_deinit(handle);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        state=disconnected;
        return;
    } else if(err<0 || 0==err) {
        if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
            force_write=true;
        }
        // okay, no problem
        return;
    }
    // err > 0
    {
        if((signed)len==err) {
            free(w);
            to_write.pop_front();
            return;
        }
        s_ssl_data n;
        n.len=((signed)len)-err;
        n.data=malloc(n.len);
        if(!n.data) {
            sprintf(buf,"malloc(%u) failed, out of memory",n.len);
            set_error(buf);
            //gnutls_deinit(handle);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }
            destroy();
            handle=NULL;
            state=disconnected;
            return;
        }
        memcpy(n.data,&(((char*)w)[err]),n.len);
        free(w);
        to_write.pop_front();
        to_write.push_front(n);
    }
    return;
#endif

    // NOT REACH
}

void c_ssl::do_verify()
{
    clear_error_queue();

    if(!cert_ok_so_far) {
        set_error("Certificate verification failed");
#ifdef USE_OPENSSL
        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
#else
        //gnutls_deinit(handle);
#endif
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        return;
    }
#ifdef USE_OPENSSL
    bool ok=true;
    X509* client_cert = SSL_get_peer_certificate (handle);
    if (client_cert != NULL) {
        if (SSL_get_verify_result(handle) == X509_V_OK) {
            // The client sent a certificate which verified OK
            if(!expected_cert_file.empty()) {
                X509* ecc=ssl_read_cert(const_cast<char*>(expected_cert_file.c_str()));
                X509_check_purpose(ecc,-1,0); // compute SHA1 hash
                if(ecc==NULL) {
                    ok=false;
                } else {
                    char* hash1=(char*)&(ecc->sha1_hash);
                    char* hash2=(char*)&(client_cert->sha1_hash);
                    ok=true;
                    for(int i1=0; i1<sizeof(client_cert->sha1_hash); i1++, hash1++, hash2++)
                        if(*hash1!=*hash2) {
                            ok=false;
                            break;
                        }
                }
            } else
                ok=true;
            if(ok) {
                state=session_ok;
                X509_free (client_cert);
            } else {
                cert_ok_so_far=false;
                set_error("Certificate verification failed: SHA1 hash doesn't match");
                X509_free (client_cert);
                if(s) {
                    sock_close(*s);
                    delete s;
                    s=NULL;
                }

                rbio=NULL;
                wbio=NULL;

                destroy();

                SSL_free(handle);
                handle=NULL;
                return;
            }
        } else {
            cert_ok_so_far=false;
            set_error("Certificate verification failed");
            X509_free (client_cert);
            if(s) {
                sock_close(*s);
                delete s;
                s=NULL;
            }

            rbio=NULL;
            wbio=NULL;

            destroy();

            SSL_free(handle);
            handle=NULL;
            return;
        }
    } else {
        cert_ok_so_far=false;
        set_error("Certificate verification failed");
        X509_free (client_cert);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }

        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
        handle=NULL;
        return;
    }
    if(!cert_ok_so_far) {
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }

        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
        handle=NULL;
        state=disconnected;
        return;
    }
    state=session_ok;
    return;
#else
    string res=verify_certificate(handle,"",expected_cert_file.c_str());
    if(!res.empty()) {
        cert_ok_so_far=false;
        string err="Certificate verification failed: ";
        err+=res;
        set_error(err);
        //gnutls_deinit(handle);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        return;
    }
    if(!cert_ok_so_far) {
        //gnutls_deinit(handle);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        state=disconnected;
        return;
    }
    state=session_ok;
    return;
#endif

    // NOT_REACH
}

void ssl_loop()
{
    {
        vector<c_ssl*>::iterator i1;
        for(i1=ssl_descriptors.begin(); i1!=ssl_descriptors.end(); i1++) {
            (*i1)->process();
        }
    }
    {
        time_t now=time(NULL);
        vector<pair<c_ssl*,time_t> >::iterator i1;
again_a:
        for(i1=ssl_garbage.begin(); i1!=ssl_garbage.end(); i1++) {
            if((*i1).second+60<now) {
                delete (*i1).first;
                ssl_garbage.erase(i1);
                goto again_a;
            }
        }
    }

    static time_t last_dh_rsa_gen=0;
    if(0==last_dh_rsa_gen) {
#ifndef USE_OPENSSL
        string log="Generating DH and RSA parameters... (";
        char tmp[64];
        log+=ltoa((long)DH_BITS,tmp,10);
        log+=" DH bits, ";
        log+=ltoa((long)RSA_BITS,tmp,10);
        log+=" RSA bits)";
        log_ssl(log,true);
        printf("%s\n",log.c_str());
        time_t start=time(NULL);

        gnutls_dh_params_init(&dh_params);
        gnutls_rsa_params_init(&rsa_params);
        /* Generate Diffie Hellman parameters - for use with DHE
         * kx algorithms. These should be discarded and regenerated
         * once a day, once a week or once a month. Depends on the
         * security requirements.
         */
        gnutls_dh_params_generate2(dh_params, DH_BITS);
        /* Generate RSA parameters - for use with RSA-export
         * cipher suites. These should be discarded and regenerated
         * once a day, once every 500 transactions etc. Depends on the
         * security requirements.
         */
        gnutls_rsa_params_generate2(rsa_params, RSA_BITS);
        time_t stop=time(NULL);
        log="Generation of DH and RSA parameters took ";
        log+=ltoa((long)stop-start,tmp,10);
        log+=" seconds";
        log_ssl(log);
        printf("%s\n",log.c_str());
#endif
        num_transactions=0;
        last_dh_rsa_gen=time(NULL);
    }
    static time_t thread_start=0;
#ifndef USE_OPENSSL
    if(last_dh_rsa_gen+DH_RSA_REGENERATION_NUM_SECS<time(NULL) || num_transactions>=DH_RSA_REGENERATION_NUM_TRANSACTIONS) {
        num_transactions=0;
        last_dh_rsa_gen=time(NULL);

        if(!cs_init) {
            thread_start=time(NULL);
            string log="Regenerating DH and RSA parameters... (";
            char tmp[64];
            log+=ltoa((long)DH_BITS,tmp,10);
            log+=" DH bits, ";
            log+=ltoa((long)RSA_BITS,tmp,10);
            log+=" RSA bits)";
            log_ssl(log);
#if defined(_WIN32)
            try {
                InitializeCriticalSection(&cs);
                cs_init=true;
            } catch(...) {
                cs_init=false;
                ssl_dh_thread_running=false;
            }
            if(cs_init) {
                ssl_dh_thread_running=true;
                uintptr_t res=_beginthread(ssl_dh_thread,0,(void*)NULL);
                if(-1L==res) {
                    cs_init=false;
                    ssl_dh_thread_running=false;
                }
            }
#else
#   ifdef HAVE_LIBPTHREAD
            if(pthread_mutex_init(&mutex,NULL)) {
                cs_init=false;
                ssl_dh_thread_running=false;
            } else {
                cs_init=true;
                pthread_t thread;
                int rc=pthread_create(&thread,NULL,ssl_dh_thread,NULL);
                if(rc) {
                    cs_init=false;
                    ssl_dh_thread_running=false;
                } else
                    ssl_dh_thread_running=true;
            }
#   else
            ssl_dh_thread(NULL);
            ssl_dh_thread_running=false;
            cs_init=true;
#endif
#endif
        }
    }
    if(cs_init) {
#if defined(_WIN32)
again:
        try {
            EnterCriticalSection(&cs);
        } catch(...) {
            SleepEx(10,true);
            goto again;
        }
#else
#   ifdef HAVE_LIBPTHREAD
        pthread_mutex_lock(&mutex);
#   endif
#endif
        bool thread_running=ssl_dh_thread_running;
#if defined(_WIN32)
        LeaveCriticalSection(&cs);
#else
#   ifdef HAVE_LIBPTHREAD
        pthread_mutex_unlock(&mutex);
#   endif
#endif
        if(!thread_running) {
            time_t stop=time(NULL);
            string log="Regeneration of DH and RSA parameters took ";
            char tmp[64];
            log+=ltoa((long)stop-thread_start,tmp,10);
            log+=" seconds";
            log_ssl(log);

#           if defined(_WIN32)
            DeleteCriticalSection(&cs);
#           else
#               ifdef HAVE_LIBPTHREAD
            pthread_mutex_destroy(&mutex);
#               endif
#           endif

            cs_init=false;

            gnutls_dh_params_deinit(dh_params);
            gnutls_rsa_params_deinit(rsa_params);
            dh_params=thread_dh_params;
            rsa_params=thread_rsa_params;

#ifndef XTELNET
            extern time_t irc_last_keepalive_detection;
            extern bool irc_keepalive_sent;
            irc_last_keepalive_detection=time(NULL)-IRC_KEEPALIVE_PERIOD;    // do not check for timeout now - this operation could have been slow
            irc_keepalive_sent=false;
#endif
        }
    }
#endif

#ifdef USE_OPENSSL
    static time_t last_rnd_save=0;
    if(last_rnd_save+60<time(NULL)) {
        last_rnd_save=time(NULL);
        RAND_write_file("openssl.rnd");
    }
#endif
}

string c_ssl::get_openssl_error()
{
#ifdef USE_OPENSSL
    char buff[1024];
    string result;
    unsigned long err;
    while((err=ERR_get_error())) {
        if(!result.empty())
            result+="; ";
        ERR_error_string(err,buff);
        result+=buff;
        if(ERR_lib_error_string(err)) {
            result+=", ";
            result+=ERR_lib_error_string(err);
        }
        if(ERR_func_error_string(err)) {
            result+=", ";
            result+=ERR_func_error_string(err);
        }
        if(ERR_reason_error_string(err)) {
            result+=", ";
            result+=ERR_reason_error_string(err);
        }
    }
    return result;
#else
    return "";
#endif
}

string c_ssl::get_openssl_error(unsigned long err)
{
#ifdef USE_OPENSSL
    char buff[1024];
    string result=get_openssl_error();
    unsigned long e=err;
    if(e) {
        if(!result.empty())
            result+="; ";
        ERR_error_string(e,buff);
        result+=buff;
        if(ERR_lib_error_string(e)) {
            result+=", ";
            result+=ERR_lib_error_string(e);
        }
        if(ERR_func_error_string(e)) {
            result+=", ";
            result+=ERR_func_error_string(e);
        }
        if(ERR_reason_error_string(e)) {
            result+=", ";
            result+=ERR_reason_error_string(e);
        }
    }
    return result;
#else
    return "";
#endif
}

void c_ssl::clear_error_queue()
{
#ifdef USE_OPENSSL
    while(ERR_get_error())
        ;
#endif
}

void c_ssl::shutdown()
{
    if(disconnected==state)
        return;
    state=in_shutdown;
}

void c_ssl::do_shutdown()
{
    clear_error_queue();

#ifdef USE_OPENSSL
    int err=SSL_shutdown(handle);

    SSL_get_error(handle,err); // don't care about error returned

    if(1==err) {
        // ok
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }

        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
        handle=NULL;
        state=disconnected;
        return;
    } else if(-1==err) {
        // error
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }

        rbio=NULL;
        wbio=NULL;

        destroy();

        SSL_free(handle);
        handle=NULL;
        state=disconnected;
        return;
    } else {
        // err == 0
        // no problem, we will call it again
    }
    return;
#else
    int err=gnutls_bye(handle,GNUTLS_SHUT_RDWR);
    if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
        return;
    }
    {
        //gnutls_deinit(handle);
        if(s) {
            sock_close(*s);
            delete s;
            s=NULL;
        }
        destroy();
        handle=NULL;
        state=disconnected;
        return;
    }
    return;
#endif
    // NOT REACH
}

bool c_ssl::is_open()
{
    if(!to_read.empty()) {
        // to allow read something from the read queue, even if disconnected
        return true;
    }
    return state!=disconnected;
}

/*!
    \brief Closes all descriptors
    \version 1
    \date 2006-07-16
    \author VooDooMan
*/
void ssl_done()
{
again:
    vector<c_ssl*>::iterator i1=ssl_descriptors.begin();
    if(i1!=ssl_descriptors.end()) {
        delete *i1;
        goto again;
    }
    ssl_garbage.clear();
}

/*!
    \brief Garbage collector
    \version 1
    \date 2006-07-17
    \author VooDooMan
    \param ssl Descriptor
*/
void ssl_put_to_garbage(c_ssl* ssl)
{
    if(!ssl)
        return;
    ssl->shutdown();
    ssl_garbage.push_back(make_pair(ssl,time(NULL)));
}

void c_ssl::destroy()
{
#ifdef USE_OPENSSL
    if(handle) {
        map<SSL*,pair<int,ssl_meta_data*> >::iterator i1=extra_data.find(handle);
        if(extra_data.end()!=i1) {
            delete (*i1).second.second;
            extra_data.erase(i1);
        }
    }
#endif
}

#if defined(_WIN32) || defined(_WIN64)
#ifndef USE_OPENSSL
static struct gcry_thread_cbs gcry_threads_other;   //!< Descriptor for thread locking

/*!
    \brief Initialization of thread library
    \author VooDooMan
    \version 1
*/
static int thread_init()
{
    return 0;
}

/*!
    \brief Initialization of mutex
    \author VooDooMan
    \version 1
*/
static int thread_mutex_init(void **priv)
{
    static HANDLE h=NULL;
    if(NULL==h) {
        h=CreateMutex(NULL,FALSE,NULL);
    }
    *priv=(void*)h;
    return 0;
}

/*!
    \brief Closing of mutex
    \author VooDooMan
    \version 1
*/
static int thread_mutex_destroy(void **priv)
{
    HANDLE h=(HANDLE)*priv;
    CloseHandle(h);
    return 0;
}

/*!
    \brief Obtaining a mutex exclusion
    \author VooDooMan
    \version 1
*/
static int thread_mutex_lock(void **priv)
{
    HANDLE h=(HANDLE)*priv;
    DWORD rc=WaitForSingleObject(h,INFINITE);
    return 0;
}

/*!
    \brief Releasing of mutex exclusion
    \author VooDooMan
    \version 1
*/
static int thread_mutex_unlock(void **priv)
{
    HANDLE h=(HANDLE)*priv;
    ReleaseMutex(h);
    return 0;
}

/*!
    \brief Introducing threads into gcrypto library
    \author VooDooMan
    \version 1
*/
void gnutls_init_msvc_thread()
{
    memset(&gcry_threads_other,0,sizeof(gcry_threads_other));
    gcry_threads_other.option=GCRY_THREAD_OPTION_USER;
    gcry_threads_other.init=thread_init;
    gcry_threads_other.mutex_init=thread_mutex_init;
    gcry_threads_other.mutex_destroy=thread_mutex_destroy;
    gcry_threads_other.mutex_lock=thread_mutex_lock;
    gcry_threads_other.mutex_unlock=thread_mutex_unlock;
    gcry_control (GCRYCTL_SET_THREAD_CBS, &gcry_threads_other);
}
#endif
#endif

