/***************************************************************************
                          xtelnet.cpp  -  description
                             -------------------
    begin                : 2004
    copyright            : (C) 2004-2009  Marian VooDooMan Meravy
    email                : ghostvoodooman@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************

XTelnet - a simple telnet client
Copyright (C) 2004-2009  Marian VooDooMan Meravy
  (ghostvoodooman@users.sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#if !defined(VERSION)
#   define VERSION  "0.4.4"
#endif

#ifdef _WIN32
#   include <winsock2.h>
#   include <Ws2tcpip.h>
#   include <windows.h>
#endif

// For compilers that support precompilation, includes "wx/wx.h".
//#include "wx/prec.h"

// the application icon (under Windows and OS/2 it is in resources)
#if defined(__WXGTK__) || defined(__WXMOTIF__) || defined(__WXMAC__) || defined(__WXMGL__) || defined(__WXX11__)
//#    include "mondrian.xpm"
#endif

#ifdef __BORLANDC__
#   pragma hdrstop
#endif

#include <string>
#include <fstream>
#include <stdio.h>
#include <time.h>
#include <list>

// for all others, include the necessary headers
#ifndef WX_PRECOMP
#   include <wx/app.h>
#   include <wx/log.h>
#   include <wx/frame.h>
#   include <wx/panel.h>
#   include <wx/stattext.h>
#   include <wx/menu.h>
#   include <wx/layout.h>
#   include <wx/msgdlg.h>
#   include <wx/checkbox.h>
#   include <wx/listbox.h>
#   include <wx/socket.h>
#   include <wx/timer.h>
#endif

#include <wx/socket.h>
#include <wx/textctrl.h>
#include <wx/event.h>
#include <wx/busyinfo.h>
#include <wx/clipbrd.h>
#include <wx/dataobj.h>
#include <wx/checkbox.h>
#include <wx/listbox.h>

#ifdef _MSC_VER
#   include <crtdbg.h>
#   ifdef _DEBUG
#       define DEBUG_CLIENTBLOCK   new( _CLIENT_BLOCK, __FILE__, __LINE__)
#       define new DEBUG_CLIENTBLOCK
#       define SET_CRT_DEBUG_FIELD(a) \
            _CrtSetDbgFlag((a) | _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG))
#       define CLEAR_CRT_DEBUG_FIELD(a) \
            _CrtSetDbgFlag(~(a) & _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG))
#   else
#       define DEBUG_CLIENTBLOCK
#       define SET_CRT_DEBUG_FIELD(a) ((void) 0)
#       define CLEAR_CRT_DEBUG_FIELD(a) ((void) 0)
#   endif // _DEBUG
#endif

#include "compat.h"
#include "cnf.h"
#include "sock_shared.h"
#include "sock.h"
#include "ssl.h"
#include "match.h"
#include "thread.h"

using namespace std;

///////////////

enum
{
    XTelnet_Tools_About = 100,  //!< Menu item: File/About
    XTelnet_Tools_Quit,         //!< Menu item: File/Quit
    XTelnet_Tools_Disconnect,
    XTelnet_Tools_Paste,
    XTelnet_Tools_Paste_Record,
    XTelnet_Tools_RecordLog,
    XTelnet_Tools_ClearBuff,
    XTelnet_Tools_Hide,

    XTelnet_Selected,

    XTelnet_Up,
    XTelnet_Down,
    XTelnet_F7,
    XTelnet_Esc,

    XTelnet_Sockek_Evt,

    // controls
    XTelnet_Enter_Ctrl,
    XTelnet_Pasw_Ctrl,
    XTelnet_Iterator_Ctrl,

    XTelnet_Timer,
    XTelnet_Loop,
};

///////////////

sClient client;
c_thread* t;
wxMutex mutex;

class MyFrame;

class MyApp : public wxApp
{
public:
    // override base class virtuals
    // ----------------------------

    // this one is called on application startup and is a good place for the app
    // initialization (doing it here and not in the ctor allows to have an error
    // return: if OnInit() returns false, the application terminates)
    /*! Event for initializing the application
    */
    virtual bool OnInit();

    /*! Event for save configuration file(s)
    */
    //virtual int OnExit();

    /*! Main frame object
    */
    MyFrame *frame;
};

class MyPanel : public wxPanel
{
public:
    MyPanel(wxFrame *frame);

    wxTextCtrl* m_log;
    wxTextCtrl* m_enter, *m_enter2;
    wxCheckBox* m_pasw;

    wxListBox* m_iterator;
    bool m_iterator_showing;

    wxTimer* m_timer;
    wxTimer* m_loop;

    void OnTimer(wxTimerEvent& event);
    void OnLoop(wxTimerEvent& event);

private:

    DECLARE_EVENT_TABLE()
};

class cMyEvtHandler : public wxEvtHandler
{
    virtual bool ProcessEvent(wxEvent& event);
};

class MyFrame : public wxFrame
{
public:
    // ctor(s)
    MyFrame(const wxString& title, const wxPoint& pos, const wxSize& size);

    // event handlers (these functions should _not_ be virtual)
    /*! This event is fired when user clicks on Exit in the menu bar
    */
    void OnQuit(wxCommandEvent& event);

    void OnClose(wxCloseEvent& event);

    /*! This event is fired when user clicks on About in the menu bar
    */
    void OnAbout(wxCommandEvent& event);

    void OnEnter(wxCommandEvent& event);

    void OnDisconnect(wxCommandEvent& event);

    void OnPaste(wxCommandEvent& event);
    void OnPasteRecord(wxCommandEvent& event);

    void OnPassword(wxCommandEvent& event);

    void OnHide(wxCommandEvent& event);

    void OnRecordLog(wxCommandEvent& event);
    void OnClearBuff(wxCommandEvent& event);

    void OnUp(wxCommandEvent& event);
    void OnDown(wxCommandEvent& event);
    void OnF7(wxCommandEvent& event);
    void OnEsc(wxCommandEvent& event);

    void OnChar(wxKeyEvent& event);

    void OnSelected(wxCommandEvent& event);

    void AddToLog(wxTextCtrl* log, bool data, const wxString& str);

    void OnIdle(wxIdleEvent& event);

public:
    MyPanel *m_panel;

    wxMenu *menuTools;
    wxMenuBar *menuBar;

    list<wxString> lines;
    int lines_i;
    bool lines_iterating;

    int conn(sClient* client);
    void loop(sClient* client);

    void GetSockError(wxSocketBase* socket, wxString& msg);

    void ShowIterator(bool show);

    bool IteratorShown();

protected:
    bool bIteratorShown;

private:
    void Send(sClient* client, wxTextCtrl* enter, wxTextCtrl* log, bool Password);

    // any class wishing to process wxWindows events must use this macro
    DECLARE_EVENT_TABLE()
};

MyFrame *main_frame;

const char* get_eol()
{
    const char* eol;
    switch(client.Conf.iEOL) {
        case 2:
            eol="\n";
            break;
        case 3:
            eol="\r";
            break;
        case 4:
            eol="\n\r";
            break;
        case 1:
        default:
            eol="\r\n";
    }
    return eol;
}

bool in_ssl_init=false;

#include "event_table.h"

void log_ssl(const string& log, bool WXUNUSED(anyway) = false)
{
    main_frame->AddToLog(main_frame->m_panel->m_log,false,wxString(log.c_str(),wxConvLocal));
    main_frame->AddToLog(main_frame->m_panel->m_log,false,_T("\n"));
    if(in_ssl_init)
        wxGetApp().Yield();
}

int MyFrame::conn(sClient* client)
{
    // mutex is already locked
    if(client->sc || client->sc_ssl) {
        delete client->sc_ssl;
        client->sc_ssl=NULL;
        if(client->sc)
            client->sc->clear();
        delete client->sc;
        client->sc=NULL;
    }

    int errnum=0;
    c_sock_log::e_type errtype;
    in_addr a=sock_resolve(client->host.c_str(),&errnum,&errtype);
    if(errnum)
        return 1;

    char ip[64];
    inet_ntop(AF_INET,(void*)&a,ip,sizeof(ip)-1);

    bool async=client->Conf.bSSL?false:true;
    async=true;

    {
        errnum=0;
        client->sc=new s_socket(sock_connect("0.0.0.0",ip,client->port,errnum,async,""));

        if(errnum || !client->sc->cmp()) {
            delete client->sc;
            client->sc=NULL;
            return 1;
        }

        if(client->Conf.bSSL) {
            client->sc_ssl=new c_ssl(*client->sc,NULL,client->Conf.client_cert,client->Conf.client_key);
        }
    }

    return 0;
}

void MyFrame::AddToLog(wxTextCtrl* log, bool data, const wxString& str)
{
    if(data && client.bHideOutput)
        return;

    bool was_iterator_shown=IteratorShown();
    if(was_iterator_shown) {
        ShowIterator(false);
        m_panel->m_iterator->Refresh();
    }

    m_panel->m_log->Refresh();

    if(data) {
        log->SetDefaultStyle(wxTextAttr(*wxBLACK,wxNullColour,wxFont(12,wxMODERN,wxNORMAL,wxNORMAL)));
    } else {
        log->SetDefaultStyle(wxTextAttr(*wxRED,wxNullColour,wxFont(12,wxDEFAULT,wxNORMAL,wxNORMAL)));
    }
    if(!str.empty()) {
        wxString tmp;
        for(size_t i1=0; i1<str.Len(); i1++) {
            if(!data && str[i1]==wxChar('\n')) {
                tmp << _T("\n");
                continue;
            }
            if(str[i1]==wxChar('\r') || str[i1]==wxChar('\n')) {
                {
                    switch(client.Conf.iEOL) {
                        case 1:
                            if(i1+1<str.Len() && str[i1]==wxChar('\r') && str[i1+1]==wxChar('\n')) {
                                i1++;
                                tmp << _T("\n");
                                continue;
                            }
                            break;
                        case 2:
                            if(str[i1]==wxChar('\n')) {
                                tmp << _T("\n");
                                continue;
                            }
                            break;
                        case 3:
                            if(str[i1]==wxChar('\r')) {
                                tmp << _T("\n");
                                continue;
                            }
                            break;
                        case 4:
                            if(i1+1<str.Len() && str[i1]==wxChar('\n') && str[i1+1]==wxChar('\r')) {
                                i1++;
                                tmp << _T("\n");
                                continue;
                            }
                            break;
                    }
                }
            }
            if(str[i1]<wxChar(char(32))) {
                if(!tmp.empty()) {
                    log->AppendText(tmp);
                }

                tmp.clear();

                switch(str[i1]) {
                    case wxChar('\n'):
                        tmp << _T("\\n");
                        break;
                    case wxChar('\r'):
                        tmp << _T("\\r");
                        break;
                    case wxChar('\f'):
                        tmp << _T("\\f");
                        break;
                    case wxChar('\t'):
                        tmp << _T("\\t");
                        break;
                    case wxChar('\b'):
                        tmp << _T("\\b");
                        break;
                    case wxChar('\x7'):
                        tmp << _T("\\a");
                        break;
                    case wxChar('\0'):
                        tmp << _T("\\0");
                        break;
                    default:
                        {
                            char tmp2[16];
                            sprintf(tmp2,"\\x%02X",(unsigned char)str[i1]);
                            tmp << wxString(tmp2,wxConvLocal);
                        }
                        break;
                }

                if(client.Conf.bHighlight && data) {
                    log->SetDefaultStyle(wxTextAttr(*wxRED,wxNullColour,wxFont(12,wxMODERN,wxNORMAL,wxNORMAL)));
                }
                
                log->AppendText(tmp);
                
                if(data) {
                    log->SetDefaultStyle(wxTextAttr(*wxBLACK,wxNullColour,wxFont(12,wxMODERN,wxNORMAL,wxNORMAL)));
                } else {
                    log->SetDefaultStyle(wxTextAttr(*wxRED,wxNullColour,wxFont(12,wxDEFAULT,wxNORMAL,wxNORMAL)));
                }

                tmp.clear();
            } else
                tmp << str[i1];
        }
        if(!tmp.empty())
            log->AppendText(tmp);
    }

    {
        int lines=log->GetNumberOfLines();
        long chars=0;
        for(int i1=lines; i1>0 && lines-i1<15; i1--)
            chars+=log->GetLineLength(i1)+1;
        log->Freeze();
        log->ShowPosition(log->GetLastPosition()-chars);
        log->ShowPosition(log->GetLastPosition());
        log->ScrollLines(lines+1);
        log->Thaw();
    }

    if(was_iterator_shown) {
        ShowIterator(true);
        m_panel->m_iterator->Refresh();
    }
}

void MyFrame::ShowIterator(bool show)
{
    if(!show) {
        m_panel->m_iterator->Hide();
    } else {
        m_panel->m_iterator->Show();
        m_panel->Freeze();
        m_panel->m_log->Refresh();
        m_panel->m_iterator->Refresh();
        m_panel->Thaw();
    }
    bIteratorShown=m_panel->m_iterator_showing=show;
}

void MyFrame::loop(sClient* client)
{
    if(client->sc && client->sc->cmp()) {
        size_t size=10240;
		size_t max_size=size;
        char* in=new char[size+1];
        for(;;) {
            int err_code=0;
            bool closed=false;
            if(!client->sc_ssl) {
                {
                    wxMutexLocker locker(mutex);
                    size=sock_read(*client->sc,in,max_size,err_code,closed);
                }
                if(err_code) {
                    string err_str="Error: ";
                    char* err=sock_error(err_code,c_sock_log::TYPE_ERRNO);
                    err_str+=err;
                    delete[] err;
                    err_str+="\n    Type HOST NAME or IP address and hit ENTER for connection\n";
                    AddToLog(m_panel->m_log,false,wxString(err_str.c_str(),wxConvLocal));

                    wxMutexLocker locker(mutex);
                    sock_close(*client->sc);
                    delete client->sc;
                    client->sc=NULL;

                    break;
                }
                if(closed) {
                    AddToLog(m_panel->m_log,false,_T("EOF.\n    Type HOST NAME or IP address and hit ENTER for connection\n"));

                    wxMutexLocker locker(mutex);
                    sock_close(*client->sc);
                    delete client->sc;
                    client->sc=NULL;

                    break;
                }
            } else {
                string err_str;
                wxMutexLocker locker(mutex);
                if(client->sc_ssl->get_error(err_str)) {
                    err_str=((string)"SSL error: ")+err_str;
                    err_str+="\n    Type HOST NAME or IP address and hit ENTER for connection\n";
                    AddToLog(m_panel->m_log,false,wxString(err_str.c_str(),wxConvLocal));

                    delete client->sc_ssl;
                    client->sc_ssl=NULL;
                    if(client->sc)
                        client->sc->clear();
                    delete client->sc;
                    client->sc=NULL;

                    break;
                }
                size=client->sc_ssl->read(in,max_size);
            }
            in[size]='\0';
            if(size) {
                AddToLog(m_panel->m_log,true,wxString(in,( client->Conf.bUTF8 ? wxConvUTF8 : wxConvLibc )));
            }
            if(client->log) {
                fwrite(in,1,size,client->log);
                fflush(client->log);
            }
            if(!size)
                break;
        }
        delete[] in;
    }
}

void LoadConf(sClient* client)
{
    client->Conf.bUTF8=false;
    client->Conf.bLocalEcho=true;
    client->Conf.bUseEscape=false;
    client->Conf.iEOL=1;
    client->Conf.bHighlight=true;
    client->Conf.bSSL=false;
    client->Conf.ca="";
    client->Conf.client_cert="";
    client->Conf.client_key="";

    string st;
    ifstream in("xtelnet.conf");
    while(getline(in,st)) {
        if(!st.compare("bUTF8=0"))
            client->Conf.bUTF8=false;
        if(!st.compare("bUTF8=1"))
            client->Conf.bUTF8=true;
        if(!st.compare("bLocalEcho=0"))
            client->Conf.bLocalEcho=false;
        if(!st.compare("bLocalEcho=1"))
            client->Conf.bLocalEcho=true;
        if(!st.compare("bUseEscape=0"))
            client->Conf.bUseEscape=false;
        if(!st.compare("bUseEscape=1"))
            client->Conf.bUseEscape=true;
        if(!st.compare("iEOL=1"))
            client->Conf.iEOL=1;
        if(!st.compare("iEOL=2"))
            client->Conf.iEOL=2;
        if(!st.compare("iEOL=3"))
            client->Conf.iEOL=3;
        if(!st.compare("iEOL=4"))
            client->Conf.iEOL=4;
        if(!st.compare("bHighlight=0"))
            client->Conf.bHighlight=false;
        if(!st.compare("bHighlight=1"))
            client->Conf.bHighlight=true;
        if(!st.compare("bSSL=0"))
            client->Conf.bSSL=false;
        if(!st.compare("bSSL=1"))
            client->Conf.bSSL=true;

        string key;
        string value;
        key=st;
        if(key.find("=")!=string::npos)
            key.erase(key.find("="),key.length()-key.find("="));
        value=st;
        if(value.find("=")!=string::npos)
            value.erase(0,value.find("=")+1);

        if(!key.compare("sCAs"))
            client->Conf.ca=value;
        if(!key.compare("sCrt"))
            client->Conf.client_cert=value;
        if(!key.compare("sKey"))
            client->Conf.client_key=value;
    }
}

bool SaveConf(sClient* client)
{
    FILE *f=fopen("xtelnet.conf","w");
    if(f) {
        fprintf(f,"# XTelnet configuration file version " VERSION "\n\n");
        fprintf(f,"%s%s\n","bUTF8=",(client->Conf.bUTF8?"1":"0"));
        fprintf(f,"%s%s\n","bLocalEcho=",(client->Conf.bLocalEcho?"1":"0"));
        fprintf(f,"%s%s\n","bUseEscape=",(client->Conf.bUseEscape?"1":"0"));
        fprintf(f,"%s%d\n","iEOL=",client->Conf.iEOL);
        fprintf(f,"%s%s\n","bHighlight=",(client->Conf.bHighlight?"1":"0"));
        fprintf(f,"%s%s\n","bSSL=",(client->Conf.bSSL?"1":"0"));
        fprintf(f,"%s%s\n","sCAs=",client->Conf.ca.c_str());
        fprintf(f,"%s%s\n","sCrt=",client->Conf.client_cert.c_str());
        fprintf(f,"%s%s\n","sKey=",client->Conf.client_key.c_str());
        fclose(f);
        return true;
    }
    return false;
}

c_cnf_frame* cnf_frame;
MyFrame* m_frame;

bool MyApp::OnInit()
{
    wxApp::OnInit();
    srand(time(NULL));

    LoadConf(&client);
    client.bEnteringPort=false;
    client.bHideOutput=false;

    wxString tmp=wxString::FromAscii("XTelnet ");
    tmp+=wxString::FromAscii(VERSION);
    tmp+=wxString::FromAscii(" by VooDooMan - setup");
    cnf_frame=new c_cnf_frame(tmp,wxPoint(-1, -1), wxSize(610, 450));

    cnf_frame->m_panel->m_utf8->SetValue(client.Conf.bUTF8);
    cnf_frame->m_panel->m_highlight->SetValue(client.Conf.bHighlight);
    cnf_frame->m_panel->m_local_echo->SetValue(client.Conf.bLocalEcho);
    cnf_frame->m_panel->m_escape_chars->SetValue(client.Conf.bUseEscape);
    switch(client.Conf.iEOL) {
        case 2:
            cnf_frame->m_panel->m_eol_n->SetValue(true);
            break;
        case 3:
            cnf_frame->m_panel->m_eol_r->SetValue(true);
            break;
        case 4:
            cnf_frame->m_panel->m_eol_nr->SetValue(true);
        case 1:
        default:
            cnf_frame->m_panel->m_eol_rn->SetValue(true);
            break;
    }
    cnf_frame->m_panel->m_ssl->SetValue(client.Conf.bSSL);
    cnf_frame->m_panel->m_cas->SetValue(wxString::FromAscii(client.Conf.ca.c_str()));
    cnf_frame->m_panel->m_crt->SetValue(wxString::FromAscii(client.Conf.client_cert.c_str()));
    cnf_frame->m_panel->m_key->SetValue(wxString::FromAscii(client.Conf.client_key.c_str()));

    wxCommandEvent e;
    cnf_frame->m_panel->OnSslCheckBox(e);

    cnf_frame->Show(true);

    return true;
}

void show_main()
{
    static bool shown=false;
    if(shown)
        return;
    shown=true;

    client.Conf.bUTF8=cnf_frame->m_panel->m_utf8->GetValue();
    client.Conf.bHighlight=cnf_frame->m_panel->m_highlight->GetValue();
    client.Conf.bLocalEcho=cnf_frame->m_panel->m_local_echo->GetValue();
    client.Conf.bSSL=cnf_frame->m_panel->m_ssl->GetValue();
    client.Conf.bUseEscape=cnf_frame->m_panel->m_escape_chars->GetValue();
    client.Conf.ca=cnf_frame->m_panel->m_cas->GetValue().ToAscii();
    client.Conf.client_cert=cnf_frame->m_panel->m_crt->GetValue().ToAscii();
    client.Conf.client_key=cnf_frame->m_panel->m_key->GetValue().ToAscii();
    if(cnf_frame->m_panel->m_eol_rn->GetValue())
        client.Conf.iEOL=1;
    else if(cnf_frame->m_panel->m_eol_n->GetValue())
        client.Conf.iEOL=2;
    else if(cnf_frame->m_panel->m_eol_r->GetValue())
        client.Conf.iEOL=3;
    else if(cnf_frame->m_panel->m_eol_nr->GetValue())
        client.Conf.iEOL=4;
    else
        client.Conf.iEOL=1;

    SaveConf(&client);

    wxString tmp=wxString::FromAscii("XTelnet ");
    tmp+=wxString::FromAscii(VERSION);
    tmp+=wxString::FromAscii(" by VooDooMan");
    m_frame=new MyFrame(tmp,wxPoint(-1, -1), wxSize(610, 450));
    m_frame->Show(true);

    cnf_frame->Show(false);

    if(client.Conf.bSSL) {
#ifndef USE_OPENSSL
        log_ssl("Please wait...");
        wxBusyInfo wait(_T("Please wait..."));
#endif

#ifdef _WIN32
        wxGetApp().Yield();
#endif
        string err_str;
        in_ssl_init=true;
        if(ssl_init(err_str,const_cast<char*>(client.Conf.ca.c_str()))) {
            if(client.Conf.bSSL)
                log_ssl(err_str);
        }
        in_ssl_init=false;
        ssl_loop();
        log_ssl("Ready.");
    }

    m_frame->AddToLog(m_frame->m_panel->m_log,false,_T("    Type HOST NAME or IP address and hit ENTER for connection\n"));

    m_frame->m_panel->m_loop=new wxTimer(m_frame->m_panel,XTelnet_Loop);
    m_frame->m_panel->m_loop->Start(200);
}

MyFrame::MyFrame(const wxString& title, const wxPoint& pos, const wxSize& size)
       : wxFrame((wxFrame *)NULL, -1, title, pos, size)
{
    main_frame=this;

    menuTools=NULL;
    menuBar=NULL;

    menuTools = new wxMenu;

    menuTools->Append(XTelnet_Tools_Disconnect,_T("Disconnect"),_T("Disconnects the current session"));
    menuTools->AppendSeparator();
    menuTools->Append(XTelnet_Tools_Paste,_T("Paste from clipboard"),_T("Paste the text from the clipcoard"));
    menuTools->Append(XTelnet_Tools_Paste_Record,_T("Paste from clipboard and start record log"),_T("Paste the text from the clipcoard and after that begin to record a log"));
    menuTools->AppendSeparator();
    menuTools->Append(XTelnet_Tools_RecordLog,_T("Record log"),_T("Records a log"));
    menuTools->Append(XTelnet_Tools_ClearBuff,_T("Clear buffer"),_T("Clears the buffer"));
    menuTools->AppendSeparator();
    menuTools->AppendCheckItem(XTelnet_Tools_Hide,_T("Hide output"),_T("Clears the buffer"));
    menuTools->AppendSeparator();
    menuTools->Append(XTelnet_Tools_About,_T("About..."),_T("Shows the about box"));
    menuTools->AppendSeparator();
    menuTools->Append(XTelnet_Tools_Quit,_T("Quit"),_T("Quits the XTelnet"));

    // now append the freshly created menu to the menu bar...
    menuBar = new wxMenuBar;
    menuBar->Append(menuTools,_T("Tools"));

    // ... and attach this menu bar to the frame
    SetMenuBar(menuBar);

    m_panel = new MyPanel(this);

    bIteratorShown=false;
}

bool MyFrame::IteratorShown()
{
    return bIteratorShown;
}

void MyFrame::OnAbout(wxCommandEvent& WXUNUSED(event))
{
    wxString tmp=wxString::FromAscii("XTelnet ");
    tmp+=wxString::FromAscii(VERSION);
    tmp+=wxString::FromAscii(" by VooDooMan");
    wxMessageBox(tmp);
}

void MyFrame::OnRecordLog(wxCommandEvent& WXUNUSED(event))
{
    unsigned int r=rand();

    char tmp[128];
    sprintf(tmp,"%2x",r);

    strcat(tmp,".txt");
    client.log=fopen(tmp,"wb");
    char tmp2[128];
    if(client.log) {
        strcpy(tmp2,"Logging to file ");
        strcat(tmp2,tmp);
        strcat(tmp2,"\n");
    } else {
        strcpy(tmp2,"Error opening log file.\n");
    }
    AddToLog(m_panel->m_log,false,wxString(tmp2,wxConvLocal));
}

void MyFrame::OnClearBuff(wxCommandEvent& WXUNUSED(event))
{
    m_panel->m_log->Clear();
}

void MyFrame::OnUp(wxCommandEvent& WXUNUSED(event))
{
    if(lines.begin()==lines.end()) {
        lines_i=0;
        return;
    }
    if(!lines_iterating) {
        lines_i=-1;
        lines_iterating=true;
    } else
        lines_i--;

    if(!m_panel->m_enter->GetValue().empty()) {
        list<wxString>::iterator i1=lines.begin();
        bool got=false;
        while(i1!=lines.end())
            if((*(i1++))==m_panel->m_enter->GetValue()) {
                got=true;
                break;
            }
        if(!got) {
            lines.push_back(m_panel->m_enter->GetValue());
            m_panel->m_iterator->Append(m_panel->m_enter->GetValue());
        }
    }

    int i1=0;
    list<wxString>::iterator i2;
    for(i2=lines.begin(); i2!=lines.end(); i2++, i1++)
        ;
    while(lines_i<0)
        lines_i+=i1;
    lines_i=lines_i%i1;
    i1=0;
    for(i2=lines.begin(); i2!=lines.end(); i2++, i1++) {
        if(i1==lines_i) {
            m_panel->m_enter->SetValue(*i2);
            break;
        }
    }

    m_panel->m_iterator->SetSelection(lines_i,true);
    ShowIterator(true);
}

void MyFrame::OnDown(wxCommandEvent& WXUNUSED(event))
{
    if(lines.begin()==lines.end()) {
        lines_i=0;
        return;
    }
    if(!lines_iterating) {
        lines_i=0;
        lines_iterating=true;
    } else
        lines_i++;

    if(strlen(m_panel->m_enter->GetValue().ToAscii())) {
        list<wxString>::iterator i1=lines.begin();
        bool got=false;
        while(i1!=lines.end())
            if((*(i1++))==m_panel->m_enter->GetValue()) {
                got=true;
                break;
            }
        if(!got) {
            lines.push_back(m_panel->m_enter->GetValue());
            m_panel->m_iterator->Append(m_panel->m_enter->GetValue());
        }
    }

    int i1=0;
    list<wxString>::iterator i2;
    for(i2=lines.begin(); i2!=lines.end(); i2++, i1++)
        ;
    while(lines_i<0)
        lines_i+=i1;
    lines_i=lines_i%i1;
    i1=0;
    for(i2=lines.begin(); i2!=lines.end(); i2++, i1++) {
        if(i1==lines_i) {
            m_panel->m_enter->SetValue(*i2);
            break;
        }
    }

    m_panel->m_iterator->SetSelection(lines_i,true);
    ShowIterator(true);
}

void MyFrame::OnF7(wxCommandEvent& WXUNUSED(event))
{
    if(lines.begin()==lines.end()) {
        lines_i=0;
        return;
    }
    if(!lines_iterating) {
        wxCommandEvent e;
        OnUp(e);
    }
}

void MyFrame::OnEsc(wxCommandEvent& WXUNUSED(event))
{
    if(lines_iterating) {
        ShowIterator(false);
        lines_iterating=false;
        m_panel->m_enter->SetValue(_T(""));
    }
}

void MyFrame::OnChar(wxKeyEvent& event)
{
    if(event.GetKeyCode()!=WXK_ESCAPE) {
        event.Skip();
        return;
    }
    if(lines_iterating) {
        ShowIterator(false);
        lines_iterating=false;
        m_panel->m_enter->SetValue(_T(""));
    }
}

void MyFrame::OnSelected(wxCommandEvent& WXUNUSED(event))
{
    if(lines.begin()==lines.end()) {
        lines_i=0;
        return;
    }
    lines_i=m_panel->m_iterator->GetSelection();
    m_panel->m_enter->SetValue(m_panel->m_iterator->GetString(lines_i));
}

void MyFrame::OnDisconnect(wxCommandEvent& WXUNUSED(event))
{
    wxMutexLocker locker(mutex);

    if(client.sc && client.sc->cmp()) {
        if(client.sc_ssl) {
            delete client.sc_ssl;
            client.sc_ssl=NULL;
            if(client.sc)
                client.sc->clear();
            delete client.sc;
            client.sc=NULL;
        }
        client.sc=NULL;
        AddToLog(m_panel->m_log,false,_T("Disconnect caused by you.\n\n    Type HOST NAME or IP address and hit ENTER for connection\n"));
        if(client.log)
            fclose(client.log);
        client.log=NULL;
    } else {
        AddToLog(m_panel->m_log,false,_T("Not connected.\n\n    Type HOST NAME or IP address and hit ENTER for connection\n"));
        if(client.log)
            fclose(client.log);
        client.log=NULL;
    }
}

void MyFrame::OnQuit(wxCommandEvent& WXUNUSED(event))
{
    Close(true);
}

void MyFrame::Send(sClient* client, wxTextCtrl* enter, wxTextCtrl* log, bool Password)
{
    wxMutexLocker locker(mutex);

    wxString s=enter->GetValue();
    if(!client->sc || !client->sc->cmp()) {
        {
            if(!client->bEnteringPort) {
                if(s.Len() && s!=_T("")) {
                    client->host=s.ToAscii();
                    wxIPV4address a;

                    wxBusyInfo wait(_T("Looking up host..."));

                    bool bOk=true;
                    {
                        AddToLog(log,false,_T("\nLooking up host... "));
                        wxGetApp().Yield();
                        if(!a.Hostname(wxString::FromAscii(client->host.c_str()))) {
                            AddToLog(log,false,_T("HOST NOT FOUND. Type another.\n"));
                            bOk=false;
                        }
                    }
                    if(bOk) {
                        AddToLog(log,false,_T("Okay\nLooking up more host info... "));
                        wxGetApp().Yield();
                        wxString h=a.Hostname();
                        wxString ip=a.IPAddress();
                        if(h==_T(""))
                            h=_T("*** UNKNOWN ***");
                        if(ip==_T(""))
                            ip=_T("*** UNKNOWN ***");
                        string l="Okay.\nHost: ";
                        l+=h.ToAscii();
                        l+="\nIP: ";
                        l+=ip.ToAscii();
                        l+="\nType PORT NUMBER\n";
                        AddToLog(log,false,wxString(l.c_str(),wxConvLocal));
                        client->bEnteringPort=true;
                    }
                }
            } else {
                client->bEnteringPort=false;

                client->port=atol(s.ToAscii());
                string l="Port: ";
                char tmp[64];
                l+=ltoa((long)client->port,tmp,10);
                AddToLog(log,false,wxString(l.c_str(),wxConvLocal));
                AddToLog(log,false,_T("\n     Connecting...\n"));
                wxGetApp().Yield();

                wxBusyInfo wait(_T("Connecting..."));

                if(wxGetApp().frame->conn(client)==1) {
                    AddToLog(log,false,_T("FAILED!\n"));

                    AddToLog(log,false,_T("    Type HOST NAME or IP address and hit ENTER for connection\n"));
                    client->bEnteringPort=false;
                } else {
                    AddToLog(log,false,_T("CONNECTED!\n"));
                }
            }
            return;
        }
    } else {
        wxString str;

        bool bEsc=false;

        for(unsigned int i=0; i<s.Len(); i++) {
            if(client->Conf.bUseEscape) {
                if(s.GetChar(i)==wxChar('\\')) {
                    if(bEsc) {
                        bEsc=false;
                        str << _T("\\");
                    } else
                        bEsc=true;
                    continue;
                }
                if(bEsc && (s.GetChar(i)==wxChar('x') || s.GetChar(i)==wxChar('X'))) {
                    bEsc=false;
                    char res=0;
                    char ch[2];
                    if(++i<s.Len())
                        ch[0]=s.GetChar(i);
                    if(++i<s.Len())
                        ch[1]=s.GetChar(i);

                    if(ch[0]>=wxChar('0') && ch[0]<=wxChar('9')) {
                        ch[0]-=wxChar('0');
                    } else if(ch[0]>=wxChar('a') && ch[0]<=wxChar('f')) {
                            ch[0]-=wxChar('a')-10;
                    } else {
                        if(ch[0]>=wxChar('A') && ch[0]<=wxChar('F'))
                            ch[0]-=wxChar('A')-10;
                    }

                    if(ch[1]>=wxChar('0') && ch[1]<=wxChar('9')) {
                        ch[1]-=wxChar('0');
                    } else if(ch[1]>=wxChar('a') && ch[1]<=wxChar('f')) {
                        ch[1]-=wxChar('a')-10;
                    } else {
                        if(ch[1]>=wxChar('A') && ch[1]<=wxChar('F'))
                            ch[1]-=wxChar('A')-10;
                    }

                    res=(((int)ch[0])<<8)+ch[1];
                    str << wxChar(res);
                    continue;
                }
                if(bEsc && (s.GetChar(i)==wxChar('n') || s.GetChar(i)==wxChar('N'))) {
                    bEsc=false;
                    str << _T("\n");
                    continue;
                }
                if(bEsc && (s.GetChar(i)==wxChar('r') || s.GetChar(i)==wxChar('R'))) {
                    bEsc=false;
                    str << _T("\r");
                    continue;
                }
                if(bEsc && (s.GetChar(i)==wxChar('f') || s.GetChar(i)==wxChar('F'))) {
                    bEsc=false;
                    str << _T("\f");
                    continue;
                }
                if(bEsc && (s.GetChar(i)==wxChar('t') || s.GetChar(i)==wxChar('T'))) {
                    bEsc=false;
                    str << _T("\t");
                    continue;
                }
                if(bEsc && (s.GetChar(i)==wxChar('b') || s.GetChar(i)==wxChar('B'))) {
                    bEsc=false;
                    str << _T("\b");
                    continue;
                }
                if(bEsc && (s.GetChar(i)==wxChar('a') || s.GetChar(i)==wxChar('A'))) {
                    bEsc=false;
                    str << _T("\x7");
                    continue;
                }
                if(bEsc && (s.GetChar(i)==wxChar('\'') || s.GetChar(i)==wxChar('\''))) {
                    bEsc=false;
                    str << _T("'");
                    continue;
                }
                if(bEsc && s.GetChar(i)==wxChar('0')) {
                    bEsc=false;
                    str << _T("\0");
                    continue;
                }
                if(bEsc && s.GetChar(i)==wxChar('"')) {
                    bEsc=false;
                    str << _T('"');
                    continue;
                }
                str << s.GetChar(i);
            } else {
                str << s.GetChar(i);
            }
        }
		str << _T("\r");
        str << _T("\n");
        if(client->Conf.bLocalEcho) {
            if(!Password)
                AddToLog(log,true,str);
            else
                AddToLog(log,false,_T("*** PASSWORD ***\n"));
        }
        int err=0;
        string err_str;

        wxWritableCharBuffer b;
        if(client->Conf.bUTF8) {
            b=str.char_str(wxConvUTF8);
        } else {
            b=str.char_str(wxConvLibc);
        }

        if(!client->sc_ssl) {
            if(client->sc) {
                sock_send(*client->sc,b.data(),strlen(b.data()),err);
            }
        } else {
            wxMutexLocker locker(mutex);
            client->sc_ssl->write(b.data(),strlen(b.data()));
        }
        if(client->log) {
            fwrite(b.data(),1,strlen(b.data()),client->log);
            fflush(client->log);
        }
    }
}

MyPanel::MyPanel(wxFrame *frame)
       : wxPanel(frame, -1)
{
    SetAutoLayout(TRUE);

    wxLayoutConstraints *c;

    m_pasw=new wxCheckBox(this, XTelnet_Pasw_Ctrl, _T("Password"));
    c=new wxLayoutConstraints;
    c->top.SameAs(this,wxHeight,-30);
    c->right.SameAs(this, wxWidth, 0);
    c->left.SameAs(this, wxWidth, -70);
    c->height.Absolute(30);
    m_pasw->SetConstraints(c);
    m_pasw->Show();

    m_enter=new wxTextCtrl(this, XTelnet_Enter_Ctrl, _T(""), wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER);
    c=new wxLayoutConstraints;
    c->left.Absolute(0);
    c->top.SameAs(this,wxHeight,-30);
    c->right.LeftOf(m_pasw);
    c->height.Absolute(30);
    m_enter->SetConstraints(c);
    m_enter->Show();

    {
        wxAcceleratorEntry entries[4];
        entries[0].Set(wxACCEL_NORMAL,WXK_UP,XTelnet_Up);
        entries[1].Set(wxACCEL_NORMAL,WXK_DOWN,XTelnet_Down);
        entries[2].Set(wxACCEL_NORMAL,WXK_F7,XTelnet_F7);
        entries[3].Set(wxACCEL_NORMAL,WXK_ESCAPE,XTelnet_Esc);
        wxAcceleratorTable accel(4,entries);
        m_enter->SetAcceleratorTable(accel);
    }

    m_enter2=new wxTextCtrl(this, XTelnet_Enter_Ctrl, _T(""), wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER | wxTE_PASSWORD);
    c=new wxLayoutConstraints;
    c->left.Absolute(0);
    c->top.SameAs(this,wxHeight,-30);
    c->right.LeftOf(m_pasw);
    c->height.Absolute(30);
    m_enter2->SetConstraints(c);
    m_enter2->Hide();

    m_log=new wxTextCtrl(this, -1, _T(""), wxDefaultPosition, wxDefaultSize, wxTE_MULTILINE | wxTE_RICH | wxTE_AUTO_URL | wxHSCROLL | wxTE_AUTO_SCROLL | wxTE_READONLY);
    c=new wxLayoutConstraints;
    c->left.Absolute(0);
    c->top.Absolute(0);
    c->right.SameAs(this, wxWidth, 0);
    c->bottom.SameAs(this,wxBottom,+30+8);
    m_log->SetConstraints(c);
    m_log->Show();

    c=new wxLayoutConstraints;
    c->top.PercentOf(this,wxWidth,20);
    c->left.PercentOf(this, wxHeight,20);
    c->right.PercentOf(this, wxWidth, 80);
    c->bottom.PercentOf(this, wxHeight, 80);
    m_iterator=new wxListBox(this,XTelnet_Iterator_Ctrl);
    m_iterator->SetConstraints(c);
    m_iterator->Hide();

    m_enter->SetFocus();

    t=new c_thread();
    t->Create();
    t->Run();

    m_timer=new wxTimer(this,XTelnet_Timer);
    m_timer->Start(2000);
}

void MyPanel::OnTimer(wxTimerEvent& WXUNUSED(event))
{
    Freeze();
    m_log->Refresh();
    if(m_iterator_showing) {
        m_iterator->Refresh();
    }
    Thaw();
}

void MyFrame::OnEnter(wxCommandEvent& WXUNUSED(event))
{
    if(m_panel->m_pasw->IsChecked()) {
        m_panel->m_pasw->SetValue(false);
        
        Send(&client,m_panel->m_enter2, m_panel->m_log, true);

        m_panel->m_enter2->Clear();
        m_panel->m_enter2->Hide();
        m_panel->m_enter->Clear();
        m_panel->m_enter->Show();
        m_panel->m_enter->SetFocus();
    } else {
        Send(&client,m_panel->m_enter, m_panel->m_log, false);

        list<wxString>::iterator i1=lines.begin();
    again:
        bool got=false;
        while(i1!=lines.end()) {
            if((*i1)==m_panel->m_enter->GetValue()) {
                got=true;
                break;
            }
            i1++;
        }
        if(!got) {
            lines.push_back(m_panel->m_enter->GetValue());
            m_panel->m_iterator->Append(m_panel->m_enter->GetValue());
        } else {
            delete[] (*i1);
            lines.erase(i1);
            m_panel->m_iterator->Clear();
            i1=lines.begin();
            while(i1!=lines.end()) {
                m_panel->m_iterator->Append(*(i1++));
            }
            goto again;
        }
        lines_iterating=false;
        ShowIterator(false);
    }
    m_panel->m_enter->Clear();
}

void MyFrame::OnPassword(wxCommandEvent& WXUNUSED(event))
{
    if(m_panel->m_pasw->IsChecked()) {
        m_panel->m_enter->Clear();
        m_panel->m_enter->Hide();
        m_panel->m_enter2->Clear();
        m_panel->m_enter2->Show();
        m_panel->m_enter2->SetFocus();
    } else {
        m_panel->m_enter2->Clear();
        m_panel->m_enter2->Hide();
        m_panel->m_enter->Clear();
        m_panel->m_enter->Show();
        m_panel->m_enter->SetFocus();
    }
}

void MyFrame::GetSockError(wxSocketBase* socket, wxString& msg)
{
    wxString r;
    switch(socket->LastError()) {
        case wxSOCKET_NOERROR:
            r=_T("no error");
            break;
        case wxSOCKET_INVOP:
            r=_T("invalid operation");
            break;
        case wxSOCKET_IOERR:
            r=_T("I/O error");
            break;
        case wxSOCKET_INVADDR:
            r=_T("invalid address");
            break;
        case wxSOCKET_INVSOCK:
            r=_T("internal error (invalid socket / uninitialized socket)");
            break;
        case wxSOCKET_NOHOST:
            r=_T("no host found");
            break;
        case wxSOCKET_INVPORT:
            r=_T("invalid port");
            break;
        case wxSOCKET_WOULDBLOCK:
            r=_T("non-critical error - socket would be block-socket");
            break;
        case wxSOCKET_TIMEDOUT:
            r=_T("timeout");
            break;
        case wxSOCKET_MEMERR:
            r=_T("out of memory");
            break;
        default:
            r=_T("unknown error");
            break;
    }
    msg=r;
}

void MyFrame::OnPaste(wxCommandEvent& WXUNUSED(event))
{
    wxMutexLocker locker(mutex);

    if(!client.sc || !client.sc->cmp()) {
        AddToLog(m_panel->m_log,false,_T("Not connected.\n\n    Type HOST NAME or IP address and hit ENTER for connection\n"));
        return;
    }

    wxClipboard c;
    if(c.Open()) {
        if(c.IsSupported(wxDF_TEXT)) {
            wxBusyInfo wait(_T("Sending text from clipboard..."));
            wxGetApp().Yield();

            wxTextDataObject data;
            c.GetData(data);
            {
                bool Password;
                if(m_panel->m_pasw->IsChecked()) {
                    m_panel->m_pasw->SetValue(false);
        
                    Password=true;

                    m_panel->m_enter2->Clear();
                    m_panel->m_enter2->Hide();
                    m_panel->m_enter->Clear();
                    m_panel->m_enter->Show();
                    m_panel->m_enter->SetFocus();
                } else
                    Password=false;
                m_panel->m_enter->Clear();

                size_t size=strlen(data.GetText().ToAscii());
                wxString t;
                t=data.GetText();
                const char* buff=t.ToAscii();
                if(client.Conf.bLocalEcho)
                    if(!Password)
                        AddToLog(main_frame->m_panel->m_log,true,wxString(buff,wxConvLocal));
                        else
                        AddToLog(main_frame->m_panel->m_log,false,_T("*** PASSWORD (PASTE) ***\n"));
                int err=0;
                string err_str;
                if(!client.sc_ssl) {
                    wxMutexLocker locker(mutex);
                    if(client.sc)
                        sock_send(*client.sc,buff,size,err);
                } else {
                    wxMutexLocker locker(mutex);
                    client.sc_ssl->write(buff,size);
                }
                if(client.log) {
                    fwrite(buff,1,size,client.log);
                    fflush(client.log);
                }
            }
        } else
            wxMessageBox(_T("The Clipboard doesn't contains text data."));
        c.Close();
    } else
        wxMessageBox(_T("Cannot open the clipboard."));
}

void MyFrame::OnPasteRecord(wxCommandEvent& event)
{
    OnPaste(event);
    OnRecordLog(event);
}

void MyFrame::OnHide(wxCommandEvent& WXUNUSED(event))
{
    client.bHideOutput=!client.bHideOutput;
    menuBar->Check(XTelnet_Tools_Hide,client.bHideOutput);
}

void MyFrame::OnClose(wxCloseEvent& WXUNUSED(event))
{
    delete m_panel->m_loop;
    m_panel->m_loop=NULL;

    t->Kill();
    t=NULL;

    delete m_panel->m_timer;
    m_panel->m_timer=NULL;

    if(client.sc_ssl) {
        delete client.sc_ssl;
        if(client.sc)
            client.sc->clear();
    }
    client.sc_ssl=NULL;
    
    if(client.sc)
        delete client.sc;
    client.sc=NULL;

    list<wxString>::iterator i1;
    while((i1=lines.begin())!=lines.end()) {
        delete[] (*i1);
        lines.erase(i1);
    }
    if(cnf_frame) {
        cnf_frame->Destroy();
        delete cnf_frame;
        cnf_frame=NULL;
    }
    Destroy();
}

void MyPanel::OnLoop(wxTimerEvent& WXUNUSED(event))
{
    if(client.sc_ssl) {
        wxMutexLocker locker(mutex);
        ssl_loop();
    }
    m_frame->loop(&client);
}

void MyFrame::OnIdle(wxIdleEvent& WXUNUSED(event))
{
    if(client.sc_ssl) {
        wxMutexLocker locker(mutex);
        ssl_loop();
    }
}

